#ifndef ASSOCIATIVE_ARRAY_H

#include "hasher.h"
#include "compable.h"
#include "intobj.h"
#include "strobj.h"


class AssocArr : public HasherTree {
  Comparable *key;

  // these keys are used so we can call the array on char* or ints
  // and the object allocation is hidden from the user's call
  StringObj strkey;
  IntObj intkey;

 public:
  AssocArr(boolean dd=TRUE, boolean dk=TRUE) { key=NULL; SetDelete(dd,dk); }

  AssocArr &operator() (Comparable *c) { key=c; return *this; }
  AssocArr &operator[] (Comparable *c) { key=c; return *this; }

  AssocArr &operator() (Comparable &c) { key=&c; return *this; }
  AssocArr &operator[] (Comparable &c) { key=&c; return *this; }

  AssocArr &operator() ( char *c ) { strkey=c; key=&strkey; return *this; }
  AssocArr &operator[] ( char *c ) { strkey=c; key=&strkey; return *this; }

  AssocArr &operator() ( int i ) { intkey=i; key=&intkey; return *this; }
  AssocArr &operator[] ( int i ) { intkey=i; key=&intkey; return *this; }

  operator void*()   { return (void*)Find( key ); }
  operator char*()   { return ((StringObj*)Find( key ))->String(); }
  operator int()     { return ((IntObj*)Find( key ))->val; }
  operator Object*() { return Find( key ); }
  int operator! () { return (Find(key)!=NULL); }

  Object* operator= ( Object &d ) { return *this = &d; }
  Object* operator= ( Object *d ) {
    if( d==NULL )
      Remove(key);
    else {
      if( DelKey() ) key = (Comparable*)key->Copy();
      if( DelData() ) d = d->Copy();

      Overwrite( *d, key );
      }
    return d;
    }

  virtual void Print (ostream& out = cout) {
    Object *data = Find(key);
    if( !data ) out << "NULL";
    else out << *data; }
  };

#endif
