#include "bddcache.h"


void
BDDCacheKey::Print( ostream& out)
  {
  out << "BDDCacheKey: op=" << op << " if (" << (void *)f << ") then "
      << (void *)g << " else "
      << (void *)h << ";\n";
  }


BDDCache::BDDCache(integer prime, integer prime_bits)
  {
  puts = collisions = hits = total = 0;

  bdd_prime = prime;
  bdd_prime_bits = prime_bits;
  keys = new BDDCacheKey*[bdd_prime+1];
  result = new BDDnode*[bdd_prime+1];
  for( integer i=0 ; i<= bdd_prime ; keys[i++] = NULL);
  }

BDDCache::~BDDCache()
  {
  delete result;
  delete keys;
  }

void
BDDCache::Replace(BDDnode* r, BDDCacheKey& key )
  {
  unsigned bin = key.Hash2();
  BDD_CACHE_MOD( bin );

  result[bin] = r;
  if( keys[bin] )
    keys[bin]->Set(key);
  else
    keys[bin] = new BDDCacheKey(key);
  }

boolean
BDDCache::Get( BDDnode** r, BDDCacheKey& key)
  {
  ++total;
  unsigned bin = key.Hash2();
  BDD_CACHE_MOD( bin );
  if( (keys[bin] != NULL) && !(keys[bin]->NotEqual(&key)) )
    {
    hits++;
    *r = result[bin];
    return TRUE;
    }
  else
    return FALSE;
  }


void
BDDCache::Clear()
  {
  integer i;
  for( i=0 ; i<=bdd_prime ; i++)
    {
    if( keys[i] )
      delete keys[i];
    keys[i] = NULL;
    }
  }

void
BDDCache::ConditionalClear()
  {
  integer i;
  for( i=0 ; i<=bdd_prime ; ++i)
    {
    if( keys[i] && (keys[i]->Flush() || result[i]->Flush()) )
      delete keys[i];
    keys[i] = 0;
    }
  }

//////////////////////////////////////////////////
//
// This "Keeps" all nodes referenced in the cache
//

void
BDDCache::KeepNodes() 
  {
  integer i;
  for( i=0 ; i<=bdd_prime ; ++i)
    if( keys[i] ) {
      result[i]->Reference();
      keys[i]->ReferenceAll();
      }
  }

void
BDDCache::UnKeepNodes() 
  {
  integer i;
  for( i=0 ; i<=bdd_prime ; ++i)
    if( keys[i] ) {
      result[i]->UnReference();
      keys[i]->UnReferenceAll();
      }
  }

void
BDDCache::Print(ostream& out)
{
  out << "\nHashCache\n";
  out << "size      =  " << bdd_prime + 1 << "\n";
  out << "hits      =  " << hits << "\n";
  out << "total     =  " << total << "\n";
  out << "hit ratio =  " << (double)hits / (double)total << "\n";
}

void
BDDCache::PrintLong(ostream& out)
  {
  for( integer bin=0 ; bin<=bdd_prime ; bin++ )
    {
    if( keys[bin] )
      {
      out << bin << "\tBDDCacheKey: " << keys[bin];
      out << "\tBDDfunc: " << result[bin];
      }
    }
  }


