#include "bddhasht.h"
/*
unsigned BDDHashTable::hash_size[] = {
  1,
  2,
  3,
  7,
  13,
  23,
  59,
  113,
  241,
  503,
  1019,
  2039,
  4091,
  8179,
  11587,
  16369,
  23143,
  32749,
  46349,
  65521,
  92683,
  131063,
  185363,
  262139,
  330287,
  416147,
  524269,
  660557,
  832253,
  1048571,
  1321109,
  1664501,
  2097143,
  2642201,
  3328979,
  4194287,
  5284393,
  6657919,
  8388593,
  10568797,
  13315831,
  16777199,
  33554393,
  67108859,
  134217689,
  268435399,
  536870879,
  1073741789,
  2147483629
*/
unsigned BDDHashTable::hash_size[] = {
  1,
  7,
  31,
  127,
  509,
  2039,
  8191,
  32749,
  131071,
  524287,
  2097143,
  8388607,
  33554431,
  134217727,
  536870911,
  2147483476
};


BDDHashTable::BDDHashTable()
{
  hsize = START_SIZE;
  bincount = hash_size[hsize];
  bin = new BDDnode*[bincount];
  for( int i=0 ; i < bincount ; bin[i++] = NULL );
#ifdef DEBUG
  hits = total = puts = rehashCount = maxDepth = 0;
#endif
  count = 0;
}

BDDHashTable::BDDHashTable(integer sz )
{
  hsize = sz;
  bincount = hash_size[hsize];
  bin = new BDDnode*[bincount];
  for( integer i=0 ; i < bincount ; bin[i++] = NULL );
#ifdef DEBUG
  hits = total = puts = rehashCount = maxDepth = 0;
#endif
  count = 0;
}

void
BDDHashTable::Clear()
{
  for (integer i = 0; i < bincount; i++) {
    BDDnode *p;

    while(bin[i]) {
      p = bin[i];
      bin[i] = bin[i]->Next();
      delete p;
    }
  }
#ifdef DEBUG
  hits = total = puts = rehashCount = maxDepth = 0;
#endif
  count = 0;
  Rehash( START_SIZE-hsize);
}

void
BDDHashTable::ConditionalClear()
{
  for (integer i = 0; i < bincount; i++) {
    BDDnode **p = &bin[i];

    while( *p ) {
      if( (*p)->Flush() ) {
	*p = (*p)->Next();
	delete p;
	count--;
      }
      p = &((*p)->next);
    }
  }
  Shrink();
}

void
BDDHashTable::Shrink()
{
  integer i;
  for( i=0 ; hash_size[hsize+i] > count ; i-- );
  
  if( hsize+i < START_SIZE )
    i = START_SIZE - hsize;
  if( i<-1 )
    Rehash(i+1);
}



void
BDDHashTable::Rehash(integer change)
{
#ifdef DEBUG
  rehashCount++;
#endif
  integer old_hsize = hsize;
  integer old_bincount = bincount;
  BDDnode **old_bin = bin;
  
  hsize = hsize+change;
  bincount = hash_size[hsize];
  bin = new BDDnode*[bincount];
  count = 0;
  for( integer i=0 ; i< bincount ; bin[i++] = NULL );

  BDDnode *q, *p;
  while( --old_bincount>=0 ) {
    q = old_bin[old_bincount];
    while( q ) {
      p = q->Next();
      Add( q );
      q = p;
    }
  }
  delete old_bin;
}


// Move these inline eventually?

////////////////////////////////////////////
//
// Add doesn't check for uniqueness
//

void
BDDHashTable::Add( BDDnode *a)
{
  unsigned hashval = a->Hash2();
  BDD_HASH_TABLE_MOD( hashval );
  a->next = bin[hashval];
  bin[hashval] = a;
  count++;
}


boolean
BDDHashTable::Put( BDDnode *a)
{
  unsigned hashval = a->Hash2();
  BDD_HASH_TABLE_MOD( hashval );
  BDDnode *b = bin[hashval];
  while( b )
    {
    if( *b==*a )
      return FALSE;
    b = b->Next();
    }

  a->next = bin[hashval];
  bin[hashval] = a;
  count++;
  Grow();
  return TRUE;
}

/////////////////////////////////////
//
// ForceGet
// If its there, return what's there.
// It not, put it there and return a;
BDDnode*
BDDHashTable::ForceGet( BDDnode* a )
  {
  unsigned hashval = a->Hash2();
  BDD_HASH_TABLE_MOD( hashval );
  BDDnode *b = bin[hashval];

  while( b )
    {
    if( *b==*a )
      return b;
    b=b->Next();
    }
  a->next = bin[hashval];
  bin[hashval] = a;
  count++;
  Grow();
  return a;
  }

BDDnode*
BDDHashTable::Get( BDDnode* a ) {
  unsigned hashval = a->Hash2();
  BDD_HASH_TABLE_MOD( hashval );
  BDDnode *b = bin[hashval];

  while( b ) {
    if( *b==*a )
      return b;
    b = b->Next();
  }
  return NULL;
}

boolean
BDDHashTable::Remove(BDDnode *a)
{
  unsigned hashval = a->Hash2();
  BDD_HASH_TABLE_MOD( hashval );
  BDDnode **b = &(bin[hashval]);
  
  while( *b ) {
    if( *b==a ) {
      *b = (*b)->Next();
      count--;
      return TRUE;
      }
    else
      b = &((*b)->next);
  }
  return FALSE;
}



void
BDDHashTable::PrintStats(ostream &out)
{
  out << "Hash Table Stats for #<HashTable " << (void *)this << "> :\n";
  out << "\tNodes = " << count << "\n";
  out << "\tbins = " << bincount << "\n";
  out << "\tnodes/bin = " << (double)count/bincount << "\n";
#ifdef DEBUG
  out << "\tRehashes = " << rehashCount << "\n";
  out << "\tMax Depth = " << maxDepth << "\n";
  out << "\tPuts = " << puts << endl;
  out << "\tHits = " << hits << endl;
  out << "\tTotal Gets = " << total << endl;
#endif
  
  integer used_bin_count = 0;
  for (integer i = 0; i < bincount; i++) {
    if (bin[i]) {
      used_bin_count++;
    }
  }
  if (count) 
    out << (double)(bincount - used_bin_count)/used_bin_count << "\n";
  else
    out << 0 << "\n";
#ifdef DEBUG
  out << "\tHit ratio = " << (double)hits/total << "\n";
#endif
}


void
BDDHashTable::Print(ostream &out)
{
   out << "Hash Table (" << count << " nodes):\n";
   BDDHashTable_ForEach_Node(this,n) {
	 out << n;
   } BDDHashTable_End_ForEach
}

void
BDDHashTable::PrintHistogram( ostream &out)
{
  int sum=0;
  double   deviation=0;
  int cnt;
  int hist[1000];
  int i=0;
  for( i=0 ; i<1000 ; hist[i++] = 0 );

  BDDnode *o;
  for( i=0 ; i < bincount ; i++ )
    {
    cnt=0;
    for( o = bin[i] ; o ; o=o->Next() )
      cnt++;
    hist[cnt]++;
    sum+=cnt;
    }

  double average = (double)sum/ (double) bincount;

  double count2;
  for( i=0 ; i < bincount ; i++ )
    {
    count2=0;
    for( o = bin[i] ; o ; o=o->Next() )
      count2++;
    deviation += (average - count2) * (average - count2);
    }
  out << "BDDHashTable Stats:\n";
  out << "average = " << average << endl;
  out << "sigma^2 = " << ((double) deviation / (double) bincount) << endl;
  out << "numbins = " << bincount;
  out << "Histogram\t:length\t:number\n";
  for( i=0 ; i<1000 ; i++ )
    {
    if( hist[i] )
      out << "\t\t:" << i << "\t:" << hist[i] << endl;
    }
}
  

