#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "bdd.h"
#include "strobj.h"
#include "hash.h"
#include "list.h"
#include "blifdata.h"


////////////////////////////////////////////////////////////////////////////
//
// A C++ BDD package                                        Andrew Seawright
//
// See Brace, Rudell, and Bryant, "Efficient Implementation of a BDD Package,"
// 27th DAC 1990. Complement edges are not implemented in this version.
// This was done to trade-off efficiency vs. ease of implemetation and ease
// of modification.
//
//	Revised (7/19/93) F. Brewer
//	* changed the inconsistent node creation procedure to prevent
//	some bugs leading to unknown numbers of inputs by the manager.
//
////////////////////////////////////////////////////////////////////////////

void
BDDmanager::PrintLabeledCubes( BDDnode *f, char **var_names, int nl, char eol, ostream& out) {
  int i;

  out << "{\n";
  if (f==one)
    out << "Tautology\n";
  else if (f==zero)
    out << "False\n";
  else {
    string cube = new char[NumInputs()*(nl+2)+1];
    for (i=0; i<NumInputs()*(nl+2); ++i)
      cube[i] = ' ';
    cube[i] = '\0';
    PrintLabeledCubesR(f, cube, out, eol, var_names);
  }
  out << "}";
}
void
BDDmanager::PrintLabeledCubesR( BDDnode *f, string cube, ostream& out,
				char eol, char** var_names, int pos ) {

  if (f == one) {
    cube[pos] = '\0';
    out << cube << eol;
    return;
  } else if (f == zero) {
    return;
  } else {

    int nl = strlen( var_names[f->GetIndex()] );
    cube[pos] = ' ';
    strncpy(cube+pos+1, var_names[f->GetIndex()], nl);
    PrintLabeledCubesR(f->Then(),  cube, out, eol, var_names, pos+nl+1);
    cube[pos] = '~';
    strncpy(cube+pos+1, var_names[f->GetIndex()],nl);
    PrintLabeledCubesR(f->Else(), cube, out, eol, var_names, pos+nl+1);
    cube[pos] = '\0';
  }
}


////////////////////////////////////////////////////////////////////////////
//
// This routine dumps the structure of a single BDD function
// - each node in the function is output once.
// 

void
BDDmanager::PrintStructure(BDDnode *f, ostream& out) {

  HasherTree *ht = new HasherTree;
  out << "Function: " << (void *) f << "\n";
  PrintStructureR(f, ht, out);
  out << "ZERO = " << (void *) zero << "\n";
  out << "ONE = " << (void *) one << "\n";
  delete ht;
}

void
BDDmanager::PrintStructureR(BDDnode *f, HasherTree *ht, ostream& out) {

  if (f == one || f == zero)
    return;

  if (ht->Put(f)) {

    f->PrintWithManager(this,ht,out);
    PrintStructureR(f->Then(),  ht, out);
    PrintStructureR(f->Else(), ht, out);
  }
}

////////////////////////////////////////////////////////////////////////////

IndexArray*
BDDmanager::Minterm2IA( BDDnode *func ) {
  BDDnode *t = func, *f1, *f0;

  int n=0;
  while( t != One() ) {
    f1 = t->Then(); f0 = t->Else();

    if( f0==Zero() ) { n++; t = f1; }
    else             { t = f0;      }
    }
  t = func;
  IndexArray *ia = new IndexArray(n);
  while( t != One() ) {
    f1 = t->Then(); f0 = t->Else();

    if( f0==Zero() ) { ia->Put(t->GetIndex()); t = f1; }
    else             {                         t = f0; }
    }
  return ia;
  }
