#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "bdd.h"


////////////////////////////////////////////////////////////////////////////
//
// A C++ BDD package                                        Andrew Seawright
//
// See Brace, Rudell, and Bryant, "Efficient Implementation of a BDD Package,"
// 27th DAC 1990. Complement edges are not implemented in this version.
// This was done to trade-off efficiency vs. ease of implemetation and ease
// of modification.
//
//	Revised (7/19/93) F. Brewer
//	* changed the inconsistent node creation procedure to prevent
//	some bugs leading to unknown numbers of inputs by the manager.
//
////////////////////////////////////////////////////////////////////////////


void
BDDmanager::doGCfast(boolean clear_hcs) {

  int kept_check=0;
  one->reference = TNODE_MARK;
  zero->reference = TNODE_MARK;

  Hasher_ForEach_Type(keptNodes, IntObj, k, num) {
    BDDnode *key = (BDDnode*)k;
    RGCfast( key );
    key->SetReference( num->val+key->NumReferences()-1 );
    kept_check+= num->val;
    } Hasher_End_ForEach

  if( kept_check != num_keeps )
    cerr << "Warning. Error detected in Kept nodes during GC.\n";

  if( clear_hcs ) {
    computed->Clear();
    cofactor_hc->Clear();
    restrict_hc->Clear();
    }

  BDDHasherArray_ForEach_Type(unique, node) {
    if (node->Marked()) {
      node->UnMark();
    } else {
      unique->Remove(node);
      delete node;
      --num_nodes;
    }
  } BDDHasherArray_End_ForEach;

  one->UnMark(); zero->UnMark();
}

void
BDDmanager::RGCfast(BDDnode *i) {

  i->Reference();

  if( i->Marked() )
    return;

  // Mark it and erase set the ref count to 0
  i->reference = TNODE_MARK;

  RGCfast(i->Then());
  RGCfast(i->Else());
}

