#include "bitstack.h"

BitStack::BitStack() { 

  depth = 0; 
  num_words = 1; 
  stack = new bit_stack_word;
  stack[0] = 0;
}

BitStack::BitStack(BitStack& x) { 

  depth = x.depth; 
  num_words = x.num_words; 
  stack = new bit_stack_word[num_words];
  integer i;
  for (i=0; i<num_words; ++i)
    stack[i] = x.stack[i];
}


BitStack::BitStack(BitStack *x) { 

  depth = x->depth; 
  num_words = x->num_words; 
  stack = new bit_stack_word[num_words];
  integer i;
  for (i=0; i<num_words; ++i)
    stack[i] = x->stack[i];
}

void
BitStack::Push(boolean b) {

  integer i, j;
  Locate(i,j);
  if (b)
    stack[i] |= (BIT_STK_TOP_ONE >> j);
  else
    stack[i] &= ((BIT_STK_TOP_ONE >> j) ^ BIT_STK_ALL_ONES);

  ++depth;
  Locate(i,j);

  if (i+1 > num_words) { // need to grow ?
    bit_stack_word *new_stack = new bit_stack_word[num_words + 1]; 
    for (i=0; i<num_words; ++i)
      new_stack[i] = stack[i];
    delete stack;
    stack = new_stack;
    stack[num_words] = 0;
    ++num_words;
  }
}

void
BitStack::Push(string v) {
 
  for (  ; *v; ++v)
    if (*v == '0') Push0(); else Push1();
}
 
boolean
BitStack::Pop() { 

  integer i, j;
  bit_stack_word w;

  ASSERT( depth>0, "BitStack : pop from empty stack");
  --depth;
  Locate(i, j);
  w = stack[i] & (BIT_STK_TOP_ONE >> j);
  stack[i] &= ((BIT_STK_TOP_ONE >> j) ^ BIT_STK_ALL_ONES);
  return (w?TRUE:FALSE);
}

void
BitStack::ReAlloc() {

  integer i;
  integer min_words = depth / BIT_STK_WORD_SIZE + 1;
  bit_stack_word *new_stack = new bit_stack_word[min_words]; 
  for (i=0; i<min_words; ++i)
    new_stack[i] = stack[i];
  delete stack;
  stack = new_stack;
}

integer 
BitStack::Compare(Comparable* comp) {

  integer i, j;
  bit_stack_word a, b;

  BitStack *x = (BitStack *) comp;
  integer min_depth = MIN(x->depth, depth);
  j = min_depth;

  for (i=0; j >= BIT_STK_WORD_SIZE; ++i, j -= BIT_STK_WORD_SIZE) {

    if (stack[i] > x->stack[i]) return  1;
    if (stack[i] < x->stack[i]) return -1;
  }
  
  if (j) {
    a =    stack[i] & (BIT_STK_ALL_ONES << (BIT_STK_WORD_SIZE-j));
    b = x->stack[i] & (BIT_STK_ALL_ONES << (BIT_STK_WORD_SIZE-j));
    if (a>b) return  1;
    if (a<b) return -1;
  }

  if (depth == x->depth)
    return 0;

  if (depth > x->depth) {
    if  (   stack[i] & (BIT_STK_TOP_ONE >> j))
      return  1;
    else 
      return -1;
  } else
    if  (x->stack[i] & (BIT_STK_TOP_ONE >> j))
      return -1;
    else 
      return  1;
}

unsigned 
BitStack::Hash(unsigned n) { 

  integer i, v;
  v = n;
  for (i=0; i<num_words; ++i)
    v ^= stack[i];
  return (unsigned)v;
}

void
BitStack::Print(ostream& out) {

  integer i;

  out << "\nBitStack : depth=" << depth << " ";
  for (i=0; i<num_words; ++i)
    out << hex << stack[i] << " ";
  out << "\n";
}


