#ifndef BITSTACK_H
#define BITSTACK_H

//
// bitstack.h 
//

// Andrew Seawright

#include "compable.h"

/////////////////////////////////////////// // machine dependant stuff here //

typedef unsigned int bit_stack_word;
#define BIT_STK_WORD_SIZE             32
#define BIT_STK_ALL_ONES              ((bit_stack_word)0xffffffff)
#define BIT_STK_ZERO                  ((bit_stack_word)0)
#define BIT_STK_ONE                   ((bit_stack_word)0x00000001)
#define BIT_STK_TOP_ONE               ((bit_stack_word)0x80000000)

//////////////////////////////////////////////////////////////////////////////

class BitStack : public Comparable {

  integer         depth;
  integer         num_words;
  bit_stack_word *stack;

  void Locate(integer &word, integer &bit) {
    word = depth / BIT_STK_WORD_SIZE;
    bit  = depth % BIT_STK_WORD_SIZE;
  } 

public:
  
  BitStack();
  BitStack(BitStack&);
  BitStack(BitStack*);
 ~BitStack() { if (stack) delete stack;}

  void    Push(boolean);                    // push bit b onto stack    
  void    Push(string);            
  void    Push1() { Push((boolean)TRUE);}   // push a 1 onto the stack
  void    Push0() { Push((boolean)FALSE);}  // push a 0 onto the stack

  boolean Pop();                            // pop a bit off the stack

  integer Depth() { return depth;}
  integer Count() { return depth;}
  void    ReAlloc();               // recycle memory so current stack just fits

  virtual integer Compare(Comparable*);
  virtual unsigned Hash(unsigned = 0);
  virtual void Print(ostream& out = cout);
  virtual integer isa() { return BIT_STACK; }
};

inline ostream& operator << (ostream &out, BitStack& f)
{ f.Print(out); return out; }

inline ostream& operator << (ostream &out, BitStack* f)
{ out << "(ptr->)"; f->Print(out);   return out; }

#endif // BITSTACK_H

