#include "blifdata.h"


BLIFnode::BLIFnode( BLIFnode &bn ) {
  name = new StringObj( bn.name );
  Hasher_ForEach_Type( bn.inputs, BDDnode, k, f ) {
    inputs.Put( f, new StringObj( (StringObj*)k ) );
    }Hasher_End_ForEach
  func = bn.func;
  globalIndex = bn.globalIndex;
  translate = new HasherTree( bn.translate,FALSE );
  }


BLIFdata::BLIFdata( BLIFdata &bd ) {
  name2bddvar.SetDelete(FALSE,FALSE);
  bddvar2name.SetDelete(TRUE,FALSE);
  name2node.SetDelete(TRUE,TRUE);
  keepGlobalFunctions = bd.keepGlobalFunctions;

  model=bd.model;
  reset = bd.reset;
  bddm = bd.bddm;

  List_ForEach_Type( bd.po, StringObj, poname ){
    po.Add( new StringObj(poname) );
    }List_End_ForEach
  po.Reverse();
  List_ForEach_Type( bd.pi, StringObj, piname ){
    pi.Add( new StringObj(piname) );
    }List_End_ForEach
  pi.Reverse();
  List_ForEach_Type( bd.ffout, StringObj, ffoutname ){
    ffout.Add( new StringObj(ffoutname) );
    }List_End_ForEach
  ffout.Reverse();
  List_ForEach_Type( bd.ffin, StringObj, ffinname ){
    ffin.Add( new StringObj(ffinname) );
    }List_End_ForEach
  ffin.Reverse();

  Hasher_ForEach_Type( bd.bddvar2name, StringObj, k, n ){
    StringObj *nn = new StringObj(*n);
    name2bddvar.Put( k, nn );
    bddvar2name.Put( nn, k );
    }List_End_ForEach
  
  Hasher_ForEach_Type( bd.name2node, BLIFnode, k, n ) {
    name2node.Put( new BLIFnode(*n), new StringObj(*(StringObj*)k) );
    }Hasher_End_ForEach
  
  }


BLIFdata::~BLIFdata() {

  Hasher_ForEach_Type( name2bddvar, BDDnode, k, v ) {
    bddm->UnKeep(v);
    }Hash_End_ForEach;


  Hasher_ForEach_Type( name2node, BLIFnode, n, bn ) {
    bddm->UnKeep( bn->Function() );
    }Hasher_End_ForEach;
  po.CleanOut();
  pi.CleanOut();
  ffin.CleanOut();
  ffout.CleanOut();
  }

BDDnode *
BLIFdata::FuncOfPI( StringObj *name ) {

  BLIFnode *bn = (BLIFnode*) name2node.Get( name );
  if( bn == NULL ) return (BDDnode*) name2bddvar.Get( name );

  HasherTree tr;
  int i=0;
  Hasher_ForEach_Type( bn->Inputs(), BDDnode, k, f ) {
    tr.Put( FuncOfPI( (StringObj*)k ), f );
    }List_End_ForEach
  
  BDDnode *f = bddm->Translate( bn->Function(), &tr );
  return f;
  }

List*
BLIFdata::GetPINamesL() {
  List *l = new List(pi);
  return l;
  }

HasherTree *BLIFdata::GetPINamesH(void)
{
  HasherTree *ht = new HasherTree;

  List_ForEach_Type( pi, StringObj, o ){
    ht->Put(new IntObj(GetVar(o)->GetIndex()),o);
  }List_End_ForEach;

  return ht;
}

HasherTree *BLIFdata::GetPONamesH(void)
{
  HasherTree *ht = new HasherTree;

  List_ForEach_Type( po, StringObj, o ){
    ht->Put(new IntObj(GetGlobalFunc(o)->GetIndex()),o);
  }List_End_ForEach;

  return ht;
}

List*
BLIFdata::GetPONamesL() {
  List *l = new List(po);
  return l;
  }

List*
BLIFdata::GetPSNames() {
  List *l = new List(ffout);
  return l;
  }

List*
BLIFdata::GetNSNames() {
  List *l = new List(ffin);
  return l;
  }

List*
BLIFdata::GetFuncNames() {
  List *l = new List;
  Hasher_ForEach_Type( name2node, BLIFnode, k, bn ) {
    l->Add( k );
    }Hasher_End_ForEach
  return l;
  }

// The returned list contains the inputs to fname.
// The items in the list are StringObj.
// clear the list when done, but don't delete the strings
List *BLIFdata::GetFuncInputs( StringObj *fname )
{
  BLIFnode *bn = (BLIFnode*) name2node.Get( fname );
  if( bn ) {
    Object **ins = new Object *[bn->Inputs().Count()];
    Hasher_ForEach_Type( bn->Inputs(), BDDnode, k, f ) {
      ins[f->GetIndex()] = k;
      }Hasher_End_ForEach
    int i;
    List *ret = new List;
    for( i=bn->Inputs().Count()-1 ; i>=0 ; i-- )
      ret->Add( ins[i] );
    delete ins;
    return ret;
    }
  else return NULL;
}

// Don't clear the tree or delete the keys when done
HasherTree *BLIFdata::GetFuncInputsH( StringObj *fname ) {
  BLIFnode *bn = (BLIFnode*) name2node.Get( fname );
  if( bn )
    return &(bn->Inputs());
  else
    return NULL;
  }

// The returned list contains the inputs to fname.
// The items in the list are StringObj.
List *BLIFdata::GetFuncInputs( BDDnode *fname )
{
  StringObj *n = GetVar(fname);
  return GetFuncInputs( n );
}

// Don't clear the tree or delete the keys when done
HasherTree *BLIFdata::GetFuncInputsH( BDDnode *fname ) {
  StringObj *n = GetVar(fname);
  return GetFuncInputsH( n );
  }

// Returns all the gates driven by electrical node "name".
// The items in the HasherTree are IntObj which contain the index of the variable
// The keys are the variable names
HasherTree *BLIFdata::GetDrivenFuncs( StringObj *name, HasherTree *df )
  {
  if( df==NULL )
    df = new HasherTree( TRUE, FALSE );

  Hasher_ForEach_Type( name2node, BLIFnode, k, bn ){  // all gates
    if( bn->Inputs().Get( name ) )
      df->Put( new IntObj(bn->Index()), GetGlobalFunc( (StringObj*)k ) );
    }Hasher_End_ForEach

  return df;
  }


void
BLIFdata::Print( ostream &out ) {

  out << ".model " << model << endl;
  out << ".inputs";
  List_ForEach_Type( pi, StringObj, pinp ) {
    out << " " << pinp;
    }List_End_ForEach
  out << endl;
  
  out << ".outputs";
  List_ForEach_Type( po, StringObj, pout ) {
    out << " " << pout;
    }List_End_ForEach
  out << endl;

  int i=0;
  while( i< ffout.Count() ) {
    out << ".latch " << ffin[i] << " " << ffout[i];
    if( bddm->and( bddm->Input(i), reset ) )
      out << " 1\n";
    else
      out << " 0\n";
    i++;
    }
  
  Hasher_ForEach_Type( name2node, BLIFnode, n, bn ) {
    out << ".names ";
    // need a list, so we get them ordered
    List *ins = GetFuncInputs( (StringObj*)n );
    List_ForEach_Type( ins, StringObj, i ) {
      out << i << " ";
      }List_End_ForEach
    delete ins;
    
    out << n << endl;
    bddm->PrintCubesF( GetLocalFunc( (StringObj*)n ), out, "%C 1\n", "%s" );
    }Hasher_End_ForEach
  
  out << ".end\n" << flush;
  }
  
