#ifndef BITSPARSE_H
#define BITSPARSE_H

////////////////////////////////////////////////////////////////////////////
//
//  bitsparse.h : sparse matrix class 
//  Andrew Seawright
//
////////////////////////////////////////////////////////////////////////////

#include "stddefs.h"
#include "object.h"


////////////////////////////////////////////////////////////////////////////
// macro's from hell (by your friendly kgb agent Chuck Monahan)

#define BitSparse_ForEach_RowInCol( m, c, r) \
{ \
  int r; \
  BitSparseIterator matI; \
  for ( matI.AcrossCol(m, c); !matI.End(); matI.NextRow() ) { \
            r = matI.GetRow();

#define BitSparse_ForEach_ColInRow( m, r, c) \
{ \
  int c; \
  BitSparseIterator matI; \
  for ( matI.AcrossRow(m, r); !matI.End(); matI.NextCol() ) { \
            c = matI.GetCol();

#define BitSparse_ForEach_Cell( m, r, c) \
{ \
  int c; \
  BitSparseIterator matI; \
  for ( int r = 0; r < m->Getysize(); r++) \
    for ( matI.AcrossRow(m, r); !matI.End(); matI.NextCol() ) { \
            c = matI.GetCol();

#define BitSparse_End_ForEach }}

// You are now leaving the macro zone
///////////////////////////////////////////////////////////////////////////


class BitSparseCell : public Object {
protected:

  int x;
  int y;
  BitSparseCell *nextx;
  BitSparseCell *nexty;

  friend class BitSparseIterator;
  friend class BitSparse;
};

class BitSparse : public Object {
protected:

  int             xsize;          //   number of columns
  int             ysize;          //   number of rows
  int             count;          //   total number of set points
  int            *num_in_row;     //   array of numbers of entries in each row
  int            *num_in_col;     //   array of numbers of entries in each col
  BitSparseCell **x;              
  BitSparseCell **y;

public:

  BitSparse(int, int);
  BitSparse(istream&, boolean load = TRUE);
 ~BitSparse();
  virtual void Set  (int, int);

  void    Clear(int, int);
  boolean Test (int, int);
  boolean TestCol (int);
  boolean TestRow (int);
  void    Resizex (int);
  void    RowMoveI(int, int, int, int);
  void    RowUnMoveI(int a, int b, int c, int d) {RowMoveI(a, b, c, d);}

  int     Getxsize ()            { return xsize;}
  int     Getysize()             { return ysize;}
  int     Count()                { return count;}
  int     GetNumInRow(int row)   { return num_in_row[row];}
  int     GetNumInCol(int col)   { return num_in_col[col];}
  int     GetFirstInRow(int row) { return ( y[row])? y[row]->x: -1;}
  int     GetFirstInCol(int col) { return ( x[col])? x[col]->y: -1;}

  virtual integer isa()          { return BIT_SPARSE;}
  virtual void Print( ostream& out = cout);

  void   Set(   istream& in);
  void   Clear( istream& in);

  friend class BitSparseIterator;
};

inline ostream& operator << (ostream &out, BitSparse& f)
{ f.Print(out); return out; }

inline ostream& operator << (ostream &out, BitSparse* f)
{ out << "(ptr->)"; f->Print(out); return out; }

class BitSparseIterator : public Object {
protected:
  BitSparseCell *tmp;

public:
  BitSparseIterator()                     { tmp=0;}
 ~BitSparseIterator()                     {;}
  void AcrossRow(BitSparse *m, int row)   { tmp = m->y[row];}
  void AcrossCol(BitSparse *m, int col)   { tmp = m->x[col];}
  void NextRow()                          { tmp = tmp->nexty;}
  void NextCol()                          { tmp = tmp->nextx;}
  boolean End()                           { return (boolean) (tmp==0);}
  int GetRow()                            { return tmp->y;}
  int GetCol()                            { return tmp->x;}
};

#endif // BITSPARSE_H

