#ifndef DG_H
#define DG_H
////////////////////////////////////////////////////////////////////////////
//
//  dg.h : a directed graph package
//  Andrew Seawright 
//  1/30 modified 6/26/91, 7/28/91
//
////////////////////////////////////////////////////////////////////////////

#include "list.h"

////////////////////////////////////////////////////////////////////////////
// macro's from hell

#define DG_ForEach_Node(g, n) \
{ \
  dg_node *n; \
  for ( List_iterator list__itr((g).nodes); !list__itr.end(); ++list__itr ) { \
            n = (dg_node *)list__itr.item();

#define DG_ForEach_Node_Object_Data(g, o) \
{ \
  Object *o; \
  for ( List_iterator list__itr((g).nodes); !list__itr.end(); ++list__itr ) { \
            o = ((dg_node *)list__itr.item())->Get();

#define DG_ForEach_Node_Type_Data(g, t, o) \
{ \
  t *o; \
  for ( List_iterator list__itr((g).nodes); !list__itr.end(); ++list__itr ) { \
            o = (t *) ((dg_node *)list__itr.item())->Get();

#define DG_ForEach_Edge(g, e) \
{ \
  dg_edge *e; \
  for ( List_iterator list__itr((g).edges); !list__itr.end(); ++list__itr ) { \
            e = (dg_edge *)list__itr.item();
 
#define DG_ForEach_Edge_Object_Data(g, o) \
{ \
  Object *o; \
  for ( List_iterator list__itr((g).edges); !list__itr.end(); ++list__itr ) { \
            o = ((dg_edge *)list__itr.item())->Get();

#define DG_ForEach_Edge_Type_Data(g, t, o) \
{ \
  t *o; \
  for ( List_iterator list__itr((g).edges); !list__itr.end(); ++list__itr ) { \
            o = (t *) ((dg_edge *)list__itr.item())->Get();

#define DG_ForEach_Out_Node(n, onode) \
 { \
   dg_node *onode; \
   for ( List_iterator list__itr((n).out_edges); !list__itr.end(); ++list__itr ){\
            onode = ((dg_edge *)list__itr.item())->to_node;

#define DG_ForEach_Out_Edge(n, e) \
 { \
   dg_edge *e; \
   for ( List_iterator list__itr((n).out_edges); !list__itr.end(); ++list__itr ){\
            e = (dg_edge *)list__itr.item();

#define DG_ForEach_In_Node(n, inode) \
 { \
   dg_node *inode; \
   for ( List_iterator list__itr((n).in_edges); !list__itr.end(); ++list__itr ){\
            inode = ((dg_edge *)list__itr.item())->from_node;

#define DG_ForEach_In_Edge(n, e) \
 { \
   dg_edge *e; \
   for ( List_iterator list__itr((n).in_edges); !list__itr.end(); ++list__itr ){\
            e = (dg_edge *)list__itr.item();


#define DG_End_ForEach }}

////////////////////////////////////////////////////////////////////////////

class dg_node : public Object {

  friend class dg;
  friend class dg_edge;

protected:   
  integer  node_num;   // node's number
  Object  *data;       // node's data 
  boolean  visit;

public:
  List out_edges;      // fan-out edge list
  List in_edges;       // fan-in edge list

  dg_node()          { data = 0; visit = FALSE;}
  dg_node(Object* o) { data = o; visit = FALSE;}
 ~dg_node()          {;}

  void RemoveInEdge (dg_edge *e);
  void RemoveOutEdge(dg_edge *e);

  integer Fanout()         { return out_edges.Count();}
  integer Fanin()          { return in_edges.Count();}
  integer GetNum()         { return node_num;}
  void    Put(Object* o)   { data = o;}
  Object *Get()            { return data;}
  void    Touch(boolean v) { visit = v;}
  boolean Touched()        { return visit;}

  virtual integer isa()    { return DG_NODE;}
  virtual void Print( ostream& out = cout);
  virtual void PrintII( ostream& out, boolean); 
};  

inline ostream& operator << (ostream &out, dg_node& f)
{ f.Print(out); return out; }

inline ostream& operator << (ostream &out, dg_node* f)
{ out << "(ptr->)"; f->Print(out); return out; }

////////////////////////////////////////////////////////////////////////////

class dg_edge : public Object {

  friend class dg;
  friend class dg_node;

protected:   
  Object  *data;         // edge's data object

public: 
  dg_node *to_node;      // the node this edge points to
  dg_node *from_node;    // the node this edge points from

  dg_edge()           { to_node = from_node = 0; data = 0;}
  dg_edge( Object* o) { to_node = from_node = 0; data = o;}
 ~dg_edge() {;}

  void     Put(Object* o){ data = o;}
  Object  *Get()         { return data;}
  dg_node *ToNode()      { return to_node;}
  dg_node *FromNode()    { return from_node;}

  virtual integer isa()  { return DG_EDGE;}
  virtual void Print( ostream& out = cout);
  virtual void PrintII( ostream& out, boolean); 
};

inline ostream& operator << (ostream &out, dg_edge& f)
{ f.Print(out); return out; }

inline ostream& operator << (ostream &out, dg_edge* f)
{ out << "(ptr->)"; f->Print(out); return out; }

////////////////////////////////////////////////////////////////////////////

class dg : public Object {

  void TopSort( dg_node *n, List* l );

protected:
  dg_node  *source;  //  pointer to node - typically a source, or start state
  dg_node  *sink;    //  pointer to node - typically a sink, or accepting state
  Object   *data;    //  graph's data object
  
public:
  List      nodes;   // list of all nodes in this graph 
  List      edges;   // list of all edges in this graph

  dg();
 ~dg();

  boolean  
    printNode,
    printEdge;

  void    Put(Object* o) { data = o;}  
  Object* Get()          { return data;}  

  dg_edge *AddEdge(dg_node&, dg_node&, dg_edge*);
  dg_edge *AddEdge(dg_node*, dg_node*, dg_edge*);
  boolean  RemoveEdges(dg_node*, dg_node*);
 
  dg_node *AddNode(dg_node& n) { return AddNode( &n);}
  dg_node *AddNode(dg_node*);
  dg_node *AppendNode(dg_node*);
  void     RemoveNode(dg_node*);

  void AddGraph(dg*);
  void Renumber();
  void TransferGraph(dg *);
  void AdjustSourceFirst();
  void MarkConnectedNodes(dg_node *);
  void BackMarkConnectedNodes(dg_node *);
  void PruneUnconnectedNodes(dg_node *);

  void TouchNodes(boolean b) { DG_ForEach_Node(*this,node) {
	                         node->Touch(b);
                               } DG_End_ForEach
                             }
  boolean  isAcyclic();
  boolean  isAcyclic(dg_node *node);
  List*    TopologicalSortNodes();
  List*    TopologicalSortData();

  void     Putsource(dg_node *s) { source = s;}
  void     Source(dg_node *s)    { source = s; }
  dg_node *Getsource()           { return source;}
  dg_node *Source()              { return source;}
  void     Putsink(dg_node *s)   { sink = s;}
  void     Sink(dg_node *s)      { sink = s; }
  dg_node *Getsink()             { return sink;}
  dg_node *Sink()                { return sink;}
  
  void     Read( istream &in = cin);
  virtual  integer isa()         { return DG;}
  virtual  void    Print( ostream& out = cout) ; 
  virtual  void    Scan ( istream&  in = cin) { Read(in);}
};

inline ostream& operator << (ostream &out, dg& f)
{ f.Print(out); return out; }

inline ostream& operator << (ostream &out, dg* f)
{ out << "(ptr->)"; f->Print(out); return out; }

#endif // DG_H

