// bdd_test.C
// Tony Stornetta
//
// Test program - this program generates very, very random BDD trees
// just give it a number of levels, maximum function space and number
// of iterations.  If the function space fills up, then the program
// just generates more random functions in the BDD.

#include <iostream.h>
#include <stdlib.h>
#include "bdd.h"

/*  Some sample values
#define MAXLEVELS 100
#define MAXFN 5000
#define ITERATIONS 100
*/

int MAXLEVELS, MAXFN, ITERATIONS;

// The CHOICE macro chooses an input variable 25% of the time and
// a function 75% of the time...
#define RAND(MAX) ((rand()>>4) % MAX)
#define ONE_IN_4 ((rand()>>4)%4 == 0)
#define CHOICE (ONE_IN_4 ? RAND(MAXLEVELS+2) : MAXLEVELS+2+RAND(fn-MAXLEVELS-2))

typedef BDDnode* BDDnodePtr;

void print(char *s, int a, int b)
{
  cout << s << '(';
  if (a == 0)
    cout << "Zero";
  else if (a == 1)
    cout << "One";
  else if (a < MAXLEVELS+2)
    cout << "Var #" << a - 2;
  else
    cout << "Fn #" << a - MAXLEVELS - 2;
  if (b == 0)
    cout << ", Zero";
  else if (b == 1)
    cout << ", One";
  else if (b < MAXLEVELS+2 && b >= 0)
    cout << ", Var #" << b - 2;
  else if (b >= 0)
    cout << ", Fn #" << b - MAXLEVELS - 2;
  cout << ')' << endl;
}

int main(int argc, char**argv)
{
  BDDmanager bdd;
  BDDnodePtr *functions;
  int i, fn;

  if (argc != 4)
  {
    fprintf (stderr, "Usage: %s <# levels> <# functions> <# iterations>\n",
								argv[0]);
    exit (1);
  }

  MAXLEVELS = atoi (argv[1]);
  MAXFN = atoi (argv[2]); ITERATIONS = atoi (argv[3]);

  if (MAXLEVELS+2 > MAXFN)
  {
    fprintf (stderr, "Can't have more levels than functions!\n");
    exit (1);
  }

  cout << "Levels = " << MAXLEVELS << endl;
  cout << "Functions = " << MAXFN << endl;
  cout << "Iterations = " << ITERATIONS << "\n\n";

  functions = new BDDnodePtr[MAXFN];

  // Set input variables
  functions[0] = bdd.Zero();
  functions[1] = bdd.One();
  for (i=2; i<MAXLEVELS+2; i++)
    functions[i] = bdd.Input(i-2);
  fn = MAXLEVELS+2;

  // Seed random generator
  srand (123);

  // Set an initial function
  functions[fn++] = bdd.not(functions[2]);

  for (i=0; i<ITERATIONS; i++)
  {
    int c1, c2;

    switch ((rand() >> 2) % 8)
    {
    case 0:
      // Invert
      c1 = CHOICE;
      if (fn < MAXFN)
        functions[fn++] = bdd.not(functions[c1]);
      else
        bdd.not(functions[c1]);
      print ("not ", c1, -1);
      break;
    case 1:
      // And
      c1 = CHOICE;
      c2 = CHOICE;
      if (fn < MAXFN)
        functions[fn++] = bdd.and (functions[c1], functions[c2]);
      else
        bdd.and(functions[c1], functions[c2]);
      print ("and ", c1, c2);
      break;
    case 2:
      // Or
      c1 = CHOICE;
      c2 = CHOICE;
      if (fn < MAXFN)
        functions[fn++] = bdd.or (functions[c1], functions[c2]);
      else
        bdd.or(functions[c1], functions[c2]);
      print ("or  ", c1, c2);
      break;
    case 3:
      // Xor
      c1 = CHOICE;
      c2 = CHOICE;
      if (fn < MAXFN)
        functions[fn++] = bdd.xor (functions[c1], functions[c2]);
      else
        bdd.xor(functions[c1], functions[c2]);
      print ("xor ", c1, c2);
      break;
    case 4:
      // Nand
      c1 = CHOICE;
      c2 = CHOICE;
      if (fn < MAXFN)
        functions[fn++] = bdd.nand (functions[c1], functions[c2]);
      else
        bdd.nand(functions[c1], functions[c2]);
      print ("nand", c1, c2);
      break;
    case 5:
      // Nor
      c1 = CHOICE;
      c2 = CHOICE;
      if (fn < MAXFN)
        functions[fn++] = bdd.nor (functions[c1], functions[c2]);
      else
        bdd.nor(functions[c1], functions[c2]);
      print ("nor ", c1, c2);
      break;
    case 6:
      // Xnor
      c1 = CHOICE;
      c2 = CHOICE;
      if (fn < MAXFN)
        functions[fn++] = bdd.xnor (functions[c1], functions[c2]);
      else
        bdd.xnor(functions[c1], functions[c2]);
      print ("xnor", c1, c2);
      break;
    case 7:
      // Diff
      c1 = CHOICE;
      c2 = CHOICE;
      if (fn < MAXFN)
        functions[fn++] = bdd.diff (functions[c1], functions[c2]);
      else
        bdd.diff(functions[c1], functions[c2]);
      print ("diff", c1, c2);
      break;
    }
  }

  bdd.Print();
  cout << endl;

  bdd.XDump();
  delete functions;

  return 0;
}

