////////////////////////////////////////////////////////////////////////////
//
//  dg.C : (a directed graph package)  
//  Andrw Seawright 
//  1/30/91 modified 6/26/91, 7/28/91
//
////////////////////////////////////////////////////////////////////////////

#include <iostream.h>
#include "fa.h"

////////////////////////////////////////////////////////////////////////////

void
fa_node::Print(ostream &out) {
  out << "fa_node: " << node_num << " ";
}

void
fa_node::PrintII(ostream &out, boolean printData, boolean printState) {
  out << node_num << " ";
  
  if ( printState)
    out << " {" << state << "} ";
  if ( printData) {
    if (data != 0)
      data->Print(out);
  }
}

/////////////////////////////////////////////////
/////////////////////////////////////////////////
/////////////////////////////////////////////////
/////////////////////////////////////////////////

void
fa_edge::Print( ostream& out) {
  out << "fa_edge: " << transition << " ";
}

void
fa_edge::PrintII( ostream& out, boolean printData) {
  if (printData)
    if (data != 0)
      data->Print( out);
    else
      out << " ";
}

/////////////////////////////////////////////////
/////////////////////////////////////////////////
/////////////////////////////////////////////////
/////////////////////////////////////////////////

//
// Simple input routine
//

void
fa::Read( istream&  in) {
  int i, N, E, n1, n2, t;
  
  in >> N;  in >> E;
  dg_node **tmp_nodes = (dg_node**) new fa_node*[N];
  
  for (i=0; i<N; ++i)
    tmp_nodes[i] = AddNode(new fa_node);
  
  for (i=0; i<E; ++i) {
    in >> n1; in >> n2; in >> t;
    AddEdge( tmp_nodes[n1], tmp_nodes[n2], new fa_edge(t));
  }
  Putsource(tmp_nodes[0]);
  delete tmp_nodes;
}

void
fa::PrintTransitionTable(int N, ostream& out) {
  
  dg_node *n;
  int      i;
  
  out << "\nTransition table:\n";
  out << "-----------------\n\t";
  for (i=0; i<N; ++i)
    out << '\t' << i;
  out << endl;
  
  out << "\t\t";
  for (i=0; i<N; ++i)
    out << "----\t";
  out << endl;
  
  List *l = new List;
  FA_ForEach_Node(*this, node) {
    l->Add(*node);
  } FA_End_ForEach;
  
  List_ForEach_Type(*l, fa_node, node) {
    out << node->GetNum() << ":\t";
    for (i=0; i<N; ++i) {
      FA_ForEach_Out_Edge(*node, e) {
	n = e->ToNode();
	if (e->GetTransition() == i)
	  out << "\t" << n->GetNum();
      } FA_End_ForEach;
    }
    out << "\n";
  } List_End_ForEach;
  delete l;
}

void
fa::Print( ostream& out) {
  out << "\nfa graph: ";
  if (data)
    data->Print( out);
  else
    out << "\n";
  out << "number of nodes: " << nodes.Count() << "\n";
  out << "number of edges: " << edges.Count() << "\n";
  
  FA_ForEach_Node(*this, n) {
    out << "[" << n->node_num << "]";
    n->PrintII( out, printNode, printState);  
    out << endl;
    
    out << " to: ";
    FA_ForEach_Out_Edge(*n, oe) {
      dg_node *on = oe->ToNode();
      out << on->GetNum() << ":";
      oe->PrintII( out, printEdge);
    } FA_End_ForEach; 
    out << endl;
    
    out << " from: ";
    FA_ForEach_In_Edge(*n, ie) {
      dg_node *in = ie->FromNode();
      out << in->GetNum() << ":";
      ie->PrintII( out, printEdge);
    } FA_End_ForEach; 
    out << "\n";
    
  } FA_End_ForEach;
}

////////////////////////////////////////////////////////////////////////////


