#ifndef FA_H
#define FA_H
////////////////////////////////////////////////////////////////////////////
//
//  dg.h : a directed graph package
//  Andrew Seawright 
//  1/30 modified 6/26/91, 7/28/91
//
////////////////////////////////////////////////////////////////////////////

#include "dg.h"

#define EPSILON -1 

////////////////////////////////////////////////////////////////////////////
// macro's from hell

#define FA_ForEach_Node(x, o) \
{ \
  fa_node *o; \
  for ( List_iterator list__itr((x).nodes); !list__itr.end(); ++list__itr ) { \
            o = (fa_node *)list__itr.item();

#define FA_ForEach_Edge(x, o) \
{ \
  fa_edge *o; \
  for ( List_iterator list__itr((x).edges); !list__itr.end(); ++list__itr ) { \
            o = (fa_edge *)list__itr.item();
 
#define FA_ForEach_Out_Node(x, o) \
 { \
   fa_node *o; \
   for ( List_iterator list__itr((x).out_edges); !list__itr.end(); ++list__itr ){\
            o = (fa_node *) ((dg_edge *)list__itr.item())->to_node;

#define FA_ForEach_Out_Edge(x, e) \
 { \
   fa_edge *e; \
   for ( List_iterator list__itr((x).out_edges); !list__itr.end(); ++list__itr ){\
            e = (fa_edge *)list__itr.item();

#define FA_ForEach_In_Node(x, o) \
 { \
   fa_node *o; \
   for ( List_iterator list__itr((x).in_edges); !list__itr.end(); ++list__itr ){\
            o = (fa_node *) ((dg_edge *)list__itr.item())->from_node;

#define FA_ForEach_In_Edge(x, e) \
 { \
   fa_edge *e; \
   for ( List_iterator list__itr((x).in_edges); !list__itr.end(); ++list__itr ){\
            e = (fa_edge *)list__itr.item();

#define FA_End_ForEach }}

////////////////////////////////////////////////////////////////////////////

class fa_node : public dg_node {

  friend class fa;
  friend class fa_edge;

public:
  integer  state;      // generic 'state' number
   
  fa_node()                    :dg_node( ){ state = -1;}
  fa_node( Object* o, int s=-1):dg_node(o){ state =  s;}
 ~fa_node() {;}

  void    SetState(integer n) { state = n;}
  integer GetState()          { return state;}
  integer State()             { return state;}
 
  virtual integer isa()       { return FA_NODE;}
  virtual void Print( ostream& out = cout);
  virtual void PrintII( ostream& out, boolean b) { PrintII( out, b, TRUE);}
  virtual void PrintII( ostream& out, boolean, boolean);
};  

inline ostream& operator << (ostream &out, fa_node& f)
{ f.Print(out); return out; }

inline ostream& operator << (ostream &out, fa_node* f)
{ out << "(ptr->)"; f->Print(out); return out; }

////////////////////////////////////////////////////////////////////////////

class fa_edge : public dg_edge {

  friend class fa;
  friend class fa_node;

protected:   
  boolean  visit;      // generic flag

public: 
  integer  transition;    // generic 'transition' value
  
  fa_edge()         :dg_edge() { visit = FALSE; transition = -1;}
  fa_edge(integer t):dg_edge() { visit = FALSE; transition =  t;}
  fa_edge(Object* o, integer t=-1):dg_edge(o) { visit = FALSE; transition =  t;}
 ~fa_edge() {;}

  void    Touch(boolean v)         { visit = v;}
  boolean Touched()                { return visit;}
  void    SetTransition(integer n) { transition = n;}
  integer GetTransition()          { return transition;}
  integer Transition()             { return transition;}

  virtual integer isa()            { return FA_EDGE;}
  virtual void Print( ostream& out = cout);     
  virtual void PrintII( ostream& out, boolean); 
};

inline ostream& operator << (ostream &out, fa_edge& f)
{ f.Print(out); return out; }

inline ostream& operator << (ostream &out, fa_edge* f)
{ out << "(ptr->)"; f->Print(out); return out; }

////////////////////////////////////////////////////////////////////////////

class fa : public dg {

public:
  boolean   printState;

  fa():dg() {;}
 ~fa()      {;}

  void TouchEdges(boolean b) { FA_ForEach_Edge(*this,edge) {
                                 edge->Touch(b);
                               } FA_End_ForEach
                             }
  void    Read ( istream&  in = cin);

  virtual integer isa() { return FA;}
  virtual void    PrintTransitionTable( int timesteps, ostream& out = cout);
  virtual void    Print( ostream& out = cout);
  virtual void    Scan ( istream&  in = cin ) { Read( in);}
};

inline ostream& operator << (ostream &out, fa& f)
{ f.Print(out); return out; }

inline ostream& operator << (ostream &out, fa* f)
{ out << "(ptr->)"; f->Print(out); return out; }

#endif // FA_H

