#ifndef FLOATOBJ_H      
#define FLOATOBJ_H

#include <iostream.h>
#include "compable.h"

#define FLOAT_PRIME 7919

class FloatObj: public Comparable {
public:
  float val;

  FloatObj( void)    { val = 0;}
  FloatObj( float v) { val = v;}

  float operator=( float i) { val = i; return i;}
  operator float()      { return val;}

  virtual integer isa() { return FLOAT_OBJ;}
  virtual Object* Copy(){ return new FloatObj( val);}   

  virtual integer Compare( Comparable* c) {    // Cannot just return diff (round off errors!
    ASSERT (c->isa() == FLOAT_OBJ, "expecting FloatObj");
    float diff = val - ((FloatObj*) c)->val;
    if ( diff == 0.0) return  0;
    if ( diff <  0.0) return -1;
    return 1;
  }

  virtual uinteger Hash( unsigned i=0) {
    return (uinteger) (((uinteger)val+i)^0x5ab);}
  virtual void Print( ostream& out = cout){ out << val;}
  virtual void Scan ( istream& in  = cin ){  in >> val;}

#ifdef _BIG_ENDIAN
  virtual void Write( ostream& out ) {
    out.write( (unsigned char*)&val, sizeof(val) );
    }
  static void WriteFloat( ostream& out, float v ) {
    out.write( (unsigned char*)&v, sizeof(v) );
    }
  virtual void Read( istream& in ) {
    in.read( (unsigned char*)&val, sizeof(val) );
    }
  static float ReadFloat( istream& in ) {
    float v;
    in.read( (unsigned char*)&v, sizeof(v) );
    return v;
    }
#else
  virtual void Write( ostream& out ) {
    int i;
    unsigned char* p = (unsigned char*)&val;
    for( i=sizeof(val)-1 ; i>=0 ; i-- )
      out.put(p[i]);
    }
  static void WriteFloat( ostream& out, float v ) {
    int i;
    unsigned char* p = (unsigned char*)&v;
    for( i=sizeof(v)-1 ; i>=0 ; i-- )
      out.put(p[i]);
    }
  virtual void Read( istream& in ) {
    int i;
    unsigned char* p = (unsigned char*)&val;
    for( i=sizeof(val)-1 ; i>=0 ; i-- )
      in.get(p[i]);
    }
  static float ReadFloat( istream& in ) {
    int i;
    float v;
    unsigned char *p = (unsigned char*)&v;
    for( i=sizeof(v)-1 ; i>=0 ; i-- )
      in.get(p[i]);
    return v;
    }
#endif
};


class DoubleObj: public Comparable {
public:
  double val;

  DoubleObj( void)     { val = 0;}
  DoubleObj( double v) { val = v;}
  virtual ~DoubleObj() {};

  double operator=( double i) { val = i; return i;}
  operator double()     { return val;}

  virtual integer isa() { return DOUBLE_OBJ;}
  virtual Object* Copy(){ return new DoubleObj( val);}   

  virtual integer Compare( Comparable* c) {
    ASSERT (c->isa() == DOUBLE_OBJ, "expecting DoubleObj");
    double diff = val - ((DoubleObj*) c)->val;
    if ( diff == 0.0) return  0;
    if ( diff <  0.0) return -1;
    return 1;
  }

  virtual uinteger Hash( unsigned i=0) {
    return (uinteger) (((uinteger)val+i)^0x5ab);}
  virtual void Print( ostream& out = cout){ out << val;}
  virtual void Scan ( istream& in  = cin ){  in >> val;}

#ifdef _BIG_ENDIAN
  virtual void Write( ostream& out ) {
    out.write( (unsigned char*)&val, sizeof(val) );
    }
  static void WriteDouble( ostream& out, double v ) {
    out.write( (unsigned char*)&v, sizeof(v) );
    }
  virtual void Read( istream& in ) {
    in.read( (unsigned char*)&val, sizeof(val) );
    }
  static double ReadDouble( istream& in ) {
    double v;
    in.read( (unsigned char*)&v, sizeof(v) );
    return v;
    }
#else
  virtual void Write( ostream& out ) {
    int i;
    unsigned char *p = (unsigned char*)&val;
    for( i=sizeof(val)-1 ; i>=0 ; i-- )
      out.put(p[i]);
    }
  static void WriteDouble( ostream& out, double v ) {
    int i;
    unsigned char *p = (unsigned char*)&v;
    for( i=sizeof(v)-1 ; i>=0 ; i-- )
      out.put( p[i] );
    }
  virtual void Read( istream& in ) {
    int i;
    unsigned char *p = (unsigned char*)&val;
    for( i=sizeof(val)-1 ; i>=0 ; i-- )
      in.get( p[i] );
    }
  static double ReadDouble( istream& in ) {
    int i;
    double v;
    unsigned char *p = (unsigned char*)&v;
    for( i=sizeof(v)-1 ; i>=0 ; i-- )
      in.get( p[i] );
    return v;
    }
#endif

};

#endif // FLOATOBJ_H

