////////////////////////////////////////////////////////////////////////////
//
//  hash.c : hash tree object
//  Andrew Seawright
//  $Header: /usr6/research/code/homebrew/RCS/hash.C,v 1.14 1996/12/14 02:20:40 crews Exp $
//  
////////////////////////////////////////////////////////////////////////////

#include <string.h>
#include <stdlib.h>
#include <iostream.h>
#include "list.h"
#include "hash.h"

////////////////////////////////////////////////////////////////////////////

HashNode::HashNode(int hv, void *k, Object *o) {   // HashNode constructor

  right   =  left  =  0;
  hashval =  hv;
  key     =  k;
  data    =  o;
}

HashNode::HashNode(HashNode *in, boolean copy)
{
   hashval = in->hashval;
   key     = in->key;
   data    = copy ? in->data->Copy() : in->data;

   if (in->right != NULL) {
      right = new HashNode(in->right, copy);
   }  else {
      right = NULL;
   }

   if (in->left != NULL) {
      left = new HashNode(in->left, copy);
   }  else {
      left = NULL;
   }
}

////////////////////////////////////////////////////////////////////////////

HashTree::HashTree(boolean d, boolean k) {                            // HashTree constructor
  
  count   =  0;
  root    =  0;
  del_data=  (d) ? TRUE:FALSE;
  del_key =  (k) ? TRUE:FALSE;
}

HashTree::HashTree(HashTree *tree, boolean copy) {

   count = tree->count;
   if (tree->root)
     root = new HashNode(tree->root, copy);
   else
     root = 0;
}

////////////////////////////////////////////////////////////////////////////

void
HashTree::HashDelete(HashNode *p) {    // recursive destruction of tree 

  if (p->right != NULL) {
     HashDelete(p->right);
     p->right = NULL;
  }
  if (p->left != NULL) {
     HashDelete(p->left);
     p->left = NULL;
  }
  if (DelData()) delete p->data;
  if (DelKey() ) delete p->key;

  delete p;
}

////////////////////////////////////////////////////////////////////////////

boolean
HashTree::Put(Object& o, char *key)
{
  register HashNode *q = NULL, *p = root;
  register integer hashval = hashf(key, 0);

  for (;;) {
    if (p == NULL) { 
      p = new HashNode(hashval, key, &o);         // build a new node
      if (root == NULL)
	root = p;
      else {
	if (q->hashval > hashval)
	  q->right = p;
	else
	  q->left = p;
      }
      ++count;
      return TRUE;
    } 
    else if (p->hashval == hashval) { 
      if ( strcmp(key, (char*) p->key) == 0) 
	return FALSE;                    // key already in tree
      else 
	hashval = hashf(key, hashval); // hash collision, rehash 
    }
    q = p;
    
    if (p->hashval > hashval)  
      p=p->right;
    else
      p=p->left;
  }
}

////////////////////////////////////////////////////////////////////////////

void
HashTree::ForcePut(Object& o, char *key)
{
  register HashNode *q = NULL, *p = root;
  register integer hashval = hashf(key, 0);

  for (;;) {
    if (p == NULL) { 
      p = new HashNode(hashval, key, &o);         // build a new node
      if (root == NULL)
	root = p;
      else {
	if (q->hashval > hashval)
	  q->right = p;
	else
	  q->left = p;
      }
      ++count;
      return;
    } 
    else if (p->hashval == hashval) { 
      if (strcmp(key, (char*) p->key) == 0) {

        if( DelData() ) delete p->data;
	p->data = &o;                    // Jam it! 

        if( DelKey() ) delete key;	// keys should match, anyway

        return;                         
      } 
      else 
        hashval = hashf(key, hashval); // hash collision, rehash 
    }
    q = p;
    if (p->hashval > hashval) 
      p=p->right;
    else
      p=p->left;
  }
}

////////////////////////////////////////////////////////////////////////////

Object*
HashTree::Get(char *key)
{
  register integer hashval = hashf(key, 0);
  register HashNode *p = root;

  for (;;) {
    if (p == NULL) 
      return 0;
    
    else if (p->hashval == hashval) {   // hashval match
       if (strcmp(key, (char*) p->key) == 0)   // match!
	 return p->data;
       else 
	 hashval = hashf(key, hashval); // hash collision, rehash 
     }
    p = (p->hashval > hashval) ? p->right : p->left;
  }
}

////////////////////////////////////////////////////////////////////////////

void
HashTree::Print( ostream& out)
{
  if (root != NULL) {
    out << "Hash Tree : " << count << " item(s).\n";
    PrintNode(root, 0, out);
  } else
    out << "Hash Tree : empty\n";
}

////////////////////////////////////////////////////////////////////////////

void
HashTree::PrintNode(HashNode *p, int depth, ostream& out)
{
  out << "Depth=" << depth << " Key=\"" << (char*) p->key << "\" Data=";
  if (p->data)
    p->data->Print(out);

  if (p->right != NULL)
    PrintNode(p->right, depth+1, out);
  if (p->left != NULL)
    PrintNode(p->left,  depth+1, out);
}

////////////////////////////////////////////////////////////////////////////

integer
HashTree::hashf (char *s, integer h) 
{  
  int shift_out;
  for (    ; *s != '\0'; ++s) {
    h += (int)*s << 11;
    shift_out = h&0x01;
    h >>= 1;
    if (shift_out)
      h ^= 0x63a;
  }
  h &= 0x0fff;
  return (integer) h%1013;
}

////////////////////////////////////////////////////////////////////////////

Hash_iterator::Hash_iterator(HashTree& h) { // constructor
  if (h.root) {
    n = h.root;
    done = FALSE;
  } else {
    done = TRUE;
  }
}

Hash_iterator::~Hash_iterator() {     // destructor
}

void
Hash_iterator::operator ++ () {      // iterator to walk through hash tree 

  if (done == TRUE)
    return;

    if (n->left != NULL) {  // try to go left
      stack.Add(*n);     // save current node
      n = n->left;
      return;
    } 

    while (n->right == NULL) {
      
      if (stack.Count() == 0) {
        done = TRUE;
        return;
      }
      n = (HashNode *) stack.Pop();
    }
   n = n->right;
   return;
}

///////////////////////////////////////////////////////////////////////////

Object*
Hash_updater::Get(char *key)
{
  register integer hashval = ht->hashf(key, 0);
  n = ht->root;

  for (;;) {
    if (n == NULL) 
      return 0;
    
    else if (n->hashval == hashval) {   // hashval match
       if (strcmp(key, (char*) n->key) == 0)   // match!
	 return n->data;
       else 
	 hashval = ht->hashf(key, hashval); // hash collision, rehash 
     }
    n = (n->hashval > hashval) ? n->right : n->left;
  }
}

////////////////////////////////////////////////////////////////////////////
