#ifndef HASH_H
#define HASH_H
////////////////////////////////////////////////////////////////////////////
//
//  hash.h : hash Tree class, string keys
//  Andrew Seawright 
//  $Header: /usr6/research/code/homebrew/RCS/hash.h,v 1.14 1996/12/14 02:20:43 crews Exp $
//
////////////////////////////////////////////////////////////////////////////
#include "list.h"

#define HashSort	HashTree( TRUE,TRUE )

class HashNode : public Object {

  HashNode    *right;
  HashNode    *left;
  integer      hashval;
  void        *key;
  Object      *data;
  
  HashNode(int, void *, Object *);
  HashNode(HashNode *, boolean copy = TRUE); // recursive copy
 ~HashNode() {}
  friend class HashTree;
  friend class Hash_iterator;
  friend class Hash_updater;
};

////////////////////////////////////////////////////////////////////////////

class Hash_iterator : public Object {
  
  List      stack;
  boolean   done;
  HashNode* n;
public:
  Hash_iterator(HashTree& h);
 ~Hash_iterator();
  boolean     Update( Object* o) { if (n==NULL) n->data = o; return( n==NULL); }
  void        operator ++(); 
  boolean     end() {return done;}
  void*       item(Object **d) {*d = n->data; return n->key; }
};
////////////////////////////////////////////////////////////////////////////

class Hash_updater : public Object {
  
  HashTree *ht;
  HashNode *n;
public:
  Hash_updater(HashTree& h)  { ht=&h; n=NULL;}
  Hash_updater(HashTree* h)  { ht= h; n=NULL;}
 ~Hash_updater() {;}

  Object*  Get( char* key);
  boolean  Update( Object* o)	{ if (n==NULL) n->data = o; return( n==NULL); }
};
////////////////////////////////////////////////////////////////////////////

class HashTree : public Object {

  friend class Hash_iterator;
  friend class Hash_updater;
       
  virtual void    HashDelete(HashNode *);
  integer hashf(char*, integer);
  virtual void PrintNode(HashNode *, int, ostream& out = cout);

protected:

  HashNode *root;
  integer  count;
  short    del_data;
  short	   del_key;
       
  boolean DelData() {return (boolean) del_data; }
  boolean DelKey()  {return (boolean) del_key; }
public:

  HashTree(boolean del_data = FALSE, boolean del_key = TRUE);
  HashTree(HashTree *, boolean copy=FALSE); // copy me harder
 ~HashTree()         { if (root != 0) HashDelete(root);}
  void Destroy()	{ if(root !=0 ) HashDelete(root); root=NULL; count=0; }
  integer Count ()   { return count;}

  void SetDelete(boolean d, boolean k) {del_data = d; del_key = k; }
  boolean Put (Object &o, char *key);         // returns TRUE if successful
  boolean Put (Object *o, char *key)          { return(Put(*o,key)); }
  void    ForcePut (Object &o, char *key);    // jam it!
  void    ForcePut (Object *o, char *key)     { ForcePut(*o,key); }

  // new data for first matching key; FALSE otherwise
  void Update (Object* o, char* key)		{ boolean dk = del_key;
						  del_key = FALSE;
						  ForcePut( *o, key );
						  del_key = dk;
						}
  void Update (Object& o, char* key)		{ Update(&o,key); }

  Object* Get (char* key); 
  boolean Get (Object **o, char* key) { return (0 != (*o=Get( key)));}        // returns TRUE if in tree

  void    Print ( ostream& out = cout); // print the contents
  integer isa() {return HASH_TREE;}
};

inline ostream& operator << (ostream &out, HashTree& f)
{ f.Print(out); return out; }

inline ostream& operator << (ostream &out, HashTree* f)
{ out << "(ptr->)"; f->Print(out); return out; }


////////////////////////////////////////////////////////////////////////////
// macros

#define Hash_ForEach_Object(ht, k, o)  \
        { \
           Object *o; \
           char *k; \
           for ( Hash_iterator hash__itr(ht); !hash__itr.end(); ++hash__itr) { \
             k = (char*) hash__itr.item(&o);

#define Hash_ForEach_Type(ht, t, k, o)  \
        { \
           t *o; \
           char *k; \
           for ( Hash_iterator hash__itr(ht); !hash__itr.end(); ++hash__itr) { \
             k = (char*) hash__itr.item((Object **)&o);

#define Hash_End_ForEach }}

////////////////////////////////////////////////////////////////////////////

#endif // HASH_H

