////////////////////////////////////////////////////////////////////////////
//
//  hash_t.C : templated hash tree object
//  Andrew Seawright, Andy Crews
//  $Header: /usr2/users/guests/astornet/bdd/RCS/hash_t.C,v 1.1 1995/05/05 20:31:06 crews Exp $
//  
////////////////////////////////////////////////////////////////////////////


template <class DataClass>
HashNode_t<DataClass>::HashNode_t(int hv, void *k, DataClass *o) {   // HashNode constructor

  right   =  left  =  0;
  hashval =  hv;
  key     =  k;
  data    =  o;
}

template <class DataClass>
HashNodeTemp_t<DataClass>::HashNodeTemp_t(int hv, void *k, DataClass *o) {   // HashNode constructor

  right   =  left  =  0;
  hashval =  hv;
  key     =  k;
  data    =  *o;
}

template <class DataClass> void
HashNodeTemp_t<DataClass>::Print( ostream& out ) {
  out << "Key = " << Key() << ", Data = " << Data();
  out << "\nhashval=" << hashval;
  }

template <class DataClass> void
HashNode_t<DataClass>::Print( ostream& out ) {
  out << "Key = " << Key() << ", Data = " << Data();
  out << "\nhashval=" << hashval;
  }

////////////////////////////////////////////////////////////////////////////

template <class DataClass, class HashNodeType>
HashTree_t<DataClass, HashNodeType>::HashTree_t() {        // HashTree constructor
  
  count   =  0;
  root    =  0;
}

////////////////////////////////////////////////////////////////////////////

template <class DataClass, class HashNodeType> void
HashTree_t<DataClass, HashNodeType>::HashDelete(HashNodeType *p) {    // recursive destruction of tree 

  if (p->Right() != NULL) {
     HashDelete((HashNodeType*)p->Right());
     p->Right (NULL);
  }
  if (p->Left() != NULL) {
     HashDelete((HashNodeType*)p->Left());
     p->Left(NULL);
  }
  // just added this
  delete p->Key();
  delete p;
}

////////////////////////////////////////////////////////////////////////////

template <class DataClass, class HashNodeType> boolean
HashTree_t<DataClass, HashNodeType>::Put(DataClass& o, char *key)
{
  register HashNodeType *q = NULL, *p = root;
  register integer hashval = hashf(key, 0);

  for (;;) {
    if (p == NULL) { 
      p = new HashNodeType(hashval, key, &o);         // build a new node
      if (root == NULL)
	root = p;
      else {
	if (q->hashval > hashval)
	  q->Right(p);
	else
	  q->Left( p );
      }
      ++count;
      return TRUE;
    } 
    else if (p->hashval == hashval) { 
      if ( strcmp(key, (char*) p->Key()) == 0) 
	return FALSE;                    // key already in tree
      else 
	hashval = hashf(key, hashval); // hash collision, rehash 
    }
    q = p;
    
    if (p->hashval > hashval)
      p=(HashNodeType*)p->Right();
    else
      p=(HashNodeType*)p->Left();
  }
}

////////////////////////////////////////////////////////////////////////////

template <class DataClass, class HashNodeType> void
HashTree_t<DataClass, HashNodeType>::ForcePut(DataClass& o, char *key)
{
  register HashNodeType *q = NULL, *p = root;
  register integer hashval = hashf(key, 0);

  for (;;) {
    if (p == NULL) { 
      p = new HashNodeType(hashval, key, &o);         // build a new node
      if (root == NULL)
	root = p;
      else {
	if (q->hashval > hashval)
	  q->Right(p);
	else
	  q->Left(p);
      }
      ++count;
      return;
    } 
    else if (p->hashval == hashval) { 
      if (strcmp(key, (char*) p->Key()) == 0) {
	p->Data( &o );                  // Jam it! 
	return;                         
      } 
      else 
        hashval = hashf(key, hashval); // hash collision, rehash 
    }
    q = p;
    if (p->hashval > hashval) 
      p=p->Right();
    else
      p=p->Left();
  }
}

////////////////////////////////////////////////////////////////////////////

template <class DataClass, class HashNodeType> DataClass*
HashTree_t<DataClass, HashNodeType>::Get(char *key)
{
  register integer hashval = hashf(key, 0);
  register HashNodeType *p = root;

  for (;;) {
    if (p == NULL) 
      return 0;
    
    else if (p->hashval == hashval) {   // hashval match
       if (strcmp(key, (char*) p->Key()) == 0)   // match!
	 return p->Data();
       else 
	 hashval = hashf(key, hashval); // hash collision, rehash 
     }
    p = (p->hashval > hashval) ? p->Right() : p->Left();
  }
}

////////////////////////////////////////////////////////////////////////////

template <class DataClass, class HashNodeType> void
HashTree_t<DataClass, HashNodeType>::Print( ostream& out)
{
  if (root != NULL) {
    out << "Hash Tree : " << count << " item(s).\n";
    PrintNode(root, 0, out);
  } else
    out << "Hash Tree : empty\n";
}

////////////////////////////////////////////////////////////////////////////

template <class DataClass, class HashNodeType> void
HashTree_t<DataClass, HashNodeType>::PrintNode(HashNodeType *p, int depth, ostream& out)
{
  out << "Depth=" << depth << " Key=" << (char*) p->Key() << " ->";
  if (p->Data())
    p->Data()->Print(out);

  if (p->Right() != NULL)
    PrintNode(p->Right(), depth+1, out);
  if (p->Left() != NULL)
    PrintNode(p->Left(),  depth+1, out);
}

////////////////////////////////////////////////////////////////////////////

template <class DataClass, class HashNodeType> integer
HashTree_t<DataClass, HashNodeType>::hashf (char *s, integer h) 
{  
  int shift_out;
  for (    ; *s != '\0'; ++s) {
    h += (int)*s << 11;
    shift_out = h&0x01;
    h >>= 1;
    if (shift_out)
      h ^= 0x63a;
  }
  h &= 0x0fff;
  return (integer) h%1013;
}

////////////////////////////////////////////////////////////////////////////

template <class DataClass, class HashNodeType>
Hash_iterator_t<DataClass, HashNodeType>::Hash_iterator_t(HashTree_t<DataClass, HashNodeType>& h) { // constructor
  if (h.root) {
    n = h.root;
    done = FALSE;
  } else {
    done = TRUE;
  }
}
template <class DataClass, class HashNodeType>
Hash_iterator_t<DataClass, HashNodeType>::~Hash_iterator_t() {     // destructor
}

template <class DataClass, class HashNodeType> void
Hash_iterator_t<DataClass, HashNodeType>::operator ++ () {      // iterator to walk through hash tree 

  if (done == TRUE)
    return;

    if (n->Left() != NULL) {  // try to go left
      stack.Add(n);     // save current node
      n = n->Left();
      return;
    } 

    while (n->Right() == NULL) {
      
      if (stack.Count() == 0) {
        done = TRUE;
        return;
      }
      n = stack.Pop();
    }
   n = n->Right();
   return;
}

///////////////////////////////////////////////////////////////////////////
