///////////////////////////////////////////////////////////////////
// $RCSfile: hasharry.C,v $
// $Author: crews $
// $Date: 1997/03/08 00:21:36 $
//
// $Revision: 1.8 $
// $Locker:  $
//
///////////////////////////////////////////////////////////////////

#include <string.h>
#include <stdlib.h>
#include <iostream.h>
#include "hasharry.h"

////////////////////////////////////////////////////////////////////////////

BDDHasherArray::BDDHasherArray()
  {
  // Set up the initial number of tables in this hash
  count = 0;
  maxTables = MAX_BDD_VAR;
  numTables = 0;
  hashTables = new TablePtr[maxTables];
  lookup_index = new integer[maxTables+1];
  lookup_level = new integer[maxTables+1];
  lookup_index = &lookup_index[1];
  lookup_level = &lookup_level[1];
  int i;
  for (i=0; i<maxTables; i++)
    hashTables[i] = NULL;
  for (i=-1; i<maxTables; i++)
    lookup_index[i] = lookup_level[i] = i;
  }

BDDHasherArray::BDDHasherArray(integer *lui, integer *lul )
  {
  count = 0;
  maxTables = MAX_BDD_VAR;
  numTables = 0;
  hashTables = new TablePtr[maxTables];
  lookup_index = new integer[maxTables+1];
  lookup_level = new integer[maxTables+1];
  lookup_index = &lookup_index[1];
  lookup_level = &lookup_level[1];
  int i;
  for (i=0; i<maxTables; i++)
    hashTables[i] = NULL;
  for (i=-1; i<maxTables; i++)
    {
    lookup_index[i] = lui[i];
    lookup_level[i] = lul[i];
    }
  }

BDDHasherArray::BDDHasherArray(BDDHasherArray *tree, boolean copy)
  {
  // Set up the initial number of tables in this hash
  if (copy == TRUE)
    {
    count = tree->count;
    hashTables = new TablePtr[maxTables];
    lookup_index = new integer[maxTables+1];
    lookup_level = new integer[maxTables+1];
    lookup_index = &lookup_index[1];
    lookup_level = &lookup_level[1];
    int i;
    for (i=0; i<tree->maxTables; i++)
      hashTables[i] = tree->hashTables[i];
    for (i=-1; i<tree->maxTables; i++)
      {
      lookup_index[i] = tree->lookup_index[i];
      lookup_level[i] = tree->lookup_level[i];
      }
    }
  else	// used in GC to create a new array
    {
    count = 0;
    numTables = 0;
    maxTables = tree->maxTables;
    hashTables = new TablePtr[maxTables];
    for (int i=0; i<maxTables; i++)
      hashTables[i] = NULL;
    lookup_index = tree->lookup_index;
    lookup_level = tree->lookup_level;
    tree->lookup_index = NULL;
    tree->lookup_level = NULL;
    }
  }

void
BDDHasherArray::Clear() {
  for( int i=0 ; i<numTables ; i++ )
    hashTables[i]->Clear();
  }

BDDHasherArray::~BDDHasherArray() {
  for( int i=0 ; i<numTables ; i++ )
    delete hashTables[i];
  delete hashTables;
  if ( lookup_index) {
    lookup_index = &lookup_index[-1];
    delete lookup_index;
    }
  if ( lookup_level) {
    lookup_level = &lookup_level[-1];
    delete lookup_level;
    }
 }

////////////////////////////////////////////////////////////////////////////

boolean BDDHasherArray::Put(BDDnode* key)
  {
  integer index = key->GetIndex();
  return (hashTables[index]->Put( key ));
  }


BDDnode *BDDHasherArray::ForceGet(BDDnode *key)
{
  integer index = key->GetIndex();
  return hashTables[index]->ForceGet(key);
}

BDDnode *BDDHasherArray::Get(BDDnode *key)
{
  integer index = key->GetIndex();
  return hashTables[index]->Get(key);
}

boolean
BDDHasherArray::Remove(BDDnode *key)
{
  integer index = key->GetIndex();
  return (hashTables[index]->Remove( key ) != 0);
}

// This inserts a level with index "numTables" (lowest unused index)
// There is no affect on the bdds, since a new level can be inserted
// anywhere. Once this is perfected, it should replace the less
// general BDDHasherArray::AddHasher( ) routine.

void
BDDHasherArray::InsertLevel( int l ) {
  int i;
  for( i=numTables ; i>l ; i-- ) {
    lookup_index[i] = lookup_index[i-1];
    lookup_level[ lookup_index[i] ] = i;
    }
  
  lookup_index[l] = numTables;
  lookup_level[ numTables ] = l;
  hashTables[numTables] = new SubHasher;
  numTables++;
  }
    
////////////////////////////////////////////////////////////////////////////

void BDDHasherArray::ShortPrint( ostream& out) 
{ 
  out << "Elements in hash = " << count << endl;
}

void BDDHasherArray::Print( ostream& out)
  {
  for( int i=0 ; i<numTables ; i++ )
    {
    out << "\nNodes for index " << i << endl;
    hashTables[i]->Print(out);
    out << endl;
    }
  }

void
BDDHasherArray::AddHasher( ) {
  hashTables[numTables++] = new SubHasher;
  }

