///////////////////////////////////////////////////////////////////
// $RCSfile: hashbase.h,v $
// $Author: crews $
// $Date: 1998/03/13 21:04:11 $
//
// $Revision: 1.5 $
// $Locker:  $
//
///////////////////////////////////////////////////////////////////

#ifndef HASHBASE_H
#define HASHBASE_H

#include "compable.h"

// The purpose of this class is to serve as a basic virtual base
// class for all hashing types of classes

class HashBase : public Object {
protected:

  integer count;		// Number of nodes in this table
  integer puts;
  integer hits;
  integer gets;
  integer collisions;
  struct {
    boolean del_data : 1;
    boolean del_key :  1;
    } flags;

  // Pure virtual functions (must appear in inherited classes)
  virtual boolean Add      (Object&, Comparable*) = 0;
  virtual void    Overwrite (Object&, Comparable*) = 0;
  virtual Object* Find (Comparable*) = 0;
  
  boolean DelData() { return flags.del_data; }
  boolean DelKey() { return flags.del_key; }

public:
 
  virtual ~HashBase()	{;}
  int Count()		{ return count;}	// running total
  int Puts()		{ return puts;}		// successful put or forceput
  int Hits()		{ return hits;}		// successful gets
  int Gets()		{ return gets;}		// gets
  int Collisions()	{ return collisions;}	// (puts and forceputs)
  void SetDelete( boolean d, boolean k ) { flags.del_data=(d!=0);
					   flags.del_key=(k!=0); }

  virtual void Clear() {;}
  virtual void Rehash() {;}

  boolean Put (Object &o, Comparable *key) { return Add( o, key ); }
  boolean Put (Object *o, Comparable *key) { return Add (*o, key); }
  boolean Put (Comparable *o)              { return Add (*o,   o); }
  boolean Put (Comparable &o)              { return Add ( o,  &o); }

  void    ForcePut (Object&o, Comparable*key)	{ Overwrite( o, key );}
  void    ForcePut (Object *o, Comparable *key)	{ Overwrite (*o, key);}
  void    ForcePut (Comparable *o)		{ Overwrite (*o,   o);}
  void    ForcePut (Comparable &o)		{ Overwrite ( o,  &o);}

  Object* Get (Comparable*key)		{ return Find( key ); }
  Object* Get (Comparable &key)		{ return Find (&key); }
  boolean Get(Object **o, Comparable *key) { return (0 != (*o = Find(key)));}
  boolean Get(Object **o, Comparable &key) { return (0 != (*o = Find(&key)));}

  // new data for first matching key; NULL otherwise
  void Update (Object* o, Comparable* key)      { boolean dk = flags.del_key;
                                                  flags.del_key = FALSE;
                                                  Overwrite(*o, key);
                                                  flags.del_key = dk; }
  
  void Update (Object& o, Comparable* key)      { Update( &o, key ); }
  void    PrintStats (ostream& out=cout) {
    out << "count     =  " << Count() << endl;
    out << "hits      =  " << Hits() << endl;
    out << "total gets=  " << Gets() << endl;
    out << "total puts=  " << Puts() << endl;
    out << "collisions=  " << Collisions() << endl;
    out << "hit ratio =  " << (double)Hits() / (double)Gets() << endl;
    }

  virtual boolean Remove (Comparable *)
    { cerr << "Write a Remove() function!" << endl; return FALSE; }
  boolean Remove(Comparable &key) { return Remove( &key);}

  virtual integer isa() = 0;
  virtual void Print (ostream& out=cout)
    { out << "Write a Print() function!" << endl; }
  virtual void Dump(ostream& out=cout)
    { out << "Write a Dump() function!" << endl; }
  virtual void ShortPrint( ostream& out=cout) { PrintStats(out); }
};

// HashExact is defined for hash-based containers that are never overwrite
class HashExact : public HashBase { /* */ };

#endif // HASHBASE_H

