#include <iostream.h>
#include "hashcach.h"

// hashcache.C         Andrew Seawright

////////////////////////////////////////////////////////////////////////////

HashCache::HashCache (integer s, boolean dd, boolean dk) {

  ASSERT(s>0, "HashCache : size must be > 0");
  size = s;
  keys  = new Comparable*[size];
  cache = new Object*[size];
  gets = hits = count = collisions = puts = 0;
  SetDelete( dd, dk );

  integer i;
  for (i=0; i<size; ++i) {
    keys[i] = NULL;
  }
}


////////////////////////////////////////////////////////////////////////////

HashCache::~HashCache () {
  if( DelData() || DelKey() )
    Clear();
  delete cache;
  delete keys;
}

////////////////////////////////////////////////////////////////////////////

void
HashCache::Clear() {

  hits=gets=puts=collisions=0;
  if( Count()==0)
    return;

  integer i;

  for (i=0; i<size; ++i) {
    if (keys[i])
      {
      if( DelData() ) delete cache[i];
      if( DelKey() ) delete keys[i];
      keys[i] = NULL;
      }
  }
  count =0;
}

////////////////////////////////////////////////////////////////////////////
 
void
HashCache::ConditionalClear() {

  if( Count()==0)
    return;

  integer i;

  for (i=0; i<size; ++i) {
    if ( keys[i] && (keys[i]->Flush() || cache[i]->Flush()) ) {
      if( DelData() ) delete cache[i];
      if( DelKey() ) delete keys[i];
      count--;
      keys[i] = 0;
    }
  }
}
////////////////////////////////////////////////////////////////////////////

boolean
HashCache::Add(Object &o, Comparable *key) {

  puts++;
  integer bin = key->Hash() % size;
  if (keys[bin]) {
    if( DelKey() ) delete keys[bin];
    if( DelData() ) delete cache[bin];
    collisions++;
    }
  else
    count++;
  cache[bin] = &o;
  keys[bin] = key;
  return TRUE;
}

////////////////////////////////////////////////////////////////////////////

Comparable *
HashCache::Replace(Object *o, Comparable *key) {

  puts++;
  Comparable *old_key;
  integer bin = key->Hash() % size;
  if (keys[bin]) {
    if( DelData() ) delete cache[bin];
    old_key = keys[bin], collisions ++;
    }
  else {
    old_key = 0;
    count++;
    }
  cache[bin] = o;
  keys[bin] = key;
  return old_key;
}

////////////////////////////////////////////////////////////////////////////

Object*
HashCache::Find( Comparable *key) {

  ++gets;
  integer bin = key->Hash() % size;
  if ((keys[bin] != 0) && key->Compare(keys[bin]) == 0) {
    ++hits;
    return cache[bin];
  } else
    return NULL;
}

////////////////////////////////////////////////////////////////////////////

void
HashCache::PrintStats(ostream& out) {

  out << "\nHashCache\n";
  out << "size      =  " << size << "\n";
  HashBase::PrintStats(out);
}

////////////////////////////////////////////////////////////////////////////

void
HashCache::Dump( ostream& out) {
  for( int bin=0 ; bin<size ; bin++ )
    if( keys[bin] )
      {
      out << bin << "\tkey: " << keys[bin];
      out << "\tdata: " << cache[bin];
      }
  }

