#ifndef HASH_CACHE_H
#define HASH_CACHE_H
 
 //
 // hashcache.h : simple cache for Comparable Objects 
 //
 // Andrew Seawright
 //

#include "compable.h"
#include "hashbase.h"

#define HASH_CACHE_DEFAULT_SIZE 211

class HashCache : public HashBase {

protected:

  integer      size;
  Comparable **keys;
  Object     **cache;

  virtual boolean Add ( Object&, Comparable*);
  virtual void Overwrite( Object&o, Comparable*c) {
    Comparable *old = Replace(&o,c);
    if( old && DelKey() ) delete old;
    }
  virtual Object* Find( Comparable* );
       
public:

  HashCache(integer = HASH_CACHE_DEFAULT_SIZE, boolean del_data=TRUE, boolean del_key=FALSE );
 ~HashCache();

  Comparable* Replace(Object* o, Comparable* key); // returns old key ptr
  Comparable* Replace(Comparable* key) { return Replace( key, key);}

  virtual void        Clear();
  void	      ConditionalClear();	// check virtual function Flush before clearing

  void PrintStats ( ostream& out = cout);       // print the stats

  // Object virtual functions

  virtual void Print( ostream& out = cout) { PrintStats( out ); }
  virtual void Dump( ostream& out = cout);
  virtual integer isa() { return HASH_CACHE; }
};

inline ostream& operator << (ostream &out, HashCache& f)
{ f.Print(out); return out; }

inline ostream& operator << (ostream &out, HashCache* f)
{ out << "(ptr->)"; f->Print(out); return out; }

////////////////////////////////////////////////////////////////////////////

#endif // HASH_CACHE_H

