#include "list.h"
#include "hashtbl.h"

int HashObject::hashObjectCount = 0;

HashTable::HashTable(boolean d_data, boolean d_key)
{
  bucketCount = 150;
  buckets = new List[bucketCount];
  gets=puts=collisions=hits=rehashCount=maxDepth=count = 0;
  SetDelete( d_data, d_key );
}

HashTable::HashTable(int size, boolean d_data, boolean d_key)
{
  bucketCount = size;
  buckets = new List[bucketCount];
  gets=puts=collisions=hits=rehashCount=maxDepth=count = 0;
  SetDelete( d_data, d_key );
}

void
HashTable::Clear()
{
  for (int i = 0; i < bucketCount; i++) {
     List_ForEach_Type(buckets[i],HashObject,ob) {
     delete ob;
    } List_End_ForEach
  }
  ::delete buckets;
  buckets = new List[bucketCount];

  gets=puts=collisions=hits=rehashCount=maxDepth=count = 0;
}   
   
// HashTable::HashTable(HashTable *t)
// {
//    bucketCount = t->bucketCount;
//    buckets = new List[bucketCount];
// 
//    for (int i = 0; i < bucketCount; i++) {
//       List_ForEach_Type(*t->Bucket(i),void,((Object *)ob)) {
// 	 buckets[i].Add((Object *)ob);
//       } List_End_ForEach
//    }
//    objectCount = t->Count();
//}

#define REHASH_GROWTH_FACTOR 1.5
#define REHASH_RATIO 1.5

void
HashTable::Rehash()
{
   static int rehashInProgress = 0;
   double ratio = (double)Count()/bucketCount;
   
   if (ratio > REHASH_RATIO) {
//      cerr << "rehash " << (void*)this << "\n";
//      cerr << objectCount << " objects, " << bucketCount << " buckets\n";
//      cerr << ((double)objectCount / bucketCount) << " objects per bucket\n";
//      cerr << "depth = " << maxDepth << "\n";
//      cerr.flush();
      if (rehashInProgress) {
	 // new table isn't big enough for depth, so make it bigger and restart
	 // other option would be to increase rehash_depth automatically
	 // right now if more than rehash_depth items have the same
	 // hash value then this rehashing system will just loop.
	 // this will have to be fixed ...
	 rehashInProgress = 0;
	 Rehash();
      }
      rehashInProgress = 1;
      int oldBucketCount = bucketCount;
      bucketCount = (int)(bucketCount * REHASH_GROWTH_FACTOR);
      List *oldBuckets = buckets;
      buckets = new List[bucketCount];
      count = 0;
      maxDepth = 0;
      
      for (int i = 0; i < oldBucketCount; i++) {
	 List_ForEach_Type(oldBuckets[i],HashObject,ob) {
	    this->Add(ob);
	 } List_End_ForEach
	 oldBuckets[i].Clear();
      }

      rehashInProgress = 0;

      ::delete oldBuckets;
      rehashCount++;

   }
}

HashTable::~HashTable()
{
   int oldCount = HashObject::hashObjectCount;
   // delete the hash nodes
   for (int i = 0; i < bucketCount; i++) {
      List_ForEach_Type(buckets[i],HashObject,ob) {
	if (DelData()) delete ob->Data();
	if (DelKey() ) delete ob->Key();
	delete ob;
      } List_End_ForEach
      buckets[i].Clear();
   }
   // objectcenter seems to call Object delete here ...
   ASSERT(HashObject::hashObjectCount == oldCount-Count(),"~HashTable: not all objects deleted");
   ::delete buckets;
}

boolean
HashTable::Add(Object &o,Comparable *key)
{
   unsigned bucket = key->Hash() % bucketCount;
   HashObject *found = NULL;

   buckets[bucket].Add(new HashObject(&o,key));
   count++;
   puts++;
   if (buckets[bucket].Count() > maxDepth) {
      maxDepth = buckets[bucket].Count();
   }
   Rehash();
   return TRUE;
}

void
HashTable::Add(HashObject *o)
{
   unsigned bucket = o->Key()->Hash() % bucketCount;
   HashObject *found = NULL;

   buckets[bucket].Add(o);
   count++;
   puts++;
   if (buckets[bucket].Count() > maxDepth) {
      maxDepth = buckets[bucket].Count();
   }
}

boolean
HashTable::Remove(Comparable *key)
{
   unsigned bucket = (key->Hash()) % bucketCount;
   HashObject *found = NULL;

   List_ForEach_Type(buckets[bucket],HashObject,ob) {
      if (*ob->Key() == *key) {
	 found = ob;
	 break;
      }
   } List_End_ForEach
      
   if (found != NULL) {
      buckets[bucket].Remove(found);
      if( DelData() ) delete found->Data();
      if( DelKey() ) delete found->Key();
      delete found;
      count--;
      return TRUE;
   } else {
      return FALSE;
   }
}

Object *
HashTable::Find(Comparable *key)
{
   unsigned bucket = ((unsigned)key->Hash()) % bucketCount;
   HashObject *found = NULL;

   List_ForEach_Type(buckets[bucket],HashObject,ob) {
      if (*ob->Key() == *key) {
	 found = ob;
	 break;
      }
   } List_End_ForEach

   gets++;
   if (found != NULL) {
      hits++;
      return found->Data();
   } else {
      return NULL;
   }
}

void
HashTable::PrintStats(ostream &out)
{
   out << "Hash Table Stats for #<HashTable " << (void *)this << "> :\n";
   out << "\tBuckets = " << bucketCount << "\n";
   out << "\tRehashes = " << rehashCount << "\n";
   out << "\tObjects/Bucket = " << (double)Count()/bucketCount << "\n";
   out << "\tMax Depth = " << maxDepth << "\n";
   out << "\tAverage Depth = ";
   int sum = 0,cnt = 0;
   for (int i = 0; i < bucketCount; i++) {
      if (buckets[i].Count()) {
	 sum += buckets[i].Count();
	 cnt++;
      }
   }
   if (cnt) 
      out << (double)sum/cnt << "\n";
   else
      out << 0 << "\n";
  HashBase::PrintStats(out);
}

   
void
HashTable::Dump(ostream &out)
{
   out << "Hash Table (" << Count() << " objects):\n";
   HashTable_ForEach_Type(this,Object,key,ob) {
	 out << ob;
   } HashTable_End_ForEach
}

HashObject::HashObject(Object *d,Comparable *k)
{
   data = d; key = k;
   hashObjectCount++;
}

HashObject::~HashObject()
{
   hashObjectCount--;
}
