//
//
// Heap class implements a priority queue.
// 
//

#include <iostream.h>
#include "heap.h"


//
// This is a Heap constructor. It allocates an
// array of 'n' pointers to Comparable objects.
// It initialize maxsize parameter to 'n', and
// tail pointer to 0. In order to implement 
// efficient (fast) functions, the actual array
// has (n+1) elements (pointer heap[0] is unused).
// 
Heap::Heap(integer n) {
  integer i;
  tail = 0;
  maxsize = n;
  heap = new (Comparable*[n+1]);
  for(i=0; i<=n; i++) {
    heap[i] = 0;
  }
}

//
// This function returns a current number (tail) of
// elements in the priority queue.
//
//integer Heap::HeapSize() {
//  return(tail); 
//}

//
// 'DeleteMin' function removes a minimum element
// from the priority queue (actually a root of
// a heap representation of tree). 'Pop' function
// is merely an alias of 'DeleteMin'.
//
Comparable* Heap::DeleteMin() {
  if (tail==0) 
    FATAL("Cannot DeleteMin from an empty heap");
  Comparable *minimum = this->heap[1]; 
  this->heap[1] = this->heap[tail];
  tail-=1;
  integer i = 1;
  integer j;

  while(i<=(tail/2)) {
    if((*(heap[2*i]) < *(heap[2*i+1])) || (2*i==tail)) j=2*i; 
      else j=2*i+1;
    if(*(heap[i]) > *(heap[j])) {
      Comparable *temp = heap[i];
      heap[i] = heap[j];
      heap[j] = temp;
      i = j;
    }
      else return(minimum);
  }
  return(minimum);
}

//
// 'Insert' function inserts a new comparable object
// int the priority queue. 'Add' and 'Push' functions
// are merely aliases of 'Insert'.
//
void Heap::Insert(Comparable& x) {
  integer i;
  if (tail>=maxsize)
    FATAL("Cannot insert into a full queue")
  tail+=1;
  heap[tail] = &x;
  i = tail;
  while((i>1) && (*(heap[i]) < *(heap[i/2]))) {
    Comparable* temp = heap[i];
    heap[i] = heap[i/2];
    heap[i/2] = temp;
  i = i/2;
  }
}

void Heap::Print(ostream& out) {
  integer i;
  out << "maxsize: " << maxsize << "\n" ;
  out << "tail: " << tail << "\n" ;
  for(i=1; i<=tail; i++) {
    out << "heap[" << i << "]:\n" ;
    heap[i]->Print();
  }
} 

Comparable* Heap::Pop()            { return (tail)? DeleteMin():NULL;}

void Heap::Add(Comparable& x)      { Insert(x);}

void Heap::Add(Comparable *x)      { Insert(*x);}

void Heap::Push(Comparable& x)     { Insert(x);}

void Heap::Push(Comparable *x)     { Insert(*x);}

