#ifndef HGRAPH_H
#define HGRAPH_H
////////////////////////////////////////////////////////////////////////////
//
//  hgraph.h : a hyper-graph package
//  Chuck Monahan
//  12/7/94
//
////////////////////////////////////////////////////////////////////////////

#include "list.h"

////////////////////////////////////////////////////////////////////////////
// macro's from hell

#define Hgraph_ForEach_Node(hg, n) \
{ \
  HgraphNode *n; \
  for ( List_iterator list__itr((hg)->nodes); !list__itr.end(); ++list__itr ) { \
            n = (HgraphNode *)list__itr.item();

#define Hgraph_ForEach_Node_Object_Data(hg, o) \
{ \
  Object *o; \
  for ( List_iterator list__itr((hg)->nodes); !list__itr.end(); ++list__itr ) { \
            o = ((HgraphNode *)list__itr.item())->Get();

#define Hgraph_ForEach_Node_Type_Data(hg, t, o) \
{ \
  t *o; \
  for ( List_iterator list__itr((hg)->nodes); !list__itr.end(); ++list__itr ) { \
            o = (t *) ((HgraphNode *)list__itr.item())->Get();

#define Hgraph_ForEach_Edge(hg, e) \
{ \
  HgraphEdge *e; \
  for ( List_iterator list__itr((hg)->edges); !list__itr.end(); ++list__itr ) { \
            e = (HgraphEdge *)list__itr.item();
 
#define Hgraph_ForEach_Edge_Object_Data(hg, o) \
{ \
  Object *o; \
  for ( List_iterator list__itr((hg)->edges); !list__itr.end(); ++list__itr ) { \
            o = ((HgraphEdge *)list__itr.item())->Get();

#define Hgraph_ForEach_Edge_Type_Data(hg, t, o) \
{ \
  t *o; \
  for ( List_iterator list__itr((hg)->edges); !list__itr.end(); ++list__itr ) { \
            o = (t *) ((HgraphEdge *)list__itr.item())->Get();

#define Hgraph_ForEach_Out_Node(edge, onode) \
{ \
   HgraphNode *onode; \
   for ( List_iterator list__itr((edge)->to_nodes); !list__itr.end(); ++list__itr ){\
            onode = (HgraphNode*)list__itr.item();

#define Hgraph_ForEach_Out_Node_Object_Data(edge, o) \
{ \
   Object *o; \
   for ( List_iterator list__itr((edge)->to_nodes); !list__itr.end(); ++list__itr ){\
            o = ((HgraphNode*)list__itr.item())->Get();

#define Hgraph_ForEach_Out_Node_Type_Data(edge, t, o) \
{ \
   t *o; \
   for ( List_iterator list__itr((edge)->to_nodes); !list__itr.end(); ++list__itr ){\
            o = (t*) ((HgraphNode*)list__itr.item())->Get();

#define Hgraph_ForEach_Out_Edge(node, oedge) \
 { \
   HgraphEdge *oedge; \
   for ( List_iterator list__itr((node)->out_edges); !list__itr.end(); ++list__itr ){\
            oedge = (HgraphEdge*)list__itr.item();

#define Hgraph_ForEach_Out_Edge_Object_Data(node, o) \
 { \
   Object *o; \
   for ( List_iterator list__itr((node)->out_edges); !list__itr.end(); ++list__itr ){\
            o = ((HgraphEdge*)list__itr.item())->Get();

#define Hgraph_ForEach_Out_Edge_Type_Data(node, t, o) \
 { \
   t *o; \
   for ( List_iterator list__itr((node)->out_edges); !list__itr.end(); ++list__itr ){\
            o = (t*) ((HgraphEdge*)list__itr.item())->Get();

#define Hgraph_ForEach_In_Node(edge, inode) \
 { \
   HgraphNode *inode; \
   for ( List_iterator list__itr((edge)->from_nodes); !list__itr.end(); ++list__itr ){\
            inode = (HgraphNode*)list__itr.item();

#define Hgraph_ForEach_In_Node_Object_Data(edge, o) \
 { \
   Object *o; \
   for ( List_iterator list__itr((edge)->from_nodes); !list__itr.end(); ++list__itr ){\
            o = ((HgraphNode*)list__itr.item())->Get();

#define Hgraph_ForEach_In_Node_Type_Data(edge, t, o) \
 { \
   t *o; \
   for ( List_iterator list__itr((edge)->from_nodes); !list__itr.end(); ++list__itr ){\
            o = (t*) ((HgraphNode*)list__itr.item())->Get();

#define Hgraph_ForEach_In_Edge(node, iedge) \
 { \
   HgraphEdge *iedge; \
   for ( List_iterator list__itr((node)->in_edges); !list__itr.end(); ++list__itr ){\
            iedge = (HgraphEdge*)list__itr.item();

#define Hgraph_ForEach_In_Edge_Object_Data(node, o) \
 { \
   Object *o; \
   for ( List_iterator list__itr((node)->in_edges); !list__itr.end(); ++list__itr ){\
            iedge = ((HgraphEdge*)list__itr.item())->Get();

#define Hgraph_ForEach_In_Edge_Type_Data(node, t, o) \
 { \
   t *o; \
   for ( List_iterator list__itr((node)->in_edges); !list__itr.end(); ++list__itr ){\
            o = (t*) ((HgraphEdge*)list__itr.item())->Get();


#define Hgraph_End_ForEach }}

////////////////////////////////////////////////////////////////////////////
class Hgraph;
class HgraphNode;
class HgraphEdge;

class HgraphNode : public Object {
  friend class Hgraph;
  friend class HgraphEdge;

protected:   
  integer node_num;   // node's number
  Object* data;       // node's data 
  boolean visit;


public:
  List out_edges;      // fan-out edge list
  List in_edges;       // fan-in edge list

  HgraphNode( Object* o, integer n=-1) { data = o; visit = FALSE; node_num = n;}
  HgraphNode( HgraphNode& n );
 ~HgraphNode() {;}

  integer     Fanout()         { return out_edges.Count();}
  integer     Fanin()          { return in_edges.Count();}
  integer     GetNum()         { return node_num;}
  void        Put(Object* o)   { data = o;}
  Object*     Get()            { return data;}
  void        Touch(boolean v) { visit = v;}
  boolean     Touched()        { return visit;}

  boolean     SourceNode();
  boolean     SinkNode();

  HgraphEdge* GetOutEdge  ( integer     i) { return (HgraphEdge*) out_edges.GetElement(i);}
  HgraphEdge* GetInEdge   ( integer     i) { return (HgraphEdge*)  in_edges.GetElement(i);}
  integer     WhichOutEdge( HgraphEdge* e);
  integer     WhichInEdge ( HgraphEdge* e);

  void        ListDependants( List* nodes);     // touch the nodes to list
  void        ListDependancies( List* nodes);   // touch the nodes to list

  virtual integer  isa() { return HGRAPH_NODE;}
  virtual void     Print( ostream& out = cout);
  virtual void     PrintII( ostream& out, boolean); 
};  

inline ostream& operator << (ostream &out, HgraphNode& f)
{ f.Print(out); return out; }

inline ostream& operator << (ostream &out, HgraphNode* f)
{ out << "(ptr->)"; f->Print(out); return out; }

////////////////////////////////////////////////////////////////////////////

class HgraphEdge : public Comparable {
  friend class Hgraph;
  friend class HgraphNode;

protected:   
  Comparable* data;   // edge's data 

public: 
  List to_nodes;      // the nodes this edge points to
  List from_nodes;    // the nodes this edge points from

  HgraphEdge( Comparable* o=NULL) { data = o;}
 ~HgraphEdge() {;}

  integer     Fanout()           { return to_nodes.Count();}
  integer     Fanin()            { return from_nodes.Count();}
  void        Put(Comparable* o) { data = o;}
  Comparable* Get()              { return data;}

  HgraphNode* GetToNode    ( integer     i) { return (HgraphNode*)   to_nodes.GetElement(i);}
  HgraphNode* GetFromNode  ( integer     i) { return (HgraphNode*) from_nodes.GetElement(i);}
  integer     WhichToNode  ( HgraphNode* n) { return   to_nodes.IdentityNum(n);}
  integer     WhichFromNode( HgraphNode* n) { return from_nodes.IdentityNum(n);}

  virtual integer  isa() { return HGRAPH_EDGE;}
  virtual integer  Compare( Comparable* c);
  virtual unsigned Hash( unsigned i=0); 
  virtual void     Print( ostream& out = cout);     
  virtual void     PrintII( ostream& out, boolean); 
};

inline ostream& operator << (ostream &out, HgraphEdge& f)
{ f.Print(out); return out; }

inline ostream& operator << (ostream &out, HgraphEdge* f)
{ out << "(ptr->)"; f->Print(out); return out; }

////////////////////////////////////////////////////////////////////////////

class Hgraph : public Object {

protected:
  HgraphNode* source;  //  pointer to node - typically a source, or start state
  HgraphNode* sink;    //  pointer to node - typically a sink, or accepting state
  Object*     data;    //  graph's data object
  
public:
  List    nodes;   // list of all nodes in this graph 
  List    edges;   // list of all edges in this graph
  boolean printNode;
  boolean printEdge;

  Hgraph();
 ~Hgraph();

  integer   NumNodes() { return nodes.Count();}
  integer   NumEdges() { return edges.Count();}

  void      Put(Object* o) { data = o;}  
  Object*   Get()          { return data;}  

  // you can just add nodes directly to "nodes" List, or
  HgraphNode* NewNode( boolean append = FALSE, Object* o = NULL);
  HgraphNode* NewNode( Object* o, boolean append = FALSE) { return NewNode( append, o);}

  // you can just add edges directly to "edges" List, or
  HgraphEdge* NewEdge( boolean append = FALSE, Comparable* o = NULL);
  HgraphEdge* NewEdge( Comparable* o, boolean append = FALSE) { return NewEdge( append, o);}

  void    Connect             ( HgraphNode* from, HgraphNode* to, HgraphEdge* e);
  void    ConnectToNet        ( HgraphNode* from, HgraphEdge* e);
  void    ConnectFromNet      ( HgraphNode* to,   HgraphEdge* e);
  void    ConnectUnique       ( HgraphNode* from, HgraphNode* to, HgraphEdge* e);
  void    ConnectUniqueToNet  ( HgraphNode* from, HgraphEdge* e);
  void    ConnectUniqueFromNet( HgraphNode* to,   HgraphEdge* e);
  boolean Disconnect          ( HgraphNode* from, HgraphNode* to, HgraphEdge* e);

  void    ReplaceNode( HgraphNode* oldn, HgraphNode* newn); 
  void    ReplaceEdge( HgraphEdge* olde, HgraphEdge* newe); 

  boolean RemoveEdge( HgraphEdge* e);
  boolean RemoveNode( HgraphNode* n);
  boolean BypassNode( HgraphNode* n);
  void    RemoveFloatingEdges();
  void    RemoveFloatingNodes();
 
  void Renumber();
  void Addgraph(Hgraph*);
  void TransferGraph(Hgraph *);
  void AdjustSourceFirst();
  void MarkConnectedNodes(HgraphNode *);
  void BackMarkConnectedNodes(HgraphNode *);
  void PruneUnconnectedNodes(HgraphNode *);

  void OrderNodeList();          // reorders list of nodes for in-order traversal
  void ReverseOrderNodeList();   // reorders list of nodes for reverse-order traversal

  void TouchNodes(boolean b) { Hgraph_ForEach_Node( this, node) {
                                 node->Touch(b);
                               } Hgraph_End_ForEach;
                             }

  HgraphNode* MakeSource( Comparable *o = NULL) { return source = new HgraphNode(o);}
  HgraphNode* Getsource()                   { return source;}
  HgraphNode* Source()                      { return source;}

  HgraphNode* MakeSink( Comparable *o = NULL)   { return sink = new HgraphNode(o);}
  HgraphNode* Getsink()                     { return sink;}
  HgraphNode* Sink()                        { return sink;}
  
  virtual  integer isa()    { return HGRAPH;}
  virtual  void    Print( ostream& out = cout); 
};

inline boolean
HgraphNode::SourceNode() {
  List_ForEach_Type( in_edges, HgraphEdge, edge) {
    if (edge->Fanin() != 0) return FALSE;
  } List_End_ForEach;
  return TRUE;
}
inline boolean
HgraphNode::SinkNode() {
  List_ForEach_Type( out_edges, HgraphEdge, edge) {
    if (edge->Fanout() != 0) return FALSE;
  } List_End_ForEach;
  return TRUE;
}
inline integer
HgraphNode::WhichOutEdge( HgraphEdge* e) { 
  return out_edges.IdentityNum( e);
}
inline integer
HgraphNode::WhichInEdge ( HgraphEdge* e) {
  return  in_edges.IdentityNum( e);
}


inline ostream& operator << (ostream &out, Hgraph& f)
{ f.Print(out); return out; }

inline ostream& operator << (ostream &out, Hgraph* f)
{ out << "(ptr->)"; f->Print(out); return out; }

#endif // HGRAPH_H

