#include "indarray.h"
#include "intobj.h"
#include "bdd.h"


IndexArray::IndexArray( IndexArray* arr) {
  timestamp = arr->timestamp; 
  size      = arr->size;
  count     = arr->count;
  ptr       = arr->ptr;
  index     = new integer[ size];
  level     = new integer[ size];
  for (integer i=0; i<count; i++) {
    index[i] = arr->index[i];
    level[i] = arr->level[i];
  }
}

IndexArray::IndexArray( List* list) {
  timestamp = -1; 
  size  = list->Count();
  count = 0;
  ptr   = 0;
  index = new integer[ size];
  level = new integer[ size];
  List_ForEach_Object( list, o) {
    ASSERT( o->isa() == INT_OBJ, "Specify indexes by list of IntObj's");
    index[ count++] = ((IntObj*) o)->val;
  } List_End_ForEach;
}

IndexArray::IndexArray( OList* list) {
  timestamp = -1; 
  size  = list->Count();
  count = 0;
  ptr   = 0;
  index = new integer[ size];
  level = new integer[ size];
  OList_ForEach_Object( list, o) {
    ASSERT( o->isa() == INT_OBJ, "Specify indexes by list of IntObj's");
    index[ count++] = ((IntObj*) o)->val;
  } OList_End_ForEach;
}

IndexArray::IndexArray( ODList* list) {
  timestamp = -1; 
  size  = list->Count();
  count = 0;
  ptr   = 0;
  index = new integer[ size];
  level = new integer[ size];
  ODList_ForEach_Object_Descend( list, o) {
    ASSERT( o->isa() == INT_OBJ, "Specify indexes by list of IntObj's");
    index[ count++] = ((IntObj*) o)->val;
  } ODList_End_ForEach;
}

IndexArray::IndexArray( integer beg, integer end, integer mod, integer off) {
  timestamp = -1; 
  while( beg%mod < off )
    beg++;
  size  = (end-beg)/mod + 1;

  index = new integer[ size];
  level = new integer[ size];
  ptr   = 0;
  count = 0;

  for (; beg<=end; beg+=mod) {
    index[count++] = beg;
  }
}

IndexArray::IndexArray( BDDmanager *bdd, BDDnode *vars ) {
  BDDnode *tptr = vars;

  size = 0;
  timestamp = -1;
  count = 0;
  ptr = 0;

  while( tptr != bdd->One() ) {
    ASSERT( tptr != bdd->Zero() &&
	    (tptr->Then() == bdd->Zero() || tptr->Else() == bdd->Zero()),
	    "Non cube passed to index array constructor." );
    if( tptr->Then() == bdd->Zero() )
      tptr = tptr->Else();
    else {
      tptr = tptr->Then();
      size++;
      }
    }
  index = new integer[ size];
  level = new integer[ size];
  tptr = vars;
  while( tptr != bdd->One() ) {
    if( tptr->Else() == bdd->Zero() ) {
      index[count++] = tptr->GetIndex();
      tptr = tptr->Then();
      }
    else
      tptr = tptr->Else();
    }
  }

IndexArray::IndexArray( IndexArray* arr, integer lv, integer hv) { 
  timestamp = arr->timestamp; 

  size      = hv-lv;
  count     = hv-lv;
  ptr       = 0;

  index     = new integer[ size];
  level     = new integer[ size];


  for (integer i=0; i<count; i++) {
    index[i] = arr->index[lv+i];
    level[i] = arr->level[lv+i];
  }
}

void
IndexArray::Merge( IndexArray* arr) {
  integer* old_index = index;
  integer* old_level = level;

  timestamp = -1; 
  size      = count + arr->count;
  count     = count;	// temporary, incremented below
  ptr       = 0;
  index     = new integer[ size];
  level     = new integer[ size];

  int i;
  for ( i=0; i<count; i++) {
    index[i] = old_index[i];
    level[i] = old_level[i];
  }
  for (i=0; i<arr->count; i++) {
    index[count  ] = arr->index[i];
    level[count++] = arr->level[i];
  }

  delete old_index;
  delete old_level;
}

void
IndexArray::Diff( IndexArray* arr) {
  if (arr->Count() == 0) 
    return;

  ASSERT( arr->timestamp == timestamp, "Order array before calling diff");
  integer* old_index = index;
  integer* old_level = level;
  integer  old_count = count;
  integer  arr_ptr   = 0;

  index = new integer[ size];
  level = new integer[ size];

  for (ptr=0, count=0; ptr<old_count; ptr++) {
    while ( old_level[ptr] < arr->level[arr_ptr] && arr_ptr+1 < arr->Count())
      arr_ptr++;

    if ( old_level[ptr] != arr->level[arr_ptr]) {
      index[count  ] = old_index[ptr];
      level[count++] = old_level[ptr];
    }
  }
  ptr = 0;

  delete old_index;
  delete old_level;
}

void
IndexArray::Qsort_Descend( integer left, integer right ) {  // K&R page 110
  if ( left >= right) return;
  Swap( left, (left+right)/2);

  integer last = left;

  for (integer i=left+1; i<=right; i++)
    if ( level[i] > level[left])
      Swap( ++last, i);
  Swap( left, last);
  Qsort_Descend( left,   last-1 );
  Qsort_Descend( last+1, right );
  }


// this doesn't work right (but the above does)

void
IndexArray::Qsort_Ascend( integer left, integer right ) {  // K&R page 110
  if ( left >= right) return;
  Swap( left, (left+right)/2);

  integer last = left;

  for (integer i=left+1; i<=right; i++)
    if ( level[i] < level[left])
      Swap( ++last, i);
  Swap( left, last);
  Qsort_Ascend( left,   last-1 );
  Qsort_Ascend( last+1, right );
  }

void
IndexArray::Order( integer *bddlevel, integer ts )
  {
  if( ts == timestamp )
    return;
    
  for( int i=0 ; i<count ; i++ ) {
    level[i] = bddlevel[index[i]];
  }
  
  Qsort_Descend( 0, count-1 );

  timestamp = ts;
  }


boolean
IndexArray::Find( integer lev) {
  ASSERT( count > 0, "Tried to use empty Index Array");
  ptr = MIN( ptr, count-1);

  if ( level[ptr]<lev) {
    while ( ptr>0 && level[ptr]<lev)
      ptr--;
  } else if ( level[ptr]>lev) {
    while ( ptr<count-1 && level[ptr]>lev)
      ptr++;
  }
  return level[ptr]==lev;
}

int
IndexArray::GetHeight( int lev) {
  ASSERT( count > 0, "Tried to use empty Index Array");
  if ( level[ptr]<lev) {
    while ( ptr>0 && level[ptr]<lev)
      ptr--;
  } else if ( level[ptr]>lev) {
    while ( ptr<count-1 && level[ptr]>lev)
      ptr++;
  }
  if( level[ptr]==lev )
    return count - ptr - 1;
  else
    return -1;
}

void
IndexArray::Print( ostream& out) { 
  cout << "Max Size = " << size << endl;
  cout << "Current Size = " << count << endl;
  cout << "timestamp = " << timestamp << endl;
  if (timestamp != -1) {
    for (integer i=0; i<count; i++) {
      out << index[i] << '(' << level[i] << ")\n";
    }
  } else {
    for (integer i=0; i<count; i++) {
      out << index[i] << '(' << level[i] << ")\n";
    }
  }
}
