#ifndef INDARRAY_H
#define INDARRAY_H

#include "object.h"
#include "list.h"
#include "odlist.h"

class BDDnode;
class BDDmanager;

class IndexArray: public Object {


private:
  integer  size;
  integer  timestamp;
  integer  count;

  integer  ptr;

  integer* index;
  integer* level;

  void Swap( integer i, integer j) {
    integer t = index[i];
    index[i]  = index[j];
    index[j]  = t;

    t        = level[i];
    level[i] = level[j];
    level[j] = t;
  }

  void Qsort_Descend( integer left, integer right );
  void Qsort_Ascend( integer left, integer right );

protected:
  // functions used by BDDmanager exclusively
  integer operator[] (integer pos)		{ return level[pos]; }
  void	  StampIt(integer ts)			{ timestamp = ts; }
  void    Order( integer *levels, integer ts=-2 );	// this also stamps the time

public:
  // user functions

  IndexArray( integer s) { size = s; timestamp = -1; count = 0; ptr = 0;
                           index = new integer[ size];
			   level = new integer[ size];
			  }
  IndexArray( IndexArray* arr);
  IndexArray( List* list);
  IndexArray( OList* list);
  IndexArray( ODList* list);
  IndexArray( integer beg, integer end, integer mod=1, integer off=0);
  IndexArray( BDDmanager *bdd, BDDnode *vars);
  IndexArray( IndexArray* arr, integer lv, integer hv);  // pv is the pointer partition value

 ~IndexArray()			{ delete index; delete level;}

  void    Clear()               { timestamp = -1; count = 0; ptr = 0;}

  void Merge( IndexArray* arr);    //insert indexes from arr into present array
  void Diff( IndexArray* arr);     //deletes indexes which also exist in arr

  boolean Find( integer level);
  integer NextLevel()           { ASSERT(ptr<count-1, ""); return level[ ptr+1];}
  integer NextIndex()           { ASSERT(ptr<count-1, ""); return index[ ptr+1];}
  integer CurrLevel()           { return level[ ptr];}
  integer CurrIndex()           { return index[ ptr];}
  integer PrevLevel()           { ASSERT(ptr>0      , ""); return level[ ptr-1];}
  integer PrevIndex()           { ASSERT(ptr>0      , ""); return index[ ptr-1];}

  integer GoNextLevel()         { if (ptr<count-1) ++ptr; return level[ptr]; } 
  integer GoPrevLevel()         { if (ptr>0) --ptr; return level[ptr]; }
  integer GoHighestLevel()      { ptr=0; return level[ptr];}
  integer GoLowestLevel()       { ptr=count-1; return level[ptr]; }

  // Risky Functions!
  void SetCurrLevel(int cl)     { level[ ptr ] = cl; }  
  void SpecialOrder()           { Qsort_Descend( 0, count-1); }


  integer GetHeight( int level );
  integer TimeStamp() 		{ return timestamp; }
  integer Size () 		{ return size; }
  integer Count() 		{ return count; }
  integer Position()          { return count-ptr-1; }
  boolean AtLowestLevel()       { return ptr == count-1;}
  integer LowestLevel()         { return level[ count-1];}
  integer LowestIndex()         { return index[ count-1];}
  boolean AtHighestLevel()      { return ptr <= 0;}	// can be -1 if array is empty
  integer HighestLevel()        { return level[ 0];}
  integer HighestIndex()        { return index[ 0];}
  integer Level( integer pos) 	{ ASSERT( pos>=0 && pos<size, "violated array bounds");
			    	  return level[ pos]; }
  integer Index( integer pos) 	{ ASSERT( pos>=0 && pos<size, "violated array bounds");
			    	  return index[ pos]; }
  void Put( integer i )		{ ASSERT( count < size, "violated array bounds");
				  index[count++] = i; timestamp = -1; }
  virtual integer isa()		{ return IND_ARRAY; }
  virtual void Print( ostream& out = cout);

  friend class BDDmgr;
};

#endif // INDARRAY_H

