#ifndef INTOBJ_H      
#define INTOBJ_H

#include <iostream.h>
#include "compable.h"

#define INT_PRIME 7919

class IntObj: public Comparable {
public:
  integer val;

  IntObj( void)         { val = 0;}
  IntObj( integer v)    { val = v;}

  integer operator=( integer i) { val = i; return i;}

  // You can use IntObj's as int's
  operator int() { return val; }

  virtual integer isa() { return INT_OBJ;}
  virtual Object* Copy(){ return new IntObj( val);}   
  virtual void ReadBinary( istream &in ) {
    in.read( (unsigned char*)&val, sizeof(val) ); }
  virtual void WriteBinary( ostream &out ) {
    out.write( (unsigned char*)val, sizeof(val) ); }
  static void WriteBinary( int v, ostream &out ) {
    out.write( (unsigned char*)v, sizeof(v) ); }
  static void ReadBinary( int &v, istream &in ) {
    in.read( (unsigned char*)&v, sizeof(v) ); }

  virtual integer Compare( Comparable* c) {
    ASSERT (c->isa() == INT_OBJ, "expecting IntObj");
    return val - ((IntObj*) c)->val;
  }

  virtual uinteger Hash( uinteger i=0) {
    return (uinteger) ( (val << 29) ^
			(val << 24) ^
			(val << 23) ^
			(val << 19) ^
			(val << 17) ^
			(val << 14) ^
			(val << 11) ^
			(val << 9) ^
			(val << 8) ^
			(val << 4) ^
			(val << 1) ^
			val ^
			i ^
			0x75bcd15 ); }
  virtual void Print( ostream& out = cout){ out << val;}
  virtual void Scan ( istream&  in = cin ){ in  >> val;}

#ifdef _BIG_ENDIAN
  virtual void Write( ostream& out ) {
    out.write( (unsigned char*)&val, sizeof(val) );
    }
  static void WriteInt( ostream& out, int v ) {
    out.write( (unsigned char*)&v, sizeof(v) );
    }
  virtual void Read( istream& in ) {
    in.read( (unsigned char*)&val, sizeof(val) );
    }
  static int ReadInt( istream& in ) {
    int v;
    in.read( (unsigned char*)&v, sizeof(v) );
    return v;
    }
  static void WriteShort( ostream& out, short v ) {
    out.write( (unsigned char*)&v, sizeof(v) );
    }
  static short ReadShort( istream& in ) {
    short v;
    in.read( (unsigned char*)&v, sizeof(v) );
    return v;
    }
#else
  virtual void Write( ostream& out ) {
    int i;
    unsigned char *p = (unsigned char*)&val;
    for( i=sizeof(val)-1 ; i>=0 ; i-- )
      out.put(p[i]);
    }
  static void WriteInt( ostream& out, int v ) {
    int i;
    unsigned char *p =(unsigned char*) &v;
    for( i=sizeof(v)-1 ; i>=0 ; i-- )
      out.put(p[i]);
    }
  virtual void Read( istream& in ) {
    int i;
    unsigned char *p =(unsigned char*) &val;
    for( i=sizeof(val)-1 ; i>=0 ; i-- )
      in.get(p[i]);
    }
  static int ReadInt( istream& in ) {
    int i,v;
    unsigned char *p =(unsigned char*) &v;
    for( i=sizeof(v)-1 ; i>=0 ; i-- )
      in.get(p[i]);
    return v;
    }
  static void WriteShort( ostream& out, short v ) {
    int i;
    unsigned char *p =(unsigned char*) &v;
    for( i=sizeof(v)-1 ; i>=0 ; i-- )
      out.put(p[i]);
    }
  static short ReadShort( istream& in ) {
    int i;
    short v;
    unsigned char *p =(unsigned char*) &v;
    for( i=sizeof(v)-1 ; i>=0 ; i-- )
      in.get(p[i]);
    return v;
    }
#endif

};

inline ostream& operator << (ostream &out, IntObj& f)
{ f.Print(out); return out; }

inline ostream& operator << (ostream &out, IntObj* f)
{ out << "(ptr->)"; f->Print(out); return out; }

#endif // INTOBJ_H

