////////////////////////////////////////////////////////////////////////////
//
//  intSparse.C : sparse integer matrix class         
//
////////////////////////////////////////////////////////////////////////////

#include <iostream.h>
#include <stdlib.h>
#include "isparse.h"

IntSparse::IntSparse(int xs, int ys) : BitSparse( xs, ys) {;}

IntSparse::IntSparse(istream& in) : BitSparse( in, FALSE) { Set( in);}

void
IntSparse::Set(int i, int j, int v) {
  ASSERT( (i < xsize && j <ysize), "array index out of range");
  if ( v == 0) {
    BitSparse::Clear( i, j);
    return;
  }

  IntSparseCell *p, *px, *py;
  for (py=0,p=(IntSparseCell*)x[i]; p!=0; py=p,p=(IntSparseCell*)p->nexty) {
    if (p->y > j)
      break;
    if (p->y == j) {
      p->val = v;
      return;
    }
  }
  for (px=0,p=(IntSparseCell*)y[j]; p!=0; px=p,p=(IntSparseCell*)p->nextx) {
    if (p->x > i)
      break;
    if (p->x == i) {
      p->val = v;
      return;
    }
  }

  IntSparseCell *q = new IntSparseCell;
  q->val = v;
  q->x   = i;
  q->y   = j;

  if (px == 0) {
    q->nextx = y[j];
    y[j] = q;
  } else {
    q->nextx = px->nextx;
    px->nextx = q;
  }
  if (py == 0) {
    q->nexty = x[i];
    x[i] = q;
  } else {
    q->nexty = py->nexty; 
    py->nexty = q;
  }
  ++count;
  ++num_in_col[i];
  ++num_in_row[j];
}

void
IntSparse::Incr( int i, int j) {
  ASSERT( (i < xsize && j <ysize), "array index out of range");
  IntSparseCell *p;

  for (p=(IntSparseCell*)x[i]; p!=0; p=(IntSparseCell*)p->nexty) {
    if (p->y > j) {
      Set( i, j, 1);
      return;
    }
    if (p->y == j) {
      if (++(p->val) == 0)
	BitSparse::Clear( i, j);
      return;
    }
  }
  Set( i, j, 1);
}

void
IntSparse::Decr( int i, int j) {
  ASSERT( (i < xsize && j <ysize), "array index out of range");
  IntSparseCell *p;

  for (p=(IntSparseCell*)x[i]; p!=0; p=(IntSparseCell*)p->nexty) {
    if (p->y > j) {
      Set( i, j, -1);
      return;
    }
    if (p->y == j) {
      if (--(p->val)==0)
	BitSparse::Clear( i, j);
      return;
    }
  }
  Set( i, j, -1);
}

int
IntSparse::Read(int i, int j) {
  ASSERT( (i < xsize && j <ysize), "array index out of range");
  IntSparseCell *p;

  for (p=(IntSparseCell*)x[i]; p!=0; p=(IntSparseCell*)p->nexty) {
    if (p->y > j)
      return 0;
    if (p->y == j)
      return p->val;
  }
  return 0;
}


void
IntSparse::Print( ostream& out) {                   // better printer function

  IntSparseCell *p;
  int i, j, n;

  out << "   ";
  for (j=0; j<xsize; ++j)
    out << "--";
  out << "-\n";
  for (i=0; i<ysize; ++i) {
    out.width(3);
    out << i;
    for (n=0, p=(IntSparseCell*)y[i]; p!=0; n=p->x+1, p=(IntSparseCell*)p->nextx) {
      for (j=0; j < (p->x - n); ++j)
	out << "| ";
      out << "|" <<p->val;
    }
    for (j=0; j < xsize-n; ++j)
      out << "| ";
    out << "|\n   ";
    for (j=0; j<xsize; ++j)
      out << "--";
    out << "-\n";
  }
}

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

void
IntSparse::Set( istream&  in) {
  integer r;
  integer c;
  integer v;
  integer t_count;

  in >> t_count;
  for (int i=0; i < t_count; i++) {
    in >> c >> r >> v;
    Set( c, r, v);
  }
}

void
IntSparse::Clear( istream&  in) {
  integer r;
  integer c;
  integer v;
  integer t_count;

  in >> t_count;
  for (int i=0; i < t_count; i++) {
    in >> c >> r >> v;
    BitSparse::Clear( c, r);
  }
}

