#ifndef INT_SPARSE_H
#define INT_SPARSE_H

////////////////////////////////////////////////////////////////////////////
//
//  IntSparse.h : sparse integer matrix class 
//  kgb (aka Chuck Monahan)
//
////////////////////////////////////////////////////////////////////////////

#include "bsparse.h"

////////////////////////////////////////////////////////////////////////////
// macro's from hell (by your friendly kgb agent Chuck Monahan)

#define IntSparse_ForEach_CellVal( m, r, c, v) \
{ \
  int c, v; \
  IntSparseIterator matI; \
  for ( int r = 0; r < m->Getysize(); r++) \
    for ( matI.AcrossRow(m, r); !matI.End(); matI.NextCol() ) { \
            c = matI.GetCol(); \
	    v = matI.GetVal();      

#define IntSparse_End_ForEach }}

// You are now leaving the macro zone
///////////////////////////////////////////////////////////////////////////


class IntSparseCell : public BitSparseCell {
  int val;

  friend class IntSparseIterator;
  friend class IntSparse;
};

class IntSparse : public BitSparse {

public:

  IntSparse(int x, int y);
  IntSparse(istream& in);

  void Incr( int x, int y);
  void Decr( int x, int y);
  
  int  Read (int x, int y);
  void Set ( int x, int y, int val);
  void Set ( int xx, int yy) { Set( xx, yy, 1);}      // overload Set( x,y) from bitSparse

  virtual integer isa()    { return INT_SPARSE;}
  virtual void Print( ostream& out = cout);

  void Set(   istream& in);
  void Clear( istream& in);

  friend class IntSparseIterator;
};

inline ostream& operator << (ostream &out, IntSparse& f)
{ f.Print(out); return out; }

inline ostream& operator << (ostream &out, IntSparse* f)
{ out << "(ptr->)"; f->Print(out); return out; }

class IntSparseIterator : public BitSparseIterator {

public:
  IntSparseIterator() : BitSparseIterator() {};

  int GetVal()  { return ((IntSparseCell *) tmp)->val;}
};

#endif // INT_SPARSE_H


