///////////////////////////////////////////////////////////////////////////
//
//  lex.C : a class interface to lex or flex, calles yylex
//  Andrew Seawright
//
////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include "lex.h"

         Lex   *Lexthis;
extern   FILE  *yyin;
extern   int    yylex();

///////////////////////////////////////////////////////////////////////////
// extras from flex... check flex man page

struct yy_buffer_state;
typedef struct yy_buffer_state* YY_BUFFER_STATE;

extern YY_BUFFER_STATE yy_create_buffer(FILE* file, int size);
extern            void yy_switch_to_buffer(YY_BUFFER_STATE new_buffer);
extern            void yy_delete_buffer(YY_BUFFER_STATE b);
extern            void yy_flush_buffer(YY_BUFFER_STATE b);
extern            void yyrestart(FILE* file);

////////////////////////////////////////////////////////////////////////////

Lex::Lex() {                   // stdin constructor
  unget = FALSE;
  linenum = 1;
  filename = 0;
  token = -1;
  id = 0;   
  val = 0;
  yyin  = stdin;              // setup up stdin as input stream;
}

////////////////////////////////////////////////////////////////////////////

Lex::Lex(char *fn) {           // input file constructor
  unget = FALSE;
  linenum = 1; 
  filename = fn;
  token = -1;
  id = 0;
  val = 0; 

  if ((yyin = fopen(filename, "r")) == NULL) {
    printf("Can't open the file %s\n", filename);
    exit (-1);
  }
}
////////////////////////////////////////////////////////////////////////////

Lex::~Lex() {          // destructor
  fclose(yyin);
}

////////////////////////////////////////////////////////////////////////////

void
Lex::Rewind() {        // reanalyze the input file
  if ( filename == NULL)
    FATAL( "Can't rewind standard input");

  unget = FALSE;
  linenum = 1; 
  token = -1;
  id = 0;
  val = 0; 
  FILE* fp;
  if ((fp = fopen(filename, "r")) == NULL) {
    printf("Can't open the file %s\n", filename);
    exit (-1);
  }
  
  fclose( yyin);
  yyrestart( fp);
}
////////////////////////////////////////////////////////////////////////////

void
Lex::Restart(char *fn) {           // analyze a new file
  unget = FALSE;
  linenum = 1; 
  filename = fn;
  token = -1;
  id = 0;
  val = 0; 
  FILE* fp;
  if ((fp = fopen(filename, "r")) == NULL) {
    printf("Can't open the file %s\n", filename);
    exit (-1);
  }
  
  fclose( yyin);
  yyrestart( fp);
}
////////////////////////////////////////////////////////////////////////////

int
Lex::Get() {          // get a token
  if (unget) {
    unget = FALSE;
    return token;
  } else {
    Lexthis = this;
    return (token = yylex());
  }
}

void Lex::unGet() { // unget a token
  unget = TRUE;
}

////////////////////////////////////////////////////////////////////////////

void
Lex::error(char *s) { // tell user the bad news

  if (filename != 0)
    (void)fprintf(stderr,"%s : ", filename);
  (void)fprintf(stderr,"Lex : line=%d : %s\n", linenum, s);
  exit (-1);
}
////////////////////////////////////////////////////////////////////////////  

void
Lex::LineNumCorrection() {
  char *s; 
  for (s=id; *s; ++s)
    if (*s == '\n')
      ++linenum;
}

////////////////////////////////////////////////////////////////////////////



