////////////////////////////////////////////////////////////////////////////
//
//  list.C : linked lists
//  Andrew Seawright et. al
//  
//  $Header: /usr6/research/code/homebrew/RCS/list.C,v 1.22 1998/09/14 23:25:40 crews Exp $
////////////////////////////////////////////////////////////////////////////

#include <iostream.h>
#include <stdlib.h>
#include "list.h"

////////////////////////////////////////////////////////////////////////////

List::List(List& l) {
  if (count = l.count) {
    Link *p, *q;
    head = p = new Link( l.head->data, 0);
    for ( q = l.head->next; q; p = p->next, q = q->next)
      p->next = new Link( q->data, 0);

  } else
    head = 0;
}

////////////////////////////////////////////////////////////////////////////

void
List::Add(Object& p) {                    // List add
  Link *l = new Link(p, head);
  head = l;
  ++count;
}

////////////////////////////////////////////////////////////////////////////

void
List::Append(Object& p) {

  if (head == 0) {
    head = new Link(p,0);
  } else {
    Link *q = head;
    for (; q->next; q=q->next) {;}
    q->next = new Link(p,0);
  }
  ++count;
}

boolean
List::AppendUnique(Object &p) {

  if (head == 0) {
    head = new Link(p,0);
    ++count;
    return TRUE;
  } else {
    Link *q = head;
    for (; q->next && q->data != &p; q=q->next);
    if (q->data != &p) {
      q->next = new Link(p,0);
      ++count;
      return TRUE;
    }
  }
  return FALSE;
}

////////////////////////////////////////////////////////////////////////////

// Top returns the head element
// Pop returns the head element and removes it from list

Object*  
List::Top() {
  return (count)? head->data: 0;
}

Object*
List::Pop() {

  if (count--) {
    Object *d = head->data;
    Link *tmp = head;            
    head = head->next;           // advance head
    delete tmp;                  // destroy Link;
    return d;
    
  } else
    FATAL ("Pop from empty list");
  return NULL;
}

Object*
List::GetElement (integer i) {
  if ( i<0 || i>=count) return 0;

  Link *p = head;
  for ( integer cnt = 0; cnt < i; cnt++) {
    p=p->next;
  } 
  return p->data;
}

Object*
List::RemoveNum (integer i) {
  if ( i<0 || i>=count) return 0;

  count--;
  Link *p = head;
  Link **prev = &head;
  for ( integer cnt = 0; cnt < i; cnt++) {
    prev = &p->next;
    p=p->next;
  } 
  *prev = p->next;
  Object *d = p->data;
  delete p;
  return d;
}

integer
List::IdentityNum(Object *o) {
  integer pos = 0;
  for( Link *p = head; p ; p=p->next, pos++ )
    if( p->data == o)         // if pointers are the same
      return pos;
  return -1;
}

integer
List::ElementNum(Comparable *o) {
  integer pos = 0;
  for( Link *p = head; p ; p=p->next, pos++ )
    if( *(Comparable*)p->data == *o )
      return pos;
  return -1;
}

/////////////////////////////////

boolean
List::Member(Object *o) {

  for (Link *p = head; p; p=p->next) {
    if ( p && p->data == o) {
      return TRUE;
    }
  } 
  return FALSE;
}

boolean
List::Member(Comparable *o) {

  for (Link *p = head; p; p=p->next) {
    if ( p && *(Comparable *)p->data == *o) {
      return TRUE;
    }
  } 
  return FALSE;
}

Comparable*
List::GetMember(Comparable *o) {

  for (Link *p = head; p; p=p->next) {
    if (p && *(Comparable *)p->data == *o) {
      return (Comparable *) p->data;
    }
  } 
  return NULL;
}


////////////////////////////////////////////////////////////////////////////
// Remove all links in list, but maintains objects

void
List::Clear () {
  Link *d;
  while ( head) {
    d    = head;
    head = d->next;
    delete d;
  }
  count = 0;
}

////////////////////////////////////////////////////////////////////////////
// Remove all links in list, and deletes objects

void
List::CleanOut() {
  Link *d;
  while ( head) {
    d    = head;
    head = d->next;
    delete d->data;
    delete d;
  }
  count = 0;
}

////////////////////////////////////////////////////////////////////////////
// Remove all matching objects in list. Return TRUE if any removed.

boolean
List::RemoveAllof (Object *o) {

  Link   *p, *q, *n;
  boolean removed = FALSE;

  for (p = head, q=0; p; p=n) {
    n = p->next;
 
    if (p->data == o) {
      if (q) 
	q->next = p->next;
      else 
	head = p->next;
      delete p;
      count--;
      removed = TRUE;
     } else 
      q = p;
  } 
  return removed;
}

boolean
List::Remove(Object *o) {                // just remove one instance

  Link *point,*lastPoint;

  for ( point=head, lastPoint=NULL; point; point = point->next) {
    
    if (point->data == o) {
      if ( lastPoint) 
	lastPoint->next = point->next;
      else 
	head = point->next;
      delete point;
      count--;
      return TRUE;
    }
    lastPoint = point;
  } 
  return FALSE;
}


////////////////////////////////////////////////////////////////////////////
// List printer

void
List::Print( ostream& out) {
  if (count == 0) 
    out << "List : empty\n";
  else {
    out << "List : " << count << " item(s) ->\n";
    for (Link *p=head; p; p=p->next)
      out << p->data;
  }
}

void
List::ListPrint(ostream& out, char *delim) {
   for (Link *p=head; p; p=p->next) {
      out <<  p->data << delim;
  }
}

////////////////////////////////////////////////////////////////////////////

void
List::AddList(List *toadd) {
  List *list = new List( *toadd);
  Transfer( list);
  delete list;
}

void
List::AppendList(List *toadd) {
  List *list = new List( *toadd);
  TransferAppend( list);
  delete list;
}

void
List::Transfer(List *toadd) {

  if ( toadd->head == 0)  
    return;
  Link* p = toadd->head;
  for ( ; p->next; p = p->next);
  p->next = head;
  head    = toadd->head;
  count  += toadd->count;

  toadd->head  = 0;
  toadd->count = 0;
}

void
List::TransferAppend(List *toadd) {

  if ( head) {
    Link* p = head;
    for ( ; p->next; p = p->next);
    p->next = toadd->head;
  } else {
    head    = toadd->head;
  }
  count += toadd->count;

  toadd->head  = 0;
  toadd->count = 0;
}

void
List::Union(List *toadd) {

  for ( Link* p = toadd->head; p; p = p->next)
    AppendUnique( p->data);
}  

////////////////////////////////////////////////////////////////////////////

List*
List::RandomizedSublist( int n) {
  ASSERT( n>0 && n<=count, "Inappropriate sublist size");
  Object**  dataset = new Object*[count];

  integer c = 0;
  for (Link* l = head; c<count; l = l->next) {
    dataset[c++] = l->data;
  }
  
  integer r;
  List*   sublist = new List;
  while ( 0 < n--) {
    r = IRANDOM(c);
    sublist->Add( dataset[r]);
    dataset[r] = dataset[--c];
  }
  delete dataset;
  return sublist;
}

////////////////////////////////////////////////////////////////////////////

void
List::Reverse() {
  Link *t, *temp = head;
  head = 0;
  while ( temp) {
    t = temp->next;
    temp->next = head;
    head = temp;
    temp = t;
  }
}


