#ifndef LIST_H
#define LIST_H

/////////////////////////////////////////////////////////////////////////
//
//  list.h : linked list of objects  
//  Andrew Seawright
//
//  $Header: /usr6/research/code/homebrew/RCS/list.h,v 1.32 1998/09/14 23:26:00 crews Exp $
//
/////////////////////////////////////////////////////////////////////////

#include "object.h"
#include "compable.h"

class List_iterator;

class Link : public Object {
protected: 

  Link   *next;
  Object *data;

public:

  Link() {;}
  Link(Object &d, Link *n) { data=&d; next=n;}
  Link(Object *d, Link *n) { data=d; next=n;}
 ~Link() {;}

  Link *Next() { return next; }
  Object *Data() { return data; }

  friend class   List;
  friend class  OList;
  friend class ODList;
  friend class   List_iterator;
};

/////////////////////////////////////////////////////////////////////////
//

class List : public Object {
protected:
  Link   *head;
  integer count;

public:
  List(List& l);
  List() { head=0; count=0;}
 ~List() { Clear();}

  // Adds object to front 
  void Add(Object&);
  void Add(Object *o) { Add(*o); }

  // Appends object
  void Append(Object&);
  void Append(Object *o) { Append(*o); }

  // Appends object if unique
  // NOTE uniqueness of objects in List is determined by address
  boolean AppendUnique(Object&);
  boolean AppendUnique(Object *o) { return AppendUnique(*o); }

  void Union(  List *l);            // Appends the unique items from l
  void AddList(List *l);            // Add List l to front in order
  void AppendList(List *l);         // Append List l in order
  void Transfer(List *l);           // Adds List l to front in order, clearing l
  void TransferAppend(List *l);     // Appends List l in order, clearing l

  void Reverse();

  void    Clear();                  // Empties list
  void    CleanOut();               // Deletes objects as it empties list
  boolean RemoveAllof(Object *);    // Removes all items with same address
  boolean Remove(Object *);         // Removes first item with same address
  Object* RemoveNum( integer i );	// Remove and return ith element

  Object* Pop();                    // Removes & returns front element
  Object* Top();                    // Returns front element
  Object* GetElement( integer i);   // Returns ith object in list w/o deleting
  Object* operator[]( integer i)    { return GetElement(i);}

  List*   RandomizedSublist( int n); // make jumbled list of n random elements
  List*   RandomizedList()          { return RandomizedSublist( Count());}

  integer IdentityNum(Object    *o); // return o's pos in list (or -1)
  integer ElementNum(Comparable *o); // return "equiv o"'s pos in list (or -1)
  boolean Member(Object *o);         // Confirm's o's existance in list
  boolean Member(Comparable *o);     // Confirm's o's existance in list
  Comparable* GetMember(Comparable *o); // Return the Object or NULL
  integer Count()  {return count;}
  void    ListPrint( ostream& out = cout, char *delim = "\n");
  virtual void Print( ostream& out = cout);
  virtual integer isa() {return LIST;}

  friend class  OList;
  friend class ODList;
  friend class   List_iterator;
};

inline ostream& operator << (ostream &out, List& f)
{ f.Print(out); return out; }

inline ostream& operator << (ostream &out, List* f)
{ out << "(ptr->)"; f->Print(out); return out; }

/////////////////////////////////////////////////////////////////////////

class List_iterator : public Object {

  List *l;
  Link **nptr;
  Link *ptr;

public:

  List_iterator(List& li) { l = &li; ptr = l->head; nptr = &l->head; }
  List_iterator(List* li) { l = li; ptr = l->head; nptr = &l->head; }
 ~List_iterator()        { ;}
  void Reset( )		 { ptr = l->head; nptr = &l->head; }
  void operator ++()     { nptr = &ptr->next; ptr = ptr->next; }
  int end()	         { return (ptr == 0);}
  int last()             { return (ptr->next == 0);}
  Object *item()         { return (ptr)? ptr->data:0;}
  Object *nextItem()     { return (ptr)?((ptr->next)?ptr->next->data:0):0;}
  Object *Remove()	 { Object *o;
			   if( ptr ) {
			     l->count--;
			     Link *tmp = ptr;
			     o = ptr->data;
			     ptr = ptr->next;
			     *nptr = ptr;
			     delete tmp;
			   } else o = NULL;
			   return o; }
  // this insert puts the object before the current data (so never looks at o)
  void Insert(Object*o)	 { Link *tmp = new Link(o,ptr); // insert before current
			   *nptr = tmp;
			   l->count++;
			   ptr = tmp; } // move forward to current again

  // use this one if you want the List_ForEach to iterate the new items too
  void InsertAfter(Object*o)
			 { Link *tmp = new Link(o,ptr->next);
			   ptr->next = tmp;
			   l->count++; }
  Object* Replace(Object*o) { Object* old;
			   ASSERT( ptr, "Cant perform list replace probably because removed first");
			   old=ptr->data;
			   ptr->data=o;
			   return old; }



  void remove(List& x) { 
    Link *tmp = ptr; 
    ptr=ptr->next; 
    delete tmp; 
    x.head=ptr; --x.count;
    }
  void remove(List* x) { remove( *x);}
  };
////////////////////////////////////////////////////////////////////////////

#define List_ForEach_Object(x, o) \
        { \
	  Object *o; \
	  for ( List_iterator list__itr(x); !list__itr.end(); ++list__itr) { \
            o = list__itr.item();

#define List_ForEach_Type2(list__itr, x, t, o) \
        { \
	  t *o; \
	  for ( List_iterator list__itr(x); !list__itr.end(); ++list__itr) { \
            o = (t *) list__itr.item();

#define List_ForEach_Type(x, t, o) \
        { \
	  t *o; \
	  for ( List_iterator list__itr(x); !list__itr.end(); ++list__itr) { \
            o = (t *) list__itr.item();

#define List_ForEach_TypeCast(x, t, o) \
        { \
	  t o; \
	  for ( List_iterator list__itr(x); !list__itr.end(); ++list__itr) { \
            o = (t) list__itr.item();

#define List_ForEach_Object_Remove(x, o) \
        { \
	  Object *o; \
	  for ( List_iterator list__itr(x); !list__itr.end(); \
	        list__itr.remove(x)) { \
            o = list__itr.item();

#define List_ForEach_Type_Remove(x, t, o) \
        { \
	  t *o; \
	  for ( List_iterator list__itr(x); !list__itr.end(); \
	        list__itr.remove(x)) { \
            o = (t *) list__itr.item();

#define List_End_ForEach }}

////////////////////////////////////////////////////////////////////////////

#endif // LIST_H

