///////////////////////////////////////////////////////////////////////////////
//
//  mem.C : Memory Allocation / Management Object
//
//  Andrew Seawright      8/10/91
//
// $Header: /usr6/research/code/homebrew/RCS/mem.C,v 1.17 1998/09/14 23:26:34 crews Exp $
//
///////////////////////////////////////////////////////////////////////////////
   
#include <stdlib.h>
#include <iostream.h>
#include <string.h>
#include "object.h"
#include "mem.h"

#ifdef MEM_STATS
#define VERMALLOC( n ) ::new char[n];
extern MemHashTree mem_verify;
#else
#define VERMALLOC( n ) ::malloc(n)
#endif

#define MALLOC( n ) ::malloc(n)

//----------------------------------------------------------------------------

MemGroup::MemGroup() {

  int i;

  num_malloced = num_inuse = num_free = 0;
  malloc_count = alloc_count = num_big = 0;
  alloc_watch = -1;

   free     = (char**)::malloc(CACHE_BIGGEST*sizeof(char*));
  free_count= (int*)::malloc(CACHE_BIGGEST*sizeof(int));
   hits     = (int*)::malloc(CACHE_BIGGEST*sizeof(int));
  allocated = (int*)::malloc(CACHE_BIGGEST*sizeof(int));

  for (i=0; i < CACHE_BIGGEST; ++ i) {
     free[i] = 0;
     hits[i] = 0;
     allocated[i] = 0;
    free_count[i] = 0;
  }
}
//----------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////
//
// Note that the addresses of fresh pages allocated by malloc
// are not kept around for deleting (during the MemGroup destructor).
// These should be kept and ::freed during the destructor call.

MemGroup::~MemGroup() {

  ::free(hits);
  ::free(free);
  ::free(allocated);
  ::free(free_count);
}
//----------------------------------------------------------------------------

void *
MemGroup::Malloc(size_t n) {

  char *p;

//////////////////////////////////////////////////////////
# ifdef MEM_STATS					//
  ASSERT (n>=MEM_WORD_SIZE, "memory allocation error");	//
  ++alloc_count;					//
# endif							//
//////////////////////////////////////////////////////////

  if (n >= CACHE_BIGGEST) {        // is object bigger than the cache limit?

    p = (char *)VERMALLOC(n);
    if (p==0)
      return (void *)p;
//////////////////////////////////////////
#   ifdef MEM_STATS			//
    num_malloced += n; num_inuse += n;	//
    ++malloc_count;			//
    ++num_big;				//
#   endif				//
//////////////////////////////////////////

  } else {


//////////////////////////////////////////
#   ifdef MEM_STATS			//
    ++allocated[n];			//
#   endif				//
//////////////////////////////////////////
    ++hits[n];
 
    if (p = free[n]) { ////////////////// check cache freelist size n bin.   

      free[n] = *((char **)p);

//////////////////////////////////////////////////////////
#     ifdef MEM_STATS					//
      num_free  -= n; num_inuse += n; free_count[n]--;	//
#     endif						//
//////////////////////////////////////////////////////////
	
    } else {                      // allocate a fresh page
      
      p = (char *)MALLOC(PAGESIZE);
      if (p==0)
	return p; 

      // add it to the list of pages allocated
//      *(char**)p = page_list;
//      page_list = p;
//      p+=MEM_WORD_SIZE

      char *q = p + PAGESIZE - 3 * n;	// place the new memory on the free list
      			//    why 3n?	// one n for left over space,
					// one n because we return the last one
					// one n because the second to last is set to NULL
      free[n] = p;
      while ( p <= q )
	*(char**)p = (char*)(p+n), p+=n;
      *(char**)p = NULL;
      p+=n;				// now, left over space = q - p + 2n;

//////////////////////////////////////////////////
#     ifdef MEM_STATS				//
      num_free += p - free[n];			//
      num_malloced += PAGESIZE ; num_inuse += n;//
      ++malloc_count;				//
      free_count[n] += PAGESIZE / n - 1;	//
#     endif					//
//////////////////////////////////////////////////
/*************************************************
      // save left over space!	-- this is optional
      int m = q + 2 * n - p;
      q = p+n;
      if( m >= MEM_WORD_SIZE )
	{
        *(char**)(q) = free[m];
        free[m] = q;
//////////////////////////
#	ifdef MEM_STATS //
	free_count[m]++;//
#	endif		//
//////////////////////////
	}
//////////////////////////
#     ifdef MEM_STATS	//
      else		//
	num_free -= m;	//
#     endif		//
//////////////////////////
**************************************************/
    }

///////////////////////////////////////////////////////////////////////////////
# ifdef MEM_STATS
  if( verify.Put( (int)n, (int)p )==FALSE ) {
    size_t oldn = verify.Get( (int)p );
    verify.ForcePut( (int)n, (int)p );
    cerr << "MEMORY ERROR, unfreed block from MemGroup::new "
	 << "reallocated by MemGroup::new "
	 << (void*)p << " " << n << " bytes, was " << oldn << " bytes.\n";
    }
  size_t oldn = mem_verify.Get( (int)p );
  if( oldn!=FALSE )
    cerr << "MEMORY ERROR, unfreed block from ::new "
	 << "reallocated by MemGroup::new "
	 << (void*)p << " " << n << " bytes, was " << oldn << " bytes.\n";

# endif
///////////////////////////////////////////////////////////////////////////////
  }
  return (void *)p;
}
//----------------------------------------------------------------------------

void
MemGroup::Free(void *p, size_t n) {

  char *q;
 
  q = (char *)p;
//////////////////////////////////////////
# ifdef MEM_STATS			//
  for( int c=0; c<n ; q[c++]=0 );	// clear freed data for faster segfault
# endif					//
//////////////////////////////////////////
  // if n is >= Cache_Biggest, then we just allocated one of it, not a
  // whole page of them (~50), so just free it.
  if (n >= CACHE_BIGGEST) {

//////////////////////////
#   ifdef MEM_STATS	//
    num_malloced -= n;	//
    num_inuse -= n;	//
#   endif		//
//////////////////////////
    ::delete p;

  } else {

    // remember how many we have
//////////////////////////////////////////
#   ifdef MEM_STATS			//
    --allocated[n];			//
#   endif				//
//////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
#   ifdef MEM_STATS

    if( verify.Remove( (int)q )==FALSE )
      cerr << "MEMORY ERROR, MemGroup::delete of unallocated block "
	   << p << " " << n << " bytes\n";

    size_t oldn = mem_verify.Get( (int)p );
    if( oldn != FALSE )
      cerr << "MEMORY ERROR, MemGroup::delete of block allocated by ::new "  
	   << p << " " << n << " bytes, was " << oldn << " bytes.\n";

#   endif
///////////////////////////////////////////////////////////////////////////////

    // set first four bytes of p to be a pointer to the next word
    // of that size--making a linked list of memory locations of size
    // n. the size is now remembered by which linked list it is in.

    *((char **)q) = free[n];
    free[n] = q;

//////////////////////////
#   ifdef MEM_STATS	//
    num_inuse -= n;	//
    num_free += n;	//
    free_count[n]++;	//
#   endif		//
//////////////////////////
  }
}
//----------------------------------------------------------------------------

void
MemGroup::PrintStats( ostream& out) {

  int i;
  double r;

  out << "Memory Group Stats:\n";
  out << "-------------------\n\n";

  out << CACHE_BIGGEST<< "\t: Size of smallest uncached object.\n";
  out << PAGESIZE     << "\t: Size of memory page.\n";  
  out << num_malloced << "\t: Total number of bytes from system malloc.\n";
  out << num_inuse    << "\t: Total number of bytes in active use.\n";
  out << num_free     << "\t: Total number of bytes of recycled available mem.\n";
  out << num_big      << "\t: Total number of uncached allocations.\n";
  out << alloc_count  << "\t: Total number of allocation requests.\n";
  out << malloc_count << "\t: Total number of calls to system malloc.\n";

  r = 1.0 - (double) malloc_count / (double) alloc_count;
  out.precision(3);
  out  << r            << "\t: Hit ratio.\n";

  out << "\nTotal number of allocations vs. size:\n";
  out << "Size\tRequests\tIn Use\tFree\t\n";
  out << "----\t--------\t------\t----\t\n";
  for (i=1; i<CACHE_BIGGEST; ++i)
    if (hits[i] || allocated[i] || free_count[i])
      {
      out << i << "\t" << hits[i] << "\t\t" << allocated[i];
      out << "\t" << free_count[i] << "\t" << endl;
      }
  
////////////////
#ifdef MEM
#ifdef MEM_STATS
  out << "Global malloc and free stats:\n";
  out << "-----------------------------\n";
  out << _global_num_alloc << " bytes allocated by new\n";
  out << _global_num_inuse << " calls to new(not counting strdups, mallocs, callocs)\n";
  out << _global_num_freed << " calls to delete\n";
  i = sizeof(char*) * CACHE_BIGGEST * 3;
  i+= sizeof(int)   * CACHE_BIGGEST * 3;
  i+= sizeof(MemGroup);
  out << "\n(" << i << " bytes allocated for use by MemGroup.)" << endl;
  out << "Stats for MemGroup::verify\n";
  verify.PrintStats( out );
  out << "Stats for ::mem_verify\n";
  mem_verify.PrintStats( out );
  out << flush;
#endif
#endif
////////////////


}
//----------------------------------------------------------------------------


