#include "object.h"
#include <fstream.h>
#include "strobj.h"
#include <sys/types.h>
#include <unistd.h>

#include "memhash.h"

#ifdef MEM_STATS

int	_global_num_alloc=0;
int	_global_num_freed=0;
int	_global_num_inuse=0;

MemHashTree mem_verify;

void *operator new(size_t n) {
  _global_num_inuse++;
  _global_num_alloc += n;
  void *p = ::malloc(n);
  if( mem_verify.Put( (int)n, (int)p )==FALSE ) {
    int oldn = mem_verify.Get( (int) p );
    mem_verify.ForcePut( (int)n, (int)p );
    cerr << "MEMORY ERROR, unfreed block from ::new reallocated by ::new "
	 << (void*)p << " " << n << " bytes, was " << oldn << " bytes.\n";
    }

  int oldn = mem_default_group.verify.Get( (int)p );
  if( oldn!=FALSE ) {
    cerr << "MEMORY ERROR, unfreed block from MemGroup::new reallocated by ::new "
	 << (void*)p << " " << n << " bytes, was " << oldn << " bytes.\n";
    }

  return p;
  }

void operator delete(void *p) {
  _global_num_freed ++;
  _global_num_inuse --;
  size_t n;
  if( (n=mem_verify.Get( (int)p ))==FALSE ) {
    cerr << "MEMORY ERROR, ::delete of unallocated block "
	 << p <<"\n";
    }
  else mem_verify.Remove( (int) p );

  _global_num_alloc -= n;
  size_t oldn = mem_default_group.verify.Get( (int)p );
  if( oldn!=FALSE ) {
    cerr << "MEMORY ERROR, ::delete of block allocated by MemGroup::new "
	 << p << ", was " << oldn << " bytes.\n";
    }
  ::free(p);
  }
#endif
