#ifndef MEM_HASH_H
#define MEM_HASH_H

#include "stddefs.h"
#include "memhash.h"

class MemHashNode {

  MemHashNode    *right;
  MemHashNode    *left;
  int             hashval;
  int             key;
  int             data;
  // place other useful memory debugging info here

public:

  MemHashNode(int, int, int);
 ~MemHashNode() {;}

  void Data (int d)		{ data = d;    }
  void Key (int k)		{ key  = k;    }

  int  Key()			{ return key; }
  int  Data()			{ return data; }

  MemHashNode *Right()		{ return right; }
  void Right (MemHashNode *r)	{ right = r; }
  MemHashNode *Left()		{ return left; }
  void Left (MemHashNode *l)	{ left = l; }
  int Hashval()			{ return hashval; }
  void *operator new (size_t n );
  void operator delete (void *p );
};

////////////////////////////////////////////////////////////////////////////

class MemHashTree {
       
protected:
  static int freecount;
  static MemHashNode *freednodes;
  static MemHashNode *freenodes;
  int allocfail, allocsuc;
  int freefail, freesuc;
  int collisions, count;
  int bytes;
  MemHashNode *root;

public:
  MemHashTree( );
 ~MemHashTree( )	{ Clear(); }

  boolean Put (int data, int key);  // returns TRUE on success
  void ForcePut (int data, int key);       // jam it!
  int Get(int key); 
  boolean Remove( int key);
       
  void Clear(MemHashNode *n );

  void Clear()	{
    if( root )  {
      Clear(root);
      root = NULL;
      bytes=freefail=freesuc=allocfail=allocsuc=collisions=count=0;
      }
    }

  void    PrintStats ( ostream& out = cout);  // just stats
//  virtual integer isa() { return ; }
  int GetHashVal( int val )	{
    return (int)      ( (val << 29) ^
                        (val << 24) ^
                        (val << 23) ^
                        (val << 19) ^
                        (val << 17) ^
                        (val << 14) ^
                        (val << 11) ^
                        (val << 9) ^
                        (val << 8) ^
                        (val << 4) ^
                        (val << 1) ^
                        val ^
                        0x75bcd15 ); }
  void *operator new (size_t n ) { return ::malloc( n ); }
  void operator delete (void *p, size_t n ) { ::free((char*)p); }
  static void *AllocNode();
  static void FreeNode( void *);
};

#endif // MEM_HASH_H

