////////////////////////////////////////////////////////////////////////////
//
//  odlist.C : ordered double-linked lists
//
//  $Header: /usr6/research/code/homebrew/RCS/odlist.C,v 1.3 1994/11/15 18:32:25 kgb Exp $
////////////////////////////////////////////////////////////////////////////

#include <iostream.h>
#include <stdlib.h>
#include "odlist.h"
#include "heap.h"

////////////////////////////////////////////////////////////////////////////

ODList::ODList(ODList& l) {
  if (count = l.count) {
    ODLink *p, *q;
    head = p = new ODLink( l.head->data, NULL, NULL);
    for ( q = l.head->next; q; p = p->next, q = q->next)
      p->next = new ODLink( q->data, p, NULL);
    tail = p;
  } else {
    head = 0;
    tail = 0;
  }
}

////////////////////////////////////////////////////////////////////////////

ODList::ODList(OList& l) {
  boolean rev = (l.type == Descend);
  if (rev)
    l.Reverse();

  if (count = l.count) {
    ODLink *p;
    OLink  *q;
    head = p = new ODLink( l.head->data, NULL, NULL);
    for ( q = l.head->next; q; p = p->next, q = q->next)
      p->next = new ODLink( q->data, p, NULL);
    tail = p;
  } else {
    head = 0;
    tail = 0;
  }

  if (rev)
    l.Reverse();
}

////////////////////////////////////////////////////////////////////////////

ODList::ODList( List& l) {

  if (count = l.count) {
    Heap *heap = new Heap( count);
    for ( Link *p = l.head; p; p = p->next) {
      ASSERT( p->data->isComparable(), "Can't order uncomp items");
      heap->Add( (Comparable *) p->data);
    }

    ODLink *q = NULL;
    while( heap->HeapSize()) {
      if (q) {
	q->next = new ODLink( heap->Pop(), q, NULL);
	q       = q->next;
      } else {
	q = head = new ODLink( heap->Pop(), NULL, NULL);
      }
    }
    tail = q;
    delete heap;
  } else
   head = 0;
   tail = 0;
}

////////////////////////////////////////////////////////////////////////////

void
ODList::Insert(Comparable& p) {                    // List add
  ODLink *q, *t;
  if (!head) {
    head = tail = new ODLink(p, NULL, NULL);
  } else if ( head->data->Compare( &p) >= 0) {
    head->prev = new ODLink(p, NULL, head);
    head = head->prev;
  } else {
    for ( q = head; q->next && q->next->data->Compare( &p) < 0; q = q->next);
    t = new ODLink( p, q, q->next);
    q->next  = t;
    if ( tail == q)
      tail = t;
  }
  ++count;
}

////////////////////////////////////////////////////////////////////////////

void
ODList::InsertUnique(Comparable& p) {
  ODLink *q, *t;
  if (!head) {
    head = tail = new ODLink(p, NULL, NULL);
  } else if ( head->data->Compare(&p) >= 0) {
    if ( *(head->data) == p)        // dont add or increment count
      return;
    head->prev = new ODLink(p, NULL, head);
    head       = head->prev;
  } else {
    for ( q = head; q->next && q->next->data->Compare(&p) <= 0; q = q->next) 
      if ( *(q->next->data) == p) 
	return;
    t = new ODLink( p, q, q->next);
    if( q->next )		// are we at the tail?
      q->next->prev = t;	// no, then set the following prev pointer
    else
      tail = t;			// yes, then set the tail to the new link
    q->next = t;		// regardless, set the next pointer

//    if ( tail == q)
//      tail = t;
  }
  ++count;
}

////////////////////////////////////////////////////////////////////////////

// Top returns the head element
// Pop returns the head element and removes it from list

Comparable*  
ODList::Top() {
  return (count)? head->data: 0;
}

Comparable*
ODList::Pop() {
  if (!head) 
    FATAL ("pop from empty list");

  count--;
  Comparable *d = head->data;
  ODLink *tmp = head;
  if ( head == tail) {           // remove link
    head = tail = 0;
  } else {
    head = head->next;
  }
  delete tmp;                    // destroy Link;
  return d;
}

boolean
ODList::Member(Comparable *o) {

  for (ODLink *p = head; p; p=p->next) {
    if (*(p->data) == *o) 
      return TRUE;
  } 
  return FALSE;
}


Comparable*
ODList::GetElement (integer i) {
  if ( i<0 || i>=count) return NULL;

  ODLink *p = head;
  for ( integer cnt = 0; cnt < i; cnt++) {
    p=p->next;
  } 
  return p->data;
}

////////////////////////////////////////////////////////////////////////////
// Deletes links in list, but retains objects

void
ODList::Clear () {
  ODLink *d;
  while ( head) {
    d    = head;
    head = d->next;
    delete d;
  }
  tail  = 0;
  count = 0;
}

////////////////////////////////////////////////////////////////////////////
// Deletes links in list, and deletes objects

void
ODList::CleanOut () {
  ODLink *d;
  while ( head) {
    d    = head;
    head = d->next;
    delete d->data;
    delete d;
  }
  tail  = 0;
  count = 0;
}

////////////////////////////////////////////////////////////////////////////
// Remove all matching objects in list. Return TRUE if any removed.

boolean
ODList::RemoveAllof (Comparable *o) {

  ODLink  *l, *p;
  boolean removed = FALSE;

  for (p = head, l=NULL; p; p=p->next) {
 
    if (*(p->data) == *o) {
      if (l) 
	l->next = p->next;
      else 
	head    = p->next;
      if ( tail == p)
	tail = p->prev;
      delete p;
      p = (l)? l:head;
      count--;
      removed = TRUE;
     } else 
      l = p;
  } 
  return removed;
}

boolean
ODList::Remove(Comparable *o) {                // just remove one instance

  ODLink  *p,*l;
  boolean removed = FALSE;

  for ( p=head, l=NULL; p; p = p->next) {
    
    if (*(p->data) == *o) {
      if ( l) 
	l->next = p->next;
      else 
	head = p->next;
      delete p;
      count--;
      return TRUE;
    }
    l = p;
  } 
  return FALSE;
}


////////////////////////////////////////////////////////////////////////////
// List printer

void
ODList::Print( ostream& out) {
  if (count == 0) 
    out << "ODList : empty\n";
  else {
    out << "ODList : " << count << " item(s) ->\n";
    for (ODLink *p=head; p; p=p->next)
      p->data->Print( out);
  }
}

void
ODList::ListPrint(ostream& out, char *delim) {
   for (ODLink *p=head; p; p=p->next) {
      out <<  p->data << delim;
  }
}

////////////////////////////////////////////////////////////////////////////
// This version of merge merges toadd into the 
// ODList without creating a new list. After Merge(), toadd will cease
// to exist.

void
ODList::Merge(ODList *toadd) { // Old version of ODList, pre Feb. 24, 1994
  ODLink *p, *q, *l = NULL;

  for ( p = head, q = toadd->head; p && q;) {
    if ( p->data->Compare( q->data) >= 0) {
      if ( l) {
	l->next = new ODLink( q->data, l, p);
	l       = l->next;
      } else {
	head    = new ODLink( q->data, NULL, p);
	l       = head;
      }
      q = q->next;
    } else  {
      l = p;
      p = p->next;
    }
  } 
  while ( q) {                // throw remains on end of list
    l->next = new ODLink( q->data, l, p);
    l = l->next;
    q = q->next;
    tail = l;
  }
  count += toadd->count;
}  

void
ODList::Merge(List *toadd) {
  ODLink      *p, *l = NULL;
  Link       *s;
  Comparable *c;

  if ( toadd->count) {
    Heap *heap = new Heap( toadd->count);
    for ( s = toadd->head; s; s = s->next) {
      ASSERT( s->data->isComparable(), "Can't order uncomp items");
      heap->Add( (Comparable *) s->data);
    }
    
    for ( p = head, c = heap->Pop(); p && c;) {
      if ( *(p->data) >= *c) {
	if ( l) {
	  l->next = new ODLink( c, l, p);
	  l       = l->next;
	} else {
	  head = new ODLink( c, NULL, p);
	  l    = head;
	}
	c = (heap->HeapSize())? heap->Pop(): 0;
      } else  {
	l = p;
	p = p->next;
      }
    }
    while ( c) {                // throw remains on end of list
      l->next = new ODLink( c, l, p);
      l = l->next;
      c = heap->Pop();
      tail = l;
    }
    count += toadd->count;
  }
}  

////////////////////////////////////////////////////////////////////////////

void
ODList::Transfer(ODList *toadd) {
  ODLink *p, *q, *t, *l = NULL;

  if (!head) {                      // Adding to empty list
    head = toadd->head;
  } else {
    for ( p = head, q = toadd->head; p!=NULL && q!=NULL;) {
      if ( p->data->Compare( q->data) >= 0) {
	t       = q;
	q       = q->next;
	t->next = p;
	t->prev = l;

	if ( l!=NULL) {
	  l->next = t;
	  l       = l->next;
	} else {
	  head    = t;
	  l       = head;
	}
      } else  {
	l = p;
	p = p->next;
      }
    } 
    if ( q) {                // throw remains on end of list
      l->next = q;
      q->prev = l;
      tail = l;
    }
  }
  count += toadd->count;
  toadd->head  = 0;
  toadd->tail  = 0;
  toadd->count = 0;
}

////////////////////////////////////////////////////////////////////////////

void
ODList::Union(ODList *toadd) {
  ODLink *p, *q, *l = NULL;

  for ( p = head, q = toadd->head; p && q;) {
    if ( p->data->Compare( q->data) >= 0) {
      if ( *(p->data) != *(q->data)) {
	count++;
	if ( l) {
	  l->next = new ODLink( q->data, l, p);
	  l       = l->next;
	} else {
	  head = new ODLink( q->data, l, p);
	  l    = head;
	}
      } 
      q = q->next;
    } else {
      l = p;
      p = p->next;
    }
  } 
  while ( q) {                // throw remains on end of list
    count++;
    if ( l) {
      l->next = new ODLink( q->data, l, p);
      l       = l->next;
    } else {
      head = new ODLink( q->data, l, p);
      l    = head;
    }
    q = q->next;
    tail = l;
  }
}  

void
ODList::Union(OList *toadd) {
  ODLink *p, *l = NULL;
  OLink *q;

  if( toadd->Order() == Descend )
    {
    for ( p = tail, q = toadd->head; p && q;) {
      if ( p->data->Compare( q->data) <= 0) {
        if ( *(p->data) != *(q->data)) {
	  count++;
	  if ( l ) {
	    l->prev = new ODLink( q->data, p, l);
	    l       = l->prev;
	  } else {
	    tail = new ODLink( q->data, p, l);
	    l    = tail;
	  }
        } 
        q = q->next;
      } else {
        l = p;
        p = p->prev;
      }
    } 
    while ( q) {                // throw remains on end of list
      count++;
      if ( l) {
        l->prev = new ODLink( q->data, p, l);
        l       = l->prev;
      } else {
        tail = new ODLink( q->data, p, l);
        l    = tail;
      }
      q = q->next;
      head = l;
    }

  }
  else {
    for ( p = head, q = toadd->head; p && q;) {
      if ( p->data->Compare( q->data) >= 0) {
        if ( *(p->data) != *(q->data)) {
	  count++;
	  if ( l ) {
	    l->next = new ODLink( q->data, l, p);
	    l       = l->next;
	  } else {
	    head = new ODLink( q->data, l, p);
	    l    = head;
	  }
        } 
        q = q->next;
      } else {
        l = p;
        p = p->next;
      }
    } 
    while ( q) {                // throw remains on end of list
      count++;
      if ( l) {
        l->next = new ODLink( q->data, l, p);
        l       = l->next;
      } else {
        head = new ODLink( q->data, l, p);
        l    = head;
      }
      q = q->next;
      tail = l;
    }
  }
}  

////////////////////////////////////////////////////////////////////////////////////////

boolean
ODList_iterator::MoveTo( Comparable* c) {
  if (*ptr->data < *c) {
    while (*ptr->data < *c && ptr->next!=NULL) {
      ptr = ptr->next;
    }
  } else {
    while (*ptr->data > *c && ptr->prev!=NULL) {
      ptr = ptr->prev;
    }
  }
  return ( *ptr->data == *c);
}
