////////////////////////////////////////////////////////////////////////////
//
//  list.C : linked lists
//
//  $Header: /usr6/research/code/homebrew/RCS/olist.C,v 1.21 1998/02/24 22:41:46 crews Exp $
////////////////////////////////////////////////////////////////////////////

#include <iostream.h>
#include <stdlib.h>
#include "olist.h"
#include "heap.h"

////////////////////////////////////////////////////////////////////////////

OList::OList(OList& l) {
  type = l.type;
  if (count = l.count) {
    OLink *p, *q;
    head = p = new OLink( l.head->data, 0);
    for ( q = l.head->next; q; p = p->next, q = q->next)
      p->next = new OLink( q->data, 0);

  } else
    head = 0;
}

////////////////////////////////////////////////////////////////////////////

OList::OList( List& l, OList_Order t) {
   Link      *p;
  OLink      *q;

  type = t;
  if (count = l.count) {
    Heap *heap = new Heap( count);
    for ( p = l.head; p; p = p->next) {
      ASSERT( p->data->isComparable(), "Can't order uncomp items");
      heap->Add( (Comparable *) p->data);
    }

    q = NULL;
    while( heap->HeapSize()) {
      if ( type == Descend) {
	head = new OLink( heap->Pop(), head);
      } else {
	if (q) {
	  q->next = new OLink( heap->Pop(), 0);
	  q       = q->next;
	} else {
	  q = head = new OLink( heap->Pop(), head);
	}
      }
    }
    delete heap;
  } else
    head = 0;
}

////////////////////////////////////////////////////////////////////////////

void
OList::Insert(Comparable& p) {                    // List add
  OLink *q, *t;
  if (head == NULL) {
    head = new OLink(p, head);
  } else if ( (head->data->Compare( &p)*type) >= 0) {
    head = new OLink(p, head);
  } else {
    for ( q = head; q->next && (q->next->data->Compare( &p)*type) < 0; q = q->next);
    t = new OLink( p, q->next);
    q->next  = t;
  }
  ++count;
}

////////////////////////////////////////////////////////////////////////////

boolean
OList::InsertUnique(Comparable& p) {
  OLink *q, *t;
  if (!head) {
    head = new OLink(p, head);
  } else if ( (head->data->Compare(&p)*type) >= 0) {
    if ( *(head->data) == p)        // dont add or increment count
      return FALSE;
    head = new OLink(p, head);
  } else {
    for ( q = head; q->next && (q->next->data->Compare(&p)*type) <= 0; q = q->next) 
      if ( *(q->next->data) == p) 
	return FALSE;
    t = new OLink( p, q->next);
    q->next = t;
  }
  ++count;
  return TRUE;
}

////////////////////////////////////////////////////////////////////////////

// Top returns the head element
// Pop returns the head element and removes it from list

Comparable*  
OList::Top() {
  return (count)? head->data: 0;
}

Comparable*
OList::Pop() {
  if (!head) 
    FATAL ("pop from empty list");

  count--;
  Comparable *d = head->data;
  OLink *tmp = head;            
  head = head->next;           // advance head
  delete tmp;                  // destroy Link;
  return d;
}

Comparable*
OList::GetEquiv(Comparable *c) {
  integer diff;

  for (OLink *p = head; p != NULL; p=p->next) {
    diff = p->data->Compare( c)*type;

    if (diff == 0)        // found it!
      return p->data;
    if (diff  > 0)        // stop looking
      break;
  } 
  
  return NULL;
}

Comparable*
OList::GetElement (integer i) {
  if ( i<0 || i>=count) return NULL;

  OLink *p = head;
  for ( integer cnt = 0; cnt < i; cnt++) {
    p=p->next;
  } 
  return p->data;
}

////////////////////////////////////////////////////////////////////////////
// Deletes links in list, but retains objects

void
OList::Clear () {
  OLink *d;
  while ( head) {
    d    = head;
    head = d->next;
    delete d;
  }
  count = 0;
}

////////////////////////////////////////////////////////////////////////////
// Deletes links in list, and deletes objects

void
OList::CleanOut () {
  OLink *d;
  while ( head) {
    d    = head;
    head = d->next;
    delete d->data;
    delete d;
  }
  count = 0;
}

////////////////////////////////////////////////////////////////////////////
// Remove all matching objects in list. Return TRUE if any removed.

boolean
OList::RemoveAllof (Comparable *o) {

  OLink  *l, *p;
  boolean removed = FALSE;

  for (p = head, l = NULL; p; p=p->next) {
    integer diff = p->data->Compare( o)*type;
    
    if (diff < 0) {            // have not reached destination
      l = p;
    } else if ( diff == 0) {   // we found the area of interest           
      if (l==NULL) 
	head    = p->next;
      else 
	l->next = p->next;
      delete p;
      p = (l==NULL)? head:l;
      count--;
      removed = TRUE;
    } else {                   // passed the area of interest
      break;
    }
  } 
  return removed;
}

boolean
OList::Remove(Comparable *o) {                // just remove one instance

  OLink  *p,*l;

  for ( p=head, l=NULL; p; p = p->next) {
    
    if (*(p->data) == *o) {
      if ( l) 
	l->next = p->next;
      else 
	head = p->next;
      delete p;
      count--;
      return TRUE;
    }
    l = p;
  } 
  return FALSE;
}

////////////////////////////////////////////////////////////////////////////

OList*
OList::Diff( OList* remove) {
  OList* new_list = new OList( Order());

  if ( Order() != remove->Order()) {
    OList *temp = new OList( *remove);
    new_list = Diff( temp);
    delete temp;
  } else {
    OLink* p = head;
    OLink* q = remove->head;
    OLink* s = NULL;
  
    while ( p!=NULL && q!=NULL) {
      integer cmp = p->data->Compare( q->data)*type;
      if ( cmp < 0) {
	if ( s == NULL) {
	  s = new_list->head = new OLink( p->data, s);
	} else {
	  s->next = new OLink( p->data, s->next);
	  s       = s->next;
	}
	p = p->next;                  // let "this" move on and catch up
	new_list->count++;
      } else if (cmp > 0)  {
	q = q->next;                  // let rem catch up
      } else {
	p = p->next;                  // Neither are of interest
	q = q->next;
      }
    } 
    while ( p!=NULL) {                // throw remains on end of list
      if ( s == NULL) {
	s = new_list->head = new OLink( p->data, s);
      } else {
	s->next = new OLink( p->data, s->next);
	s       = s->next;
      }
      p = p->next;
      new_list->count++;
    }
  }
  return new_list;
}


////////////////////////////////////////////////////////////////////////////
// List printer

void
OList::Print( ostream& out) {
  if (count == 0) 
    out << "OList : empty\n";
  else {
    out << "OList : " << count << " item(s) ->\n";
    for (OLink *p=head; p; p=p->next)
      out << p->data;
  }
}

void
OList::ListPrint(ostream& out, char *delim) {
   for (OLink *p=head; p; p=p->next) {
      out <<  p->data << delim;
  }
}

////////////////////////////////////////////////////////////////////////////
// This version of merge merges toadd into the 
// olist without creating a new list. After Merge(), toadd will cease
// to exist.

void
OList::Merge(OList *toadd, boolean alloc_mem) { // Old version of OList, pre Feb. 24, 1994
  OLink *p, *q, *l = NULL;
  Comparable *data;

  if ( type != toadd->type) {
    OList *temp = new OList( *toadd);
    Transfer( temp);
    delete temp;
  } else {
  
    for ( p = head, q = toadd->head; p!=NULL && q!=NULL;) {
      if ( (p->data->Compare( q->data)*type) >= 0) {
	data = (alloc_mem)? (Comparable*) q->data->Copy(): q->data;
	if ( l) {
	  l->next = new OLink( data, p);
	  l       = l->next;
	} else {
	  head = new OLink( data, p);
	  l    = head;
	}
	q = q->next;
      } else  {
	l = p;
	p = p->next;
      }
    } 
    while ( q!=NULL) {                // throw remains on end of list
      data = (alloc_mem)? (Comparable*) q->data->Copy(): q->data;
      if (l) {
	l->next = new OLink( data, p);
	l = l->next;
      } else {
	head = new OLink( data, p);
	l    = head;
      }
      q = q->next;
    }
    count += toadd->count;
  }
}  


void
OList::Merge(List *toadd, boolean alloc_mem) {
  OLink      *p, *l = NULL;
  Link       *s;
  Comparable *c;

  if ( type == Descend) {   // Since heap is limited in direction
    Reverse();              // this the cheapest way I know -kgb
    Merge( toadd);
    Reverse();
  }else if ( toadd->count) {
    Heap *heap = new Heap( toadd->count);
    if ( alloc_mem) {
      for ( s = toadd->head; s; s = s->next) {
	ASSERT( s->data->isComparable(), "Can't order uncomp items");
	heap->Add( (Comparable*) s->data->Copy());
      }
    } else {
      for ( s = toadd->head; s; s = s->next) {
	ASSERT( s->data->isComparable(), "Can't order uncomp items");
	heap->Add( (Comparable*) s->data);
      }
    }
  
    for ( p = head, c = heap->Pop(); p && c;) {
      if ( *(p->data) >= *c) {
	if ( l) {
	  l->next = new OLink( c, p);
	  l       = l->next;
	} else {
	  head = new OLink( c, p);
	  l    = head;
	}
	c = (heap->HeapSize())? heap->Pop(): 0;
      } else  {
	l = p;
	p = p->next;
      }
    }
    while ( c) {                // throw remains on end of list
      if ( l) {
	l->next = new OLink( c, p);
	l       = l->next;
      } else {
	head = new OLink( c, p);
	l    = head;
      }
      c = (heap->HeapSize())? heap->Pop(): 0;
    }
    count += toadd->count;
  }
}  

////////////////////////////////////////////////////////////////////////////

void
OList::Transfer(OList *toadd) {
  OLink *p, *q, *t, *l = NULL;

  if ( type != toadd->type) 
    toadd->Reverse();
  
  if (!head) {                      // Adding to empty list
    head = toadd->head;
  } else {
    for ( p = head, q = toadd->head; p && q;) {
      if ( (p->data->Compare( q->data)*type) >= 0) {
	t       = q;
	q       = q->next;
	t->next = p;
	if ( l) {
	  l->next = t;
	  l       = l->next;
	} else {
	  head = t;
	  l    = head;
	}
      } else  {
	l = p;
	p = p->next;
      }
    } 
    if ( q) {                // throw remains on end of list
      l->next = q;
    }
  }
  count += toadd->count;
  toadd->head  = 0;
  toadd->count = 0;
}

////////////////////////////////////////////////////////////////////////////

void
OList::Union(OList *toadd, boolean alloc_mem) {
  OLink *p, *q, *l = NULL;
  Comparable* data;

  if ( type != toadd->type) {
    toadd->Reverse();
    Union( toadd);
    toadd->Reverse();
  } else {
  
    for ( p = head, q = toadd->head; p && q;) {
      if ( (p->data->Compare( q->data)*type) >= 0) {
	if ( *(p->data) != *(q->data)) {
	  count++;
	  data = (alloc_mem)? (Comparable*) q->data->Copy(): q->data;
	  if ( l) {
	    l->next = new OLink( data, p);
	    l       = l->next;
	  } else {
	    head = new OLink( data, p);
	    l    = head;
	  }
	} 
	q = q->next;
      } else {
	l = p;
	p = p->next;
      }
    } 
    while ( q) {                // throw remains on end of list
      count++;
      data = (alloc_mem)? (Comparable*) q->data->Copy(): q->data;
      if ( l) {
	l->next = new OLink( data, p);
	l       = l->next;
      } else {
	head = new OLink( data, p);
	l    = head;
      }
      q = q->next;
    }
  }
}  

////////////////////////////////////////////////////////////////////////////

List*
OList::RandomizedSublist( int n) {
  ASSERT( n>0 && n<=count, "Inappropriate sublist size");
  Object**  dataset = new Object*[count];

  integer c = 0;
  for (OLink* l = head; c<count; l = l->next) {
    dataset[c++] = l->data;
  }
  
  integer r;
  List*   sublist = new List;
  while ( 0 < n--) {
    r = IRANDOM(c);
    sublist->Add( dataset[r]);
    dataset[r] = dataset[--c];
  }
  delete dataset;
  return sublist;
}

OList*
OList::RandomSublist( int n) {
  ASSERT( n>0 && n<count, "Inappropriate sublist size");
  Comparable**  dataset = new Comparable*[count];

  integer c = 0;
  for (OLink* l = head; c<count; l = l->next) {
    dataset[c++] = l->data;
  }
  
  integer r;
  OList*  sublist = new OList(type);
  while ( 0 < n--) {
    r = IRANDOM(c);
    sublist->Insert( dataset[r]);
    dataset[r] = dataset[--c];
  }
  delete dataset;
  return sublist;
}

////////////////////////////////////////////////////////////////////////////

void
OList::Reverse() {
  type = (type == Ascend)? Descend: Ascend;
  OLink *t, *temp = head;
  head = 0;
  while ( temp) {
    t = temp->next;
    temp->next = head;
    head = temp;
    temp = t;
  }
}

////////////////////////////////////////////////////////////////////////////

integer
OList::Compare( Comparable* c) {
  ASSERT( c->isa() == OLIST, "Must compare to another OList");

  OList* list2 = (OList*) c;
  if ( Count() != list2->Count())
    return Count() - list2->Count();

  OLink* link1 = head;
  OLink* link2 = list2->head;
  while ( link1 != NULL) {
    if ( link1->data->Compare( link2->data) != 0)
      return link1->data->Compare( link2->data);
    link1 = link1->next;
    link2 = link2->next;
  }
  return 0;
}
