/* -*- C++ -*- */
#ifndef RBTREE_H
#define RBTREE_H

#include "object.h"
#include "list.h"
#include <string.h>

#define RED 1
#define BLACK 0

class RBNode : public Object {
  RBNode *right;
  RBNode *left;
  char color;
  integer hashval;
  char *key;
  Object *data;

  RBNode(int, char *, Object *);
  RBNode(RBNode *, boolean copy = TRUE);
  static RBNode *rbnull;

  char red() { return color; }

  friend class RBTree;
  friend class RBIterator;

  operator == ( RBNode);
  operator != ( RBNode);
  operator <  ( RBNode);
  operator <= ( RBNode);
  operator >  ( RBNode);
  operator >= ( RBNode);

 public:
  ~RBNode() {}
  virtual integer isa() { return RB_NODE; }
};

/**********************************************************************/

class RBTree : public Object {

  void RBDestroy(RBNode *);
  virtual void PrintNode(RBNode *, int, ostream& out = cout);
  void split(RBNode *);
  RBNode *rotate(RBNode *, RBNode *);
  integer hashf(char *);

  friend class RBIterator;

  RBNode *gg, *g, *p, *x;

protected:
  RBNode *root;
  integer count;

public:
  RBTree();
  RBTree(RBTree *, boolean copy = FALSE);
  ~RBTree() { if (root != 0) RBDestroy(root); }
  integer Count() { return count; }

  boolean Put (Object &o, char *key);
  boolean Put (Object *o, char *key) { return(Put(*o, key)); }
  void ForcePut (Object &o, char *key);
  void ForcePut (Object *o, char *key) { ForcePut(*o, key); }

  Object* Get (char *key);
  boolean Get (Object **o, char *key) { return ( *o = Get( key))!=0;}
  void Print (ostream &out = cout);
  integer isa() { return RB_TREE; }
};

inline ostream& operator << (ostream &out, RBTree& f)
{ f.Print(out); return out; }

inline ostream& operator << (ostream &out, RBTree* f)
{ out << "(ptr->)"; f->Print(out); return out; }

/**********************************************************************/

class RBIterator : public Object {
  
  List stack;
  int done;
  RBNode *n;

public:
  RBIterator(RBTree& t);
  ~RBIterator();

  void operator ++(); 

  int   end() {return done;}
  char *item(Object **d) {*d = n->data; return n->key; }
};

/**********************************************************************/


// Macros

#define RB_ForEach_Object(ht, k, o)  \
        { \
           Object *o; \
           char *k; \
           for ( RBIterator rb__itr(ht); !rb__itr.end(); ++rb__itr) { \
             k = rb__itr.item(&o);

#define RB_ForEach_Type(ht, t, k, o)  \
        { \
           t *o; \
           char *k; \
           for ( RBIterator rb__itr(ht); !rb__itr.end(); ++rb__itr) { \
             k = rb__itr.item((Object **)&o);

#define RB_End_ForEach }}

#endif // RBTREE_H

