#include "sticky.h"

///////////////////////////////////////////////////////
//
// /mo/homebrew_project/temp/dlink.C
//
// Methods of Sticky class        Michael Mo
//
///////////////////////////////////////////////////////

void 
Sticky :: Add(Sticky * E) {

// To add object E in front of object.

  ASSERT( E&& E!=this, "Fatal error: NULL input pointer or try to add to itself.");
  E->Remove();
  E->prev = prev;
  if (prev) prev->next = E;
  prev = E;
  E->next = this;

}

void
Sticky :: Append(Sticky * E) {

// To append object E behind object.

  ASSERT( E&& E!=this, "Fatal error: NULL input pointer or try to append itself.");
  E->Remove();
  E->prev = this;
  if (next) next->prev = E;
  E->next = next;
  next = E;

}

void
Sticky :: AddTo(Sticky * E) {

  // To add object in front of object E.
  ASSERT( E&& E!=this, "Fatal error: NULL input pointer or try to add to itself.");
  Remove();
  next = E;
  if ( E->prev ) E->prev->next = this;
  prev = E->prev; 
  E->prev = this;
 
} 
 
void
Sticky :: AppendTo(Sticky* E) {

  // To append object behind object E.
  ASSERT( E&& E!=this, "Fatal error: NULL input pointer or try to append to itself.");
  Remove();
  prev = E;
  next = E->next;
  if ( E->next ) E->next->prev = this;
  E->next = this;
  
}

void
Sticky :: AddList(Sticky * E) {

  // Add list E in front of object.
  ASSERT( E&& E!=this, "Fatal error: NULL input pointer or try to add own list."); 
  ASSERT( !E->Member(this), "Fatal error:Can't add a list which contains it.");
  E->RemoveList();
  if( prev ) prev->next = E;
  E->prev = prev;
  prev = E->EndOfList();
  E->EndOfList()->next = this;


}

void
Sticky :: AppendList(Sticky * E) {

  // Append list E behind object.
  ASSERT( E&& E!=this, "Fatal error: NULL input pointer or try to append own list."); 
  ASSERT( !E->Member(this), "Fatal error:Can't append a list which contains it.");
  E->RemoveList();
  E->prev = this;
  if ( next ) next->prev = E->EndOfList();
  E->EndOfList()->next = next;
  next = E;

}

void
Sticky :: AddListTo(Sticky * E) {
 
  // Add object list in front of object E.
  ASSERT( E&& E!=this, "Fatal error: NULL input pointer or try to add to own list."); 
  ASSERT( !Member(E), "Fatal error: Can't add to a list that it contains.");
  RemoveList();
  if ( E->prev ) E->prev->next = this;
  prev = E->prev;
  E->prev = EndOfList();
  EndOfList()->next = E;

}

void
Sticky :: AppendListTo(Sticky * E) {

  // Append object list behind object E.
  ASSERT( E&& E!=this, "Fatal error: NULL input pointer or try to append to own list."); 
  ASSERT( !Member(E), "Fatal error: Can't append to a list that it contains.");
  RemoveList();
  prev = E;
  if ( E->next ) E->next->prev = EndOfList();
  EndOfList()->next = E->next;
  E->next = this;

}

void
Sticky :: PrintList( ostream& out ) {

  Print();
  if ( next)
     next->PrintList(out);
}




void Sticky :: Remove() {
  
  // Remove the object from link.
  
  if ( prev ) prev->next = next;
  if ( next ) next->prev = prev;
  prev = next = NULL;
  
}

void
Sticky :: RemoveList() {

  // Removes object list from link.

  if( prev ) prev->next = NULL;
  prev = NULL;

}


Sticky*
Sticky :: EndOfList() {

  // Returns the end of list E.

  Sticky* end = this;
  for ( ; end->next ; ) end = end->next;
  return(end);
  
}

boolean
Sticky :: Member(Sticky * E) {

  ASSERT( E, "Fatal Error: Null is not a member of list.");
  for( Sticky*end = this; end->next; ) {
    if ( end == E ) return(TRUE);
    end = end->next;
  }
  return(FALSE);
}






	
	
