//  $Header: /usr6/research/code/homebrew/RCS/strobj.C,v 1.19 1997/06/10 06:03:04 ihkhan Exp $

#include "strobj.h"

//////////////////////////////
//constructors and destructors
//////////////////////////////
StringObj::StringObj() {
  length = 0;
  str    = new char[length+1];
  str[0] = '\0';
}

StringObj::StringObj( StringObj& copy) {
  length = copy.length;
  str    = new char[length+1];
  for (int i=0; i < length+1; i++)
    str[i] = copy.str[i];
}

StringObj::StringObj( StringObj* copy) {
  length = copy->length;
  str    = new char[length+1];
  for (int i=0; i < length+1; i++)
    str[i] = copy->str[i];
}

StringObj::StringObj( char *copy, boolean mem) {
  length = strlen( copy);
  if (mem) {
    str    = new char[length+1];
    for (int i=0; i < length+1; i++)
      str[i] = copy[i];
  } else
    str = copy;
}

StringObj::~StringObj( void) { 
  if ( str) delete str;
};

//////////////////////////////////

unsigned
StringObj::Hash (unsigned h) {  
  unsigned shift_out;
  for (char *s = str; *s != '\0'; ++s) {
    h += (int)*s << 11;
    shift_out = h&0x01;
    h >>= 1;
    if (shift_out)
      h ^= 0x63a;
  }
  h &= 0x0fff;
  return (unsigned) h%1013;
}

char 
StringObj::operator[]( int offset) { 
  return (offset >= 0 && offset < length)? str[offset]: '\0';
}


/////////////////////////////////////////////
// Compare() like typical string comparison 
// return == 0: matched
//         < 0: local < test (alphabetically)
//         > 0: local > test (alphabetically)
/////////////////////////////////////////////
int
StringObj::Compare( char *test) {
  return (strcmp( str, test));
}

int
StringObj::Compare( Comparable *test) {
  ASSERT ( test->isa() == STRING_OBJ, "must compare stringObj with stringObj");
  return (strcmp( str, ((StringObj*) test)->str));
}


///////////////////////////////////////
// case insensitive string comparitors
///////////////////////////////////////
int       // loacl <?> test -> return <> 0
StringObj::CIcompare( char *test) {
  int  weight = 0;
  
  int pos = 0;
  for (; pos < length+1; pos++) {
    weight = str[pos] - test[pos];
    if ((weight == -0x20) && (str[pos]>0x40) && (str[pos]<0x5b)) continue;
    if ((weight ==  0x20) && (str[pos]>0x60) && (str[pos]<0x7b)) continue;

    if (weight != 0) return weight;
    if ( test[pos] == '\0') break;
  }
  if (test[pos] != '\0') return -1;
  return length - pos;
}

int       // loacl <?> test -> return <> 0
StringObj::CIcompare( StringObj& test) {
  int  weight = 0;
  
  for (int pos = 0; pos < length; pos++) {
    weight = str[pos] - test.str[pos];
    if ((weight == -0x20) && (str[pos]>0x40) && (str[pos]<0x5b)) continue;
    if ((weight ==  0x20) && (str[pos]>0x60) && (str[pos]<0x7b)) continue;

    if (weight != 0) return weight;
    if ( pos > test.length) break;
  }
  return length - test.length;
}


///////////////////////////////////////
// test or locate sub strings
///////////////////////////////////////
boolean      // TRUE: test string match from st_pos (>=0) or to st_pos from end (<0)
StringObj::MatchAt( char *test, int st_pos) {
  
  if (st_pos>=0) {
    if (st_pos >= length) return FALSE;

    for (int pos = 0; test[pos] != '\0'; pos++) {
      if (str[st_pos + pos] != test[pos]) return FALSE;
      if (str[st_pos + pos] == '\0' && test[pos] != '\0') return FALSE;
    }
  } else {
    int test_lng = strlen( test);
    if (test_lng > length) return FALSE;
    if ( -1*st_pos > length) return FALSE;

    for (int pos = 0; pos < test_lng; pos++) {
      if (str[st_pos + length -pos] != test[test_lng -1 -pos]) return FALSE;
    }
  }   
  return TRUE;
}

boolean      // TRUE: test string match from st_pos (>=0) or to st_pos from end (<0)
StringObj::MatchAt( StringObj& test, int st_pos) {

  if (st_pos>=0) {
    if (st_pos >= length) return FALSE;
    if (test.length > length) return FALSE;

    for (int pos = 0; test.length; pos++) {
      if (str[st_pos + pos] != test.str[pos]) return FALSE;
    }
  } else { 
    if (test.length > length) return FALSE;
    if ( -1*st_pos > length) return FALSE;

    for (int pos = 0; test.length; pos++) {
      if (str[st_pos + length -pos] != test.str[test.length -1 -pos]) return FALSE;
    }
  }  
  return TRUE;
}

int 
StringObj::Locate( StringObj &sub, boolean from_beg) {
  int match;
  if (sub.length > length) return -1;

  if (from_beg) {
    match = 0;
    int i=0;
    for (; i < length && match < sub.length; i++) {
      if ( str[i] != sub.str[match++])
	match = 0;
    }
    if (match < sub.length) return -1;
    return i - match;
  } else {
    match = 1;
    int i=1;
    for (; i <= length && match <= sub.length; i++) {
      if ( str[length-i] != sub.str[sub.length-match++])
	match = 1;
    }
    if (match < sub.length) return -1;
    return length - i +1;
  }
}   
 
int 
StringObj::Locate( char *sub, boolean from_beg) {
  int match;
  int test_lng = strlen(sub);
  if (test_lng > length) return -1;

  if (from_beg) {
    match = 0;
    int i=0;
    for (; i < length && match < test_lng; i++) {
      if ( str[i] != sub[match++])
	match = 0;
    }
    if (match < test_lng) return -1;
    return i - match;
  } else {
    match = 1;
    int i=1;
    for (; i <= length && match <= test_lng; i++) {
      if ( str[length-i] != sub[test_lng-match++])
	match = 1;
    }
    if (match < test_lng) return -1;
    return length - i +1;
  }
}    
  

//////////////////////////////////////////
// concat() has robust concatination will allocate req space.
//////////////////////////////////////////

StringObj  
StringObj::operator+( const StringObj& add) { 
  StringObj temp; 
  delete temp.str;
  temp.length = length+add.length;
  temp.str    = new char[ temp.length+1];

  int indx = 0;
  for (; indx < length; indx++)
    temp.str[indx] = str[indx];
  for (indx=0; indx <= add.length; indx++)
    temp.str[length+indx] = add.str[indx];
  return temp;
}

StringObj&  
StringObj::Concat( const StringObj& add) {
  int  indx;
  SetLength (length + add.length);  
  
  for ( indx=0; indx <= add.length; indx++)
    str[ length-indx] = add.str[ add.length-indx];
  return *this;
}


StringObj  
StringObj::operator+( char add) { 
  StringObj temp; 
  delete temp.str;
  temp.length = length+1;
  temp.str    = new char[ temp.length+1];


  for (int indx = 0; indx < length; indx++)
    temp.str[indx] = str[indx];
  
  temp.str[length] = add;
  temp.str[length+1] = '\0';

  return temp;
}

StringObj&  
StringObj::Concat( char add) {
  SetLength (length + 1);  
  str[ length-1] = add;
  return *this;
}


StringObj  
StringObj::operator+( const char *add) { 
  StringObj temp; 
  int       add_length = strlen( add);
  delete temp.str;
  temp.length = length+add_length;
  temp.str    = new char[ temp.length+1];

  int indx = 0;
  for (; indx < length; indx++)
    temp.str[indx] = str[indx];
  for (indx = 0; indx <= add_length; indx++)
    temp.str[length+indx] = add[indx];
  return temp;
}

StringObj&  
StringObj::Concat( const char *add) {
  int add_len = strlen( add);
  SetLength (length + add_len);  

  for ( int indx=0; indx <= add_len; indx++) 
    str[ length-indx] = add[add_len-indx];
  return *this;
}


StringObj  
StringObj::operator+( int add) { 
  char buf[MAXSTRING];
  ostrstream( buf, MAXSTRING) << add << ends;
  return *this + buf;
}

StringObj&      
StringObj::ConcatInt( int int_str, int min_lng) {
  char buf[MAXSTRING];
  ostrstream( buf, MAXSTRING) << int_str << ends;

  int int_lng = strlen( buf);
  SetLength( length + MAX( min_lng, int_lng));  
  
  int pos = length-min_lng;
  for (; pos < length-int_lng; pos++) 
    str[pos] = '0';
  for (    pos = length-int_lng; pos <= length; pos++) 
    str[pos] = buf[pos-length+int_lng];
  return *this;
}

StringObj  
StringObj::operator+( long add) { 
  char buf[MAXSTRING];
  ostrstream( buf, MAXSTRING) << add << ends;
  return *this + buf;
}


StringObj&      
StringObj::ConcatLong( long long_str, int min_lng) {
  char buf[MAXSTRING];
  ostrstream( buf, MAXSTRING) << long_str << ends;

  int long_lng = strlen( buf);
  SetLength( length + MAX( min_lng, long_lng));  
  
  int pos = length-min_lng;
  for (; pos < length-(int)long_lng; pos++) 
    str[pos] = '0';
  for (    pos = length-(int)long_lng; pos <= length; pos++) 
    str[pos] = buf[pos-length+(int)long_lng];
  return *this;
}


////////////////////////////////
// typical copy functs
////////////////////////////////

StringObj&
StringObj::operator=( const char *copy) {
  int copy_len;

  if ( (copy_len = strlen( copy)) != length) {
    if (str) delete str;
    length = copy_len;
    str    = new char[ length+1];
  }
  for ( int indx=0; indx<=length; indx++)
    str[indx] = copy[indx];
  return *this;
}

StringObj&
StringObj::operator=( const StringObj& copy) {

  if ( copy.length != length) {
    if (str) delete str;
    length = copy.length;
    str    = new char[ length+1];
  }
  for ( int indx=0; indx<=length; indx++)
    str[indx] = copy.str[indx];
  return *this;
}

///////////////////////////////////////////// 
// Copy functs for extracting substrings
// Note: starting element (starting from 0) and ending element 
// (set stop=-1 for copy to eol).  
//////////////////////////////////////////////////

void
StringObj::Copy( char *copy, int start, int stop) {
  int copy_lng = strlen( copy);
  start = MAX( 0, start);
  if (stop == -1) stop = copy_lng;
  else            stop = MIN( copy_lng, stop);
  SetLength( stop-start);

  int pos = start;
  for (; pos < stop; pos++)
    str[ pos-start] = copy[ pos];
  str[pos-start] = '\0';
}

void
StringObj::Copy( StringObj& copy, int start, int stop) {
  start = MAX( 0, start);
  if (stop == -1) stop = copy.length;
  else            stop = MIN( copy.length, stop);
  SetLength( stop-start);

  int pos = start;
  for (; pos < stop; pos++)
    str[ pos-start] = copy.str[ pos];
  str[ pos-start] = '\0';
}

StringObj
StringObj::Copy( int start, int stop) {
  StringObj new_str;

  start = MAX( 0, start);
  if (stop == -1) stop = length;
  else            stop = MIN( length, stop);
  new_str.SetLength( stop-start);

  int pos = start;
  for (; pos < stop; pos++)
    new_str.str[ pos-start] = str[ pos];
  new_str.str[ pos-start] = '\0';
  return new_str;
}

/////////////////////////////////////////////
// I/O functions
/////////////////////////////////////////////

void
StringObj::Scan( istream& in) {
  char *buf     = new char[ MAXSTRING];
  buf[0]        = '\0';

  in.width( MAXSTRING);
  in >> buf;
  Copy( buf);
}

void
StringObj::Getline( istream& in, char delim) {
  char *buf     = new char[ MAXSTRING];
  buf[0]        = '\0';

  in.getline( buf, MAXSTRING, delim);
  Copy( buf);
} // Getline()


