#ifndef STRINGOBJ_H
#define STRINGOBJ_H

//  $Header: /usr6/research/code/homebrew/RCS/strobj.h,v 1.21 1997/12/07 22:30:11 crews Exp $

#include <strstream.h>
#include <string.h>
#include "intobj.h"
#include "compable.h" 
#include "stddefs.h"

#define MAXSTRING 256

class StringObj: public Comparable {
private:
  char *str;
  int   length;
  
  void  SetLength( int l) { 
    int indx;

    if (l != length) { char *temp = str;
		       str  = new char[l+1];
		       
		       for ( indx=0; indx<length && indx<l; indx++)
			 str[indx] = temp[indx];
		       str[indx] = '\0';
		       length = l;
		       if (temp) delete temp;}
  }

public:
  StringObj();
  StringObj( StringObj &copy);
  StringObj( StringObj *copy);
  StringObj( char      *copy, boolean mem = TRUE);
 ~StringObj();

  virtual integer isa() { return STRING_OBJ;}
  virtual Object* Copy(){ return new StringObj( this);}   
  virtual unsigned Hash(unsigned i=0);

  // to view the private information
  char    operator[]( int offset);
  operator char*() { return str;}
  char*   String() { return str;}
  boolean Null()   { return (length == 0);}
  int     Length() { return length;}

  // regular string Compare() and a Case Insensitve CIcompare() 
  int     Compare( char       *test);
  integer Compare( Comparable *test);
  integer Compare( Comparable &test) { return Compare( &test);}

  boolean operator< ( char *test) { return Compare( test) >= 0;}
  boolean operator<=( char *test) { return Compare( test) >  0;}
  boolean operator==( char *test) { return Compare( test) == 0;}
  boolean operator!=( char *test) { return Compare( test) != 0;}
  boolean operator> ( char *test) { return Compare( test) <= 0;}
  boolean operator>=( char *test) { return Compare( test) <  0;}
  

  int     CIcompare( char      *test);
  int     CIcompare( StringObj &test);

  // check or locate sub-strings

  boolean MatchAt( StringObj& test, int st_pos = 0);
  boolean MatchAt( char*      test, int st_pos = 0);
  int     Locate( StringObj& sub, boolean from_beg = TRUE);
  int     Locate( char*      sub, boolean from_beg = TRUE);

  // concat class 
  // With pointers, const should be used here (avoids anachronism warnings)
  StringObj& Concat( char       add);
  StringObj& Concat( const char      *add);
  StringObj& Concat( const StringObj &add);
  StringObj& ConcatInt( int add, int min_len = -1);   // has zeros inserted to create proper lng
  StringObj& ConcatLong( long add, int min_len = -1);   // has zeros inserted to create proper lng

  StringObj  operator+( char		 add);
  StringObj  operator+( const char	*add);
  StringObj  operator+( const StringObj &add);
  StringObj  operator+( int		 add);
  StringObj  operator+( long		 add);

  // Standard and more elaborate copy functs
  StringObj& operator=( const char      *src);
  StringObj& operator=( const StringObj &src);
  void       Copy( char      *copy, int start = 0, int stop = -1);
  void       Copy( StringObj &copy, int start = 0, int stop = -1);
  StringObj  Copy( int start = 0, int stop = -1);

  // danger function: Hash and Compare will work without length set but not much else
  void       Set( char *s ) { str = s; length = 0; }

  // obvious right? 
  virtual void Print( ostream& out = cout) { if (str) out << str;}
  virtual void Scan(  istream&  in = cin );
  virtual void Getline( istream& in, char delim = '\n');
  virtual void Write( ostream &out ) {
    IntObj::WriteInt( out, length+1);
    out.write( str, length+1 );
    }
  static void WriteString( ostream &out, char *s ) {
    int l=strlen(s)+1;
    IntObj::WriteInt( out, l );
    out.write( s, l );
    }
  virtual void Read( istream &in ) {
    length = IntObj::ReadInt( in );
    str = new char[length];
    in.read( str, length-- );
    }
  static char* ReadString( istream &in ) {
    int l=IntObj::ReadInt( in );
    char *s = new char[l];
    in.read( s, l );
    return s;
    }
};

inline boolean operator< ( char *test, StringObj& str)
{ return str.Compare( test) >= 0;}
inline boolean operator<=( char *test, StringObj& str)
{ return str.Compare( test) >  0;}
inline boolean operator==( char *test, StringObj& str)
{ return str.Compare( test) == 0;}
inline boolean operator!=( char *test, StringObj& str)
{ return str.Compare( test) != 0;}
inline boolean operator> ( char *test, StringObj& str)
{ return str.Compare( test) <= 0;}
inline boolean operator>=( char *test, StringObj& str)
{ return str.Compare( test) <  0;}

// Make StrObj's act like streams, for Hien...
inline StringObj operator << (StringObj &s, char add)
{ return (s + add); }
inline StringObj operator << (StringObj &s, const char *add)
{ return (s + add); }
inline StringObj operator << (StringObj &s, const StringObj& add)
{ return (s + add); }
inline StringObj operator << (StringObj &s, int add)
{ return (s + add); }

#endif // STRINGOBJ_H

