#include "bddsh.h"

///////////////////////////////////////////////////////////
//
// Tcl specifics:
// main must call the appropriate shell main defined above
//

extern boolean useTK;

int
main(int argc,char **argv)
{
    int i;
    for( i=1 ; i<argc ; i++ )
     if( strcmp( argv[i], "-tk" )==0 )
       useTK=TRUE;

    if( useTK )
      Tk_Main(argc, argv, Tcl_AppInit);
    else
      Tcl_Main(argc, argv, Tcl_AppInit);

    return 0;
  }

///////////////////////////////////////////////////////////
//
// Tcl_AppInit is called at startup
// it is used to allocate new commands to the shell

int Tcl_AppInit(Tcl_Interp *interp) {
  if (Tcl_Init(interp) == TCL_ERROR) {
    return TCL_ERROR;
    }
  if( useTK ) {
    if (Tk_Init(interp) == TCL_ERROR) {
      return TCL_ERROR;
      }
    }

  Tcl_SetVar(interp, "tcl_rcFileName", "./.bddshrc", TCL_GLOBAL_ONLY);
  Tcl_Eval( interp, "set auto_noload 1" );	// autoload is screwed up

  BDDmanager *bdd = new BDDmanager(17,0);
//////////////////////////////////////////////////////
// now create commands for using bdd's

  // core functions
  Tcl_CreateCommand(interp, "input", BDD_InputCmd, (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "and",   BDD_AndCmd,   (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "&",     BDD_AndCmd,   (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "or",    BDD_OrCmd,    (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "|",     BDD_OrCmd,    (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "nor",   BDD_NorCmd,   (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "not",   BDD_NotCmd,   (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "!",     BDD_NotCmd,   (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "xor",   BDD_XorCmd,   (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "^",     BDD_XorCmd,   (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "thenptr", BDD_ThenCmd,  (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "elseptr", BDD_ElseCmd,  (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "read",    BDD_ReadCmd,  (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "exactly", BDD_TupleCmd, (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "tuple",   BDD_TupleCmd, (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "upto",    BDD_TupleCmd, (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "atleast", BDD_TupleCmd, (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "smooth",  BDD_SmoothCmd, (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "uabstract", BDD_UAbstractCmd, (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "pickmt",  BDD_PickMTCmd, (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "cofactor",  BDD_CofactorCmd, (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "restrict",  BDD_RestrictCmd, (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "distinguish", BDD_DistCmd, (ClientData) bdd, NULL);
#ifndef NO_MEM
  Tcl_CreateCommand(interp, "memstats", BDD_MemStatsCmd, (ClientData) bdd, NULL);
#endif
  Tcl_CreateCommand(interp, "bddhelp", BDD_HelpCmd, (ClientData) bdd, NULL);

  // functions that return non bddnode info
  Tcl_CreateCommand(interp, "index",   BDD_GetIndexCmd, (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "level",   BDD_GetLevelCmd, (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "numnodes",BDD_NumNodesCmd, (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "printcubes",BDD_PrintCubesCmd, (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "numcubes",BDD_NumCubesCmd, (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "mincube", BDD_MinCubeCmd, (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "setsize", BDD_SetSizeCmd, (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "bddstats",BDD_StatsCmd, (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "show",    BDD_ShowCmd,     (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "printorder",BDD_PrintOrderCmd,     (ClientData) bdd, NULL);

  // Garbage collection functions
  Tcl_CreateCommand(interp, "keep",     BDD_KeepCmd, (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "unkeep", BDD_UnKeepCmd, (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "gc",         BDD_GCCmd, (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "sift",     BDD_SiftCmd, (ClientData) bdd, NULL);

  Tcl_CreateCommand(interp, "readeqn",   BDD_ReadEqn,  (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "readblif",  BDD_ReadBLIFCmd, (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "readpla",   BDD_ReadPLACmd,(ClientData) bdd, NULL);

  Tcl_CreateCommand(interp, "writeblif", BDD_WriteBLIFCmd, (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "save",     BDD_SaveCmd, (ClientData) bdd, NULL);
  Tcl_CreateCommand(interp, "load",     BDD_LoadCmd, (ClientData) bdd, NULL);

  Tcl_CreateCommand(interp, "drawhgraph",HG_DrawHGraphCmd, (ClientData) NULL, NULL);


  // Report the values of Zero and One
  cout << "BDD info: " << endl;
  cout << "          one is " << (int) bdd->One() << endl;
  cout << "         zero is " <<(int)bdd->Zero() << endl;
  cout << "[enter bddhelp for help on bdd commands]" << endl;
  cout << flush;

  // Set tcl variables "one" and "zero"
  char one[10], zero[10];
  sprintf(one, "%d", (int)bdd->One() );
  sprintf(zero, "%d",(int)bdd->Zero() );
  Tcl_SetVar2(interp, "one", NULL, one,   0);
  Tcl_SetVar2(interp, "zero",NULL, zero,  0);

  // Link tcl variables to the actual bddmanager variables
  // and make them read only
  Tcl_LinkVar(interp, "num_nodes", (char*) &(bdd->num_nodes),
		TCL_LINK_INT|TCL_LINK_READ_ONLY);
  Tcl_LinkVar(interp, "num_inputs", (char*) &(bdd->num_inputs),
		TCL_LINK_INT|TCL_LINK_READ_ONLY);
  Tcl_LinkVar(interp, "num_ite", (char*) &(bdd->num_ite_calls),
		TCL_LINK_INT|TCL_LINK_READ_ONLY);

#ifndef NO_MEM
  Tcl_MMLinkVars(interp);
#endif
  return TCL_OK;
  }

//
// end of tcl/tk specific routines
/////////////////////////////////////////////////////////////////

