#ifndef BDDSH_H
#define BDDSH_H

#include "tcl.h"
#include "tk.h"
#include <fstream.h>
#include <string.h>
#include <unistd.h>
#include <iomanip.h>


#include "bdd.h"
#include "hash.h"
#include "strobj.h"
#include "blifdata.h"
#include "hgraph.h"

// default to tcl interface, unless TK defined during the compile

#ifndef TK
#define TCL_MAIN Tcl_Main
#else
#define TCL_MAIN Tk_Main
#endif

extern const char* BDDSH_ERROR_MESSAGE[];

int BDD_ReadPLACmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_ReadEqn(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_ReadBLIFCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_WriteBLIFCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_HelpCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
#ifndef NO_MEM
int BDD_MemStatsCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
#endif
int HG_DrawHGraphCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);

int BDD_NumCubesCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_MinCubeCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_ShowCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_PrintCubesCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_PrintMinCubesCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_SetSizeCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);

int BDD_NumNodesCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_StatsCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_TupleCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_SmoothCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_UAbstractCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_PickMTCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_DistCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_CofactorCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_RestrictCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);

int BDD_ReadCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_InputCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_AndCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_OrCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_XorCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_NorCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_NotCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);

int BDD_GetIndexCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_GetLevelCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_ElseCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_ThenCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);

int BDD_KeepCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_UnKeepCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_GCCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_SiftCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);

int BDD_SaveCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_LoadCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
int BDD_PrintOrderCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);

#ifndef NO_MEM

/*void Tk_MemoryManagerInit( char *argv[] ); */
int Tk_MMInit(Tcl_Interp *interp);
int Tcl_MMLinkVars(Tcl_Interp *interp);
int Tk_MMInternal(Tcl_Interp *interp);

int BDD_UpdateMemGCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[]);
#endif

// error stuff //

#define BDDSH_ERROR_IF( test, err_num, args ) \
      if( test ) { \
        sprintf(interp->result, BDDSH_ERROR_MESSAGE[err_num], args); \
        return TCL_ERROR; \
        }

#endif BDDSH_H
