#include "bddsh.h"

const char *BDDSH_ERROR_MESSAGE[] = {
/* 0 */ "Wrong args: should be: %s",
/* 1 */ "%s is not a valid bdd",
/* 2 */ "BDD index must be positive.",
/* 3 */ "Need to supply a filename.",
/* 4 */ "Can't open file %s" };

static char interp_result[10000];
boolean useTK;
/////////////////////////////////////////////////////////////////
/* Synopsis: reads function (EQN format), assigns bdd vars, returns bdd */

int BDD_ReadEqn(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
{
  BDDnode *f;
  BDDmanager *bdd = (BDDmanager*)clientdata;
  static HashTree *htname = new HashTree;
  static HashTree *htindex = new HashTree;

  BDDSH_ERROR_IF( argc!=2, 0, "eqn \"function\"" );
  f = bdd->ReadEqn( htname, htindex, argv[1] );
  sprintf( interp->result, "%d", (int) f );
  return( TCL_OK );
}

/////////////////////////////////////////////////////////////////
/* Synopsis: returns a bdd variable with the specified index */

int BDD_InputCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc>2, 0, "input [indexnum]" );

  int index;
  if( argc == 2 ) {
    index = atoi( argv[1] );
    BDDSH_ERROR_IF( index < 0, 2, NULL );
    }
  else
    index = bdd->NumInputs();
  BDDnode *result = bdd->Input(index);

  sprintf(interp->result, "%d", (int) result);
  return TCL_OK;
  }

/////////////////////////////////////////////////////////////////
/* Synopsis: Read in a set of cubes and return the bdd */

int BDD_ReadCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc>2, 0, "read { cubes }" );

  char *cubes = new char[strlen(argv[1])+4];
  sprintf(cubes,"{\n%s\n}", argv[1]);
  BDDnode *result = bdd->ReadEq(cubes);

  sprintf(interp->result, "%d", (int) result);
  delete cubes;
  return TCL_OK;
  }


/////////////////////////////////////////////////////////////////
/* Synopsis: Compute the sets of variables that distinguish 2 BDDs */

int BDD_DistCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc!=5, 0, "distinguish bdd bdd max vars" );
  int max = atoi(argv[3]);

  BDDnode *vars = (BDDnode*) atoi(argv[4]);
  BDDSH_ERROR_IF( vars==NULL, 1, argv[4] );
  IndexArray ia( bdd, vars );

  BDDnode *s1 = (BDDnode*) atoi(argv[1]);
  BDDSH_ERROR_IF( s1==NULL, 1, argv[1] );
  BDDnode *s2 = (BDDnode*) atoi(argv[2]);
  BDDSH_ERROR_IF( s2==NULL, 1, argv[2] );

  BDDnode *result = bdd->Distinguish( s1, s2, max, &ia, NULL);

  sprintf(interp->result, "%d", (int) result);
  return TCL_OK;
  }

/////////////////////////////////////////////////////////////////
/* Synopsis: Perform Boolean and of 2 or more BDDs */

int BDD_AndCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc<3, 0, "and bdd bdd [bdd...]" );

  BDDnode *result = bdd->One();
  for( int num=1 ; num<argc ; num++ ) {
    BDDnode *next = (BDDnode*) atoi(argv[num]);
    BDDSH_ERROR_IF( next==NULL, 1, argv[num] );
    result = bdd->and(result, next);
    }
  sprintf(interp->result, "%d", (int) result);
  return TCL_OK;
  }


/////////////////////////////////////////////////////////////////
/* Synopsis: Perform Boolean or of 2 or more BDDs */

int BDD_OrCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc<3, 0, "or bdd bdd [bdd...]" );

  BDDnode *result = bdd->Zero();
  for( int num=1 ; num<argc ; num++ ) {
    BDDnode *next = (BDDnode*) atoi(argv[num]);
    BDDSH_ERROR_IF( next==NULL, 1, argv[num] );
    result = bdd->or(result, next);
    }
  sprintf(interp->result, "%d", (int) result);
  return TCL_OK;
  }


/////////////////////////////////////////////////////////////////
/* Synopsis: Perform Boolean xor of 2 or more BDDs */

int BDD_XorCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc<3, 0, "xor bdd bdd [bdd...]" );

  BDDnode *result = bdd->Zero();
  for( int num=1 ; num<argc ; num++ ) {
    BDDnode *next = (BDDnode*) atoi(argv[num]);
    BDDSH_ERROR_IF( next==NULL, 1, argv[num] );

    result = bdd->xor(result, next);
    }

  sprintf(interp->result, "%d", (int) result);
  return TCL_OK;
  }


/////////////////////////////////////////////////////////////////
/* Synopsis: Perform Boolean nor of 2 or more BDDs */

int BDD_NorCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc<3, 0, "nor bdd bdd [bdd...]" );

  BDDnode *result = bdd->Zero();
  for( int num=1 ; num<argc ; num++ ) {
    BDDnode *next = (BDDnode*) atoi(argv[num]);
    BDDSH_ERROR_IF( next==NULL, 1, argv[num] );
    result = bdd->or(result, next);
    }
  result = bdd->not(result);

  sprintf(interp->result, "%d", (int) result);
  return TCL_OK;
  }

/////////////////////////////////////////////////////////////////
/* Synopsis: Perform generalized cofactor of two BDD's */

int BDD_CofactorCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc!=3, 0, "cofactor bdd bdd" );

  BDDnode *s1 = (BDDnode*) atoi(argv[1]);
  BDDSH_ERROR_IF( s1==NULL, 1, argv[1] );
  BDDnode *s2 = (BDDnode*) atoi(argv[2]);
  BDDSH_ERROR_IF( s2==NULL, 1, argv[2] );

  BDDnode *result = bdd->Cofactor( s1, s2 );

  sprintf(interp->result, "%d", (int) result);
  return TCL_OK;
  }

/////////////////////////////////////////////////////////////////
/* Synopsis: Restrict BDD 1 by BDD 2 */

int BDD_RestrictCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc!=3, 0, "restrict bdd bdd" );

  BDDnode *s1 = (BDDnode*) atoi(argv[1]);
  BDDSH_ERROR_IF( s1==NULL, 1, argv[1] );
  BDDnode *s2 = (BDDnode*) atoi(argv[2]);
  BDDSH_ERROR_IF( s2==NULL, 1, argv[2] );

  BDDnode *result = bdd->Restrict( s1, s2 );

  sprintf(interp->result, "%d", (int) result);
  return TCL_OK;
  }

/////////////////////////////////////////////////////////////////
/* Synopsis: BDD with up to, exactly, or at least num_on vars set to '1' */

int BDD_TupleCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( (argc!=3 && argc!=4) , 0, "tuplecmd num_on (#\"height\" | bddvars) pattern-bdd" );

  IndexArray *ia;
  if( argv[2][0] == '#' ) {
    int height = atoi( &argv[2][1] )-1;
    BDDSH_ERROR_IF( height<1, 2, NULL );
    ia = new IndexArray( 0, height );
    }
  else {
    BDDnode *vars = (BDDnode*)atoi(argv[2]);
    BDDSH_ERROR_IF( vars==NULL, 1, argv[2] );
    ia = new IndexArray(bdd, vars );
    }

  BDDnode *result;
  int n = atoi(argv[1]);
  BDDSH_ERROR_IF( n<0, 2, NULL );

  BDDnode *p = NULL;

  if (argc==4) {
      p = (BDDnode*)atoi(argv[3]);
      BDDSH_ERROR_IF( p==NULL, 1, argv[3] );
  }

  if( strcmp( argv[0], "tuple") ==0 || strcmp( argv[0], "exactly")==0 )
      result =  (argc==4) ? bdd->Tuple( n, ia, p ) : bdd->Tuple( n, ia );
  else if( strcmp( argv[0], "upto") ==0 )
      result =  (argc==4) ? bdd->UpToN( n, ia, p ) : bdd->UpToN( n, ia );
  else if( strcmp( argv[0], "atleast") ==0 )
      result =  (argc==4) ? bdd->AtLeastN( n, ia, p ) : bdd->AtLeastN( n, ia );

  sprintf(interp->result, "%d", (int) result);

  delete ia;
  return TCL_OK;
  }

/////////////////////////////////////////////////////////////////
/* Synopsis: Smooth bddvars from the BDD (bddvars is a BDD cube) */

 
int BDD_SmoothCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
 {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc!=3, 0, "smooth bdd bddvars" );

  BDDnode *v = (BDDnode*) atoi(argv[2]);
  BDDSH_ERROR_IF( v==NULL, 1, argv[2] );

  IndexArray ia(bdd, v);
  BDDnode *f = (BDDnode*) atoi(argv[1]);
  BDDSH_ERROR_IF( f==NULL, 1, argv[1] );

  BDDnode *result = bdd->Smooth( f, &ia );
  sprintf(interp->result, "%d", (int) result);
  return TCL_OK;
  }

/////////////////////////////////////////////////////////////////
/* Synopsis: Universal Abstraction of bddvars on the BDD (bddvars is a BDD cube) */

 
int BDD_UAbstractCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
 {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc!=3, 0, "uabstract bdd bddvars" );

  BDDnode *v = (BDDnode*) atoi(argv[2]);
  BDDSH_ERROR_IF( v==NULL, 1, argv[2] );

  IndexArray ia(bdd, v);
  BDDnode *f = (BDDnode*) atoi(argv[1]);
  BDDSH_ERROR_IF( f==NULL, 1, argv[1] );

  BDDnode *result = bdd->USmooth( f, &ia );
  sprintf(interp->result, "%d", (int) result);
  return TCL_OK;
  }

/////////////////////////////////////////////////////////////////
/* Synopsis: Perform Boolean not on BDD */

int BDD_NotCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc!=2, 0, "not bdd" );

  BDDnode *result = (BDDnode*)atoi(argv[1]);
  BDDSH_ERROR_IF( result==NULL, 1, argv[1] );

  result = bdd->not(result);

  sprintf(interp->result, "%d", (int) result);
  return TCL_OK;
  }


/////////////////////////////////////////////////////////////////
/* Synopsis: Return the BDD assuming the top node is '0' */

int BDD_ElseCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;

  BDDSH_ERROR_IF( argc != 2, 0, "elseptr bdd" );

  BDDnode *result = (BDDnode*) atoi( argv[1] );
  BDDSH_ERROR_IF( result ==NULL, 1, argv[1] );

  result = result->Else();

  sprintf(interp->result, "%d", (int) result);
  return TCL_OK;
  }


/////////////////////////////////////////////////////////////////
/* Synopsis: Return the BDD assuming the top node is '1' */

int BDD_ThenCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;

  BDDSH_ERROR_IF( argc!=2, 0, "thenptr bdd" );

  BDDnode *result = (BDDnode*) atoi( argv[1] );
  BDDSH_ERROR_IF( result ==NULL, 1, argv[1] );

  result = result->Then();

  sprintf(interp->result, "%d", (int) result);
  return TCL_OK;
  }

#ifndef NO_MEM
/////////////////////////////////////////////////////////////////
/* Synopsis: Print HomeBrew Memory Statistics */

int BDD_MemStatsCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  argv = argv;
  BDDmanager *bdd = (BDDmanager*)clientdata;

  BDDSH_ERROR_IF( argc!=1, 0, "memstats" );

  mem_default_group.PrintStats();
  cout << flush;
  sprintf(interp->result, "ok");
  return TCL_OK;
  }
#endif

/////////////////////////////////////////////////////////////////
/* Synopsis: Pick one minterm in the BDD, given the vars in BDD 2 */
int
BDD_PickMTCmd(ClientData clientdata, Tcl_Interp *interp, int argc,char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc > 3, 0, "pickmt bdd [all|bdd]");

  BDDnode *f, *result, *iabdd;
  BDDSH_ERROR_IF( (f = (BDDnode*) atoi(argv[1])) == NULL, 8, argv[1] );

  IndexArray *ia;
  if( argc == 3 && strcmp( argv[2], "all" ) != 0 ) {
    BDDSH_ERROR_IF( (iabdd = (BDDnode*) atoi(argv[2])) == NULL, 8, argv[2] );
    ia = new IndexArray(bdd, iabdd );
    }
  else
    ia = new IndexArray( 0, f->GetIndex() );

      
  result = bdd->SampleCube( f );
  result = bdd->DC2Zero(result, ia);

  sprintf(interp->result, "%d", result);
  delete ia;
  return TCL_OK;
  }

/////////////////////////////////////////////////////////////////
/* Synopsis: Return number of nodes in a list of BDD's (or in the Manager) */

int
BDD_NumNodesCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;

  if( argc == 1 ) {
    sprintf( interp->result, "%d", bdd->NumNodes());
    return TCL_OK;
    }

  int i,result=0;
  for( i=1 ; i<argc ; i++ ) {
    HasherTree ht;
    BDDnode *node = (BDDnode*) atoi( argv[1] );
    BDDSH_ERROR_IF( node ==NULL, 1, argv[1] );

    result += bdd->NumNodes( node, &ht );
    }

  sprintf(interp->result, "%d", (int) result);
  return TCL_OK;
  }

/////////////////////////////////////////////////////////////////
/* Synopsis: Return number of cubes in BDD */

int BDD_NumCubesCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[ ])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc != 2, 0, "numcubes bdd" );

  BDDnode *f = (BDDnode*) atoi(argv[1]);
  BDDSH_ERROR_IF( f==NULL, 1, argv[1] );

  int result = bdd->NumCubes(f);
  sprintf(interp->result, "%d", (int) result);

  return TCL_OK;
}

/////////////////////////////////////////////////////////////////
/* Synopsis: Return minimum # of vars that must be '1' to satisfy BDD */

int BDD_MinCubeCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[ ])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc != 2, 0, "mincube bdd" );

  BDDnode *f = (BDDnode*) atoi(argv[1]);
  BDDSH_ERROR_IF( f==NULL, 1, argv[1] );

  int result = bdd->MinCube(f);
  sprintf(interp->result, "%d", (int) result);

  return TCL_OK;
}

/////////////////////////////////////////////////////////////////
/* Synopsis: Return number of minterms in BDD given num_indexes vars */

int BDD_SetSizeCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc < 2 || argc > 3, 0, "setsize bdd [num_indexes]");

  BDDnode* f = (BDDnode*)atoi(argv[1]);
  BDDSH_ERROR_IF( f==NULL, 1, argv[1] );
  int levels=0;
  if( argc==3 )
    levels = atoi(argv[2]);
  int result = bdd->OnSetSize(f, levels);
  sprintf(interp->result, "%d", (int) result);

  return TCL_OK;
}

///////////////////////////////////////////////////////
/* Synopsis: Return the index of the top node in the BDD */

int BDD_GetIndexCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc!=2, 0, "index bdd" );

  BDDnode *node = (BDDnode*) atoi( argv[1] );
  BDDSH_ERROR_IF( node ==NULL, 1, argv[1] );

  int result = node->GetIndex();
  sprintf(interp->result, "%d", (int) result);
  return TCL_OK;
  }


/////////////////////////////////////////////////////////////////
/* Synopsis: Return the level of the top node in the BDD */

int BDD_GetLevelCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc!=2, 0, "level bdd" );

  BDDnode *node = (BDDnode*) atoi( argv[1] );
  BDDSH_ERROR_IF( node ==NULL, 1, argv[1] );

  int result = bdd->GetLevel(node);
  sprintf(interp->result, "%d", (int) result);
  return TCL_OK;
  }


/////////////////////////////////////////////////////////////////
/* Synopsis: Mark BDD as kept (for garbage collection) */

int BDD_KeepCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc!=2, 0, "keep bdd" );

  BDDnode *result = (BDDnode*) atoi( argv[1] );
  BDDSH_ERROR_IF( result ==NULL, 1, argv[1] );

  bdd->Keep(result);
  sprintf(interp->result, "%d", (int) result);
  return TCL_OK;
  }

/////////////////////////////////////////////////////////////////
/* Synopsis: Unmark BDD as kept (for garbage collection) */

int BDD_UnKeepCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc!=2, 0, "unkeep bdd" );

  BDDnode *result = (BDDnode*) atoi( argv[1] );
  BDDSH_ERROR_IF( result ==NULL, 1, argv[1] );

  bdd->UnKeep(result);
  sprintf(interp->result, "%d", (int) result);
  return TCL_OK;
  }


/////////////////////////////////////////////////////////////////
/* Synopsis: Garbage collect all unkept nodes */

int BDD_GCCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  argv = argv;
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc!=1, 0, "gc" );

  bdd->GC();
  return TCL_OK;
  }


/////////////////////////////////////////////////////////////////
/* Synopsis: Garbage collect all unkept nodes, then reorder by sifting */

int BDD_SiftCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  argv = argv;
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc!=1, 0, "sift" );

  bdd->GC_Sift();
  return TCL_OK;
  }


/////////////////////////////////////////////////////////////////
/* Synopsis: Print statistics on the BDD cache and unique table */

int BDD_StatsCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  argv = argv;
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc!=1, 0, "bddstats" );

  static char str[200];
  static ostrstream out(str, 200, ios::out);

  out << "BDD stats:\nnum_inputs = " << bdd->NumInputs() << "\n";
  out << "num_nodes = " << bdd->NumNodes() << "\n";
  out << "num_ite (calls) = " << bdd->iteCalls() << "\n";
  bdd->computed->Print(out);
  Tcl_AppendElement(interp, out.str());
  return TCL_OK;
  }

/////////////////////////////////////////////////////////////////
/* Synopsis: Show the BDD as a graph using X11 routines */

int BDD_ShowCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc>4, 0, "show [bdd [window_name [show_ref]]]" );

  if( argc == 1 )
    bdd->XDump();
  else {
    BDDnode *result = (BDDnode*) atoi( argv[1] );

    char *name;
    if( argc >= 3 ) name = argv[2];	// name given
    else            name = argv[1];	// tcl id
    bdd->XDump(result, name, argc==4);
    sprintf(interp->result, "%d", (int) result);
  }

  return TCL_OK;
  }


/////////////////////////////////////////////////////////////////
/* Synopsis: Print the on-set cubes of a function */

int BDD_PrintCubesCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc != 2, 0, "printcubes bdd" );

  BDDnode *f = (BDDnode*) atoi( argv[1] );
  BDDSH_ERROR_IF( f == NULL, 1, argv[1] );

  bdd->PrintCubes( f );
  cout << flush;

  sprintf( interp->result, "%s", argv[1] );
  return TCL_OK;
  }


/////////////////////////////////////////////////////////////////
/* Synopsis: Print the order of the bdd variables */

int BDD_PrintOrderCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  argv = argv;
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc != 1, 0, "printorder" );

  int p=0,i;
  for( i=0 ; i<bdd->NumInputs() ; i++ ) {
    int index = bdd->unique->LookupIndex(i);
    sprintf( &interp_result[p], "%d ", index );
    p+=2;
    if( index > 9 ) p++;
    if( index > 99 ) p++;
    }
  interp_result[p-1] = '\0';
  interp->result = interp_result;

  return TCL_OK;
  }


/////////////////////////////////////////////////////////////////
/* Synopsis: Save the BDD(s) to a file in compressed binary format */

int BDD_SaveCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc<3, 0, "save filename bdd [bdd...]" );
  
  ofstream bddDump(argv[1]);
  BDDSH_ERROR_IF( !bddDump, 4, argv[1] );
  
  List l;
  while( --argc > 1 ) {
    BDDnode *f = (BDDnode*) atoi( argv[argc] );
    BDDSH_ERROR_IF( f==NULL, 1, argv[argc] );
    l.Add( f );
    }
  
  l.Reverse();
  bdd->WriteBinary(&l,bddDump);

  sprintf(interp->result, "%d functions written to file \"%s\"",
		  l.Count(), argv[1] );
  return TCL_OK;
  }

////////////////////////////////////////////////////////////////////
/* Synopsis: Load BDD(s) from a compressed binary file into memory */

int BDD_LoadCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc!=2, 0, "load filename" );
  
  ifstream bddDump(argv[1]);
  BDDSH_ERROR_IF( !bddDump, 4, argv[1] );
  
  List *l = bdd->ReadBinary(bddDump);

  List_ForEach_Type( l, BDDnode, f ) {
    cout << (int) f << " : ";
    cout << bdd->NumNodes( f ) << " nodes, ";
    cout << bdd->NumCubes( f )  << " cubes.\n";
    sprintf(interp->result, "%d", (int) f );
    }List_End_ForEach
  cout << flush;

  return TCL_OK;
  }

////////////////////////////////////////////////////////////////////
/* Synopsis: Read a BLIF file, return BDD's for the PO's and latches */

int BDD_ReadBLIFCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc!=2, 0, "readblif filename" );
  
  ifstream BLIFfile(argv[1]);
  BDDSH_ERROR_IF( !BLIFfile, 4, argv[1] );
  
  BLIFdata *bd = bdd->ReadBLIF(BLIFfile);

  List *l = bd->GetPONamesL();
  cout << l->Count() << " Primary Outputs:\n";
  List_ForEach_Type( l, StringObj, n ) {
    cout << n << " : ";
    BDDnode *f = bd->FuncOfPI( n );
    cout << (int) f << endl;
    }List_End_ForEach
  delete l;

  l = bd->GetNSNames();
  cout << l->Count() << " Next State Functions:\n";
  List_ForEach_Type( l, StringObj, n ) {
    cout << n << " : ";
    BDDnode *f = bd->FuncOfPI( n );
    cout << (int) f << endl;
    }List_End_ForEach
  delete l;

  l = bd->GetPINamesL();
  cout << l->Count() << " Primary Inputs:\n";
  while( l->Count() ) {
    cout << *(StringObj*) l->Pop();
    if( l->Count() ) cout << ", ";
    }
  delete l;

  cout << endl << flush;
  sprintf(interp->result, "%d", (int) bd );

  return TCL_OK;
  }

int BDD_WriteBLIFCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc>3 || argc<2, 0, "writeblif blifdata [filename]" );

  ostream *out = &cout;
  if( argc==3 ) {
    out = new ofstream(argv[2]);
    BDDSH_ERROR_IF( !out, 4, argv[2] );
    }
  
  BLIFdata *bd = (BLIFdata*)atoi(argv[1]);
  BDDSH_ERROR_IF( bd==NULL, 1, argv[1] );
  bd->Print( *out );
  interp->result = "done";

  if( argc==3 )
    delete out;

  return TCL_OK;
  }


////////////////////////////////////////////////////////////////////
/* Synopsis: Read PLA file, return onset, offset, dcset BDD's for the PO's */

int BDD_ReadPLACmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  BDDmanager *bdd = (BDDmanager*)clientdata;
  BDDSH_ERROR_IF( argc!=2, 0, "readpla filename" );
  
  ifstream PLAfile(argv[1]);
  BDDSH_ERROR_IF( !PLAfile, 4, argv[1] );
  
  List on, off, dc;
  bdd->ReadPLA( &on, &off, &dc, PLAfile );

  int i=0;
  cout << "function   onset     offset    dcset\n";
  while( on.Count() ) {
    cout << "o(" << setw(3) << i++ << "): " << setw(10) << (int) on.Pop();
    cout << setw(10) << (int) off.Pop();
    cout << setw(10) << (int) dc.Pop();
    cout << endl;
    }

  cout << flush;

  return TCL_OK;
  }

////////////////////////////////////////////////////////////////////
/* Synopsis: Rough graph printing routine, uses graph.tcl source file */

int HG_DrawHGraphCmd(ClientData clientdata, Tcl_Interp *interp, int argc, char *argv[])
  {
  clientdata=clientdata;
  BDDSH_ERROR_IF( argc!=2, 0, "drawgraph graph" );
  
  Hgraph *graph = (Hgraph*) atoi( argv[1] );
  BDDSH_ERROR_IF( graph==0, 1, argv[1] );

  char script[100];
  sprintf( script, "%d", graph->NumNodes() );
  Tcl_SetVar(interp, "maxnodes", script, 0 );

  Hgraph_ForEach_Node( graph, n ) {
    sprintf( script, "addnode %d white -1 -1", n->GetNum() );
    cout << script << endl << flush;
    Tcl_Eval( interp, script );
    }Hgraph_End_ForEach
  
  int max=2;
  char nodename[10];
  char *escript = new char[8+2*4];
  Hgraph_ForEach_Edge( graph, e ) {
    int m = e->Fanin()+e->Fanout();
    if( m>max ) {
      max=m;
      delete escript;
      escript = new char[max*4+8];
      }
    sprintf( escript, "addedge" );
    Hgraph_ForEach_In_Node( e, n ) {
      sprintf( nodename, " %d", n->GetNum() );
      strcat( escript, nodename );
      }Hgraph_End_ForEach
    Hgraph_ForEach_Out_Node( e, n ) {
      sprintf( nodename, " %d", n->GetNum() );
      strcat( escript, nodename );
      }Hgraph_End_ForEach
    cout << escript << endl << flush;
    Tcl_Eval( interp, escript );
    }Hgraph_End_ForEach
  delete escript;

  cout << flush;
  return TCL_OK;
  }

/////////////////////////////////////////////////////////

char *Tk_MMWindowName = NULL;

#ifndef NO_MEM
extern MemGroup mem_default_group;

void*
Tk_MMInitPrivate( void * arg ) {
  char **argv = (char**)arg;
  Tk_Main( 1, argv, Tk_MMInit );
  exit(0);
  return NULL;
  }
/*
void
Tk_MemoryManagerInit( char *argv[] ) {
  thread_t mmid;
  thr_setconcurrency(3);
  int flags = 0;
  int result = thr_create( NULL, 0, Tk_MMInitPrivate, (void*) argv, flags, &mmid );
  }
*/

int Tk_MMInit(Tcl_Interp *interp) {
  if( Tcl_Init(interp) == TCL_ERROR ) {
    return TCL_ERROR;
    }
  if( Tk_Init(interp) == TCL_ERROR) {
    return TCL_ERROR;
    }
  Tcl_SetVar( interp, "tcl_interactive", "0", TCL_GLOBAL_ONLY );
  if( Tcl_MMLinkVars(interp) == TCL_ERROR) {
    return TCL_ERROR;
    }
  if( Tk_MMInternal(interp) == TCL_ERROR) {
    return TCL_ERROR;
    }
  return TCL_OK;
  }

int Tk_MMInternal( Tcl_Interp *interp ) {

  Tcl_EvalFile(interp, "mm.tcl" );

  if( Tk_MMWindowName==NULL )
    Tcl_Eval( interp, "init_memgroup no" );
  else
    Tcl_Eval( interp, "init_memgroup yes" );

  return TCL_OK;
  }

int Tcl_MMLinkVars(Tcl_Interp *interp) {
  Tcl_LinkVar(interp, "num_malloced", (char*) &(mem_default_group.num_malloced),
	      TCL_LINK_INT|TCL_LINK_READ_ONLY);
  Tcl_LinkVar(interp, "num_inuse", (char*) &(mem_default_group.num_inuse),
	      TCL_LINK_INT|TCL_LINK_READ_ONLY);
  Tcl_LinkVar(interp, "num_free", (char*) &(mem_default_group.num_free),
	      TCL_LINK_INT|TCL_LINK_READ_ONLY);
  Tcl_LinkVar(interp, "num_big", (char*) &(mem_default_group.num_big),
	      TCL_LINK_INT|TCL_LINK_READ_ONLY);
  Tcl_LinkVar(interp, "malloc_count", (char*) &(mem_default_group.malloc_count),
	      TCL_LINK_INT|TCL_LINK_READ_ONLY);
  Tcl_LinkVar(interp, "alloc_count", (char*) &(mem_default_group.alloc_count),
	      TCL_LINK_INT|TCL_LINK_READ_ONLY);
  
  for( int i=0 ; i<120 ; i++ ) {
    char name[20];
    sprintf( name, "used(%d)", i );
    Tcl_LinkVar(interp,name, (char*) &(mem_default_group.allocated[i]),
		TCL_LINK_INT|TCL_LINK_READ_ONLY);
    sprintf( name, "free(%d)", i );
    Tcl_LinkVar(interp,name, (char*) &(mem_default_group.free_count[i]),
		TCL_LINK_INT|TCL_LINK_READ_ONLY);
    }
  return TCL_OK;
  }
#endif
