#!/usr/local/bin/wish
################################################
# constants for displaying the graphs
set NODEW 60
set NODEH 20
set COLORS "white red green yellow cyan magenta orange violet blue"
set CLUSTER_RADIUS 100

set CENTER 300
set RADIUS 150

set maxnodes 10
set maxclusters 2

# variables for displaying the graphs
set centerx() 0
set centery() 0

set clusterposi [ expr $maxclusters * 3 / 4 ]
set nposi() 0
set nextx 0
set nexty 0

set colorof() 0
set graph default

################################################

# variables used in the algorithms stuff
set nodenames ""


################################################

# graph drawing routines

# set initial graph partitions
proc setnxy cluster {
  global nposi nextx nexty maxnodes RADIUS CENTER CLUSTER_RADIUS
  global centerx centery clusterposi maxclusters
  if { [ info exists centerx($cluster) ] == 0 } {
    set nposi($cluster) [ expr $maxnodes * 3 / 4 ]
    set centerx($cluster) [ expr sin( $clusterposi * 6.2832 / $maxclusters ) * \
			    $RADIUS + $CENTER ]
    puts "centerx($cluster) is $centerx($cluster)"
    set centery($cluster) [ expr cos( $clusterposi * 6.2832 / $maxclusters ) * \
			    $RADIUS + $CENTER ]
    puts "centery($cluster) is $centery($cluster)"
    incr clusterposi
    }
  set nextx [ expr sin( $nposi($cluster) * 6.2832 / $maxnodes ) * \
	      $CLUSTER_RADIUS + $centerx($cluster) ]
  puts "nextx is $nextx"
  set nexty [ expr cos( $nposi($cluster) * 6.2832 / $maxnodes ) * \
	      $CLUSTER_RADIUS + $centery($cluster) ]
  puts "nexty is $nexty"
  incr nposi($cluster)
  }
  
proc addnode { name color xpos ypos } {
  global NODEW NODEH nextx nexty INCRX INCRY cluster
  global nodenames colorof graph

  if { $xpos == -1 || $ypos == -1 } {
    if { [ info exists cluster(n$name) ] == 0 } {
      set cluster(n$name) 0
      }
    setnxy $cluster(n$name)
    set xpos $nextx
    set ypos $nexty
    }
  set x1 [ expr $xpos - ( $NODEW / 2 ) ]
  set x2 [ expr $xpos + ( $NODEW / 2 ) ]
  set y1 [ expr $ypos - ( $NODEH / 2 ) ]
  set y2 [ expr $ypos + ( $NODEH / 2 ) ]
  set colorof($name) $color
  .$graph.c create oval $x1 $y1 $x2 $y2 -fill $color -tag "n$name" -width 2
  .$graph.c create text $xpos $ypos -text $name -tag "lab$name"
  lappend nodenames $name

  }

proc getnodecoords name {
  global graph
  return [ .$graph.c coords "n$name" ]
  }

proc addedge args  {
  global nextx nexty graph

  set x() 0
  set y() 0
  set centerx 0
  set centery 0

  set marker [ lsearch -exact $args ">" ]
  set to [ lrange $args [ expr 1 + $marker ] end ]
  set from [ lrange $args 0 [ expr $marker - 1 ] ]

  # compute the center of all nodes in $args
  set num [ llength $args ]
  for { set i 0 } { $i < $num } { incr i } {
    set name [ lindex $args $i ]
    set c [ getnodecoords $name ]
    set x($name) [expr ( [lindex $c 0] + [lindex $c 2] ) / 2]
    set y($name) [expr ( [lindex $c 1] + [lindex $c 3] ) / 2]
    set centerx [ expr $centerx + $x($name) ]
    set centery [ expr $centery + $y($name) ]
    }
  set centerx [ expr $centerx / $num ]
  set centery [ expr $centery / $num ]

  foreach n $from {
    set cur [ .$graph.c create line $x($i) $y($i) $centerx $centery \
	       -arrow first ]
    for { set j 0 } { $j < $num } { incr j } {
      if { $j == $i } {
	.$graph.c addtag "ef[ lindex $args $j]" withtag $cur
	} else {
	.$graph.c addtag "et[ lindex $args $j]" withtag $cur
	}
      }
#    .$graph.c lower $cur
    }
  }

proc rmedge args {
  global graph nodenames

  foreach n $args {
    set ef($n) [.$graph.c find withtag "ef$n"]
    set et($n) [.$graph.c find withtag "et$n"]
    }
  
  set edges ""
  foreach n1 $args {
    foreach efrom $ef($n) {
      foreach n2 $args {
	if { [ lsearch -exact $et($n2) $efrom ] > -1 } {
	  lappend edges $efrom
	  }
	}
      }
    foreach eto $et($n) {
      foreach n2 $args {
	if { [ lsearch -exact $ef($n2) $eto ] > -1 } {
	  lappend edges $eto
	  }
	}
      }
    }
  foreach e $edges {
    .$graph.c del $e
    }
  }

proc clearall { } {
  global graph nodenames

  set nodenames ""
  .$graph.c delete all
  }

proc makegraph { name } {
  global graph

  if { [winfo exists .$name] } {
    puts "graph $name already exists"
    return
    }
  set graph $name
  if { $graph != "default" } {
    toplevel .$graph
    wm title .$graph "graph: $graph"
    } else {
    frame .default
    pack .default
    wm title . "graph: $graph"
    }
  canvas .$graph.c -width 600 -height 600
  pack .$graph.c
  puts stdout "current graph is $graph"
  }

proc rmgraph { name } {
  global graph

  if { [ winfo exists .$name ] == 0 } {
    puts "no suck graph \"$name\""
    return
    }
  if { $name == "default" } {
    puts "cannot destroy the default graph!"
    return
    }
  destroy .$name
  if { $graph == $name } {
    set graph default
    }
  }
   
makegraph default

##############################################################
# end generic hyper graph package
##############################################################
