#ifndef MEM_H
#define MEM_H

///////////////////////////////////////////////////////////////////////////////
//
//  mem.h : C++ Memory Allocation / Management object
//
//  Andrew Seawright      8/10/91
//
// $Header: /usr6/research/code/homebrew/RCS/mem.h,v 1.12 1998/09/14 23:26:46 crews Exp $
//
//  modeled after Magic's memory allocator
//
///////////////////////////////////////////////////////////////////////////////
#include <iostream.h>
#include <stdlib.h>
#include "stddefs.h"
#include "memhash.h"

#define CACHE_BIGGEST    120          // size of biggest object that is cached
#define CACHE_THRESHOLD  3            // # hits before starting cacheing
#define PAGESIZE         4096         // number of objects in fresh new page
#define MEM_WORD_SIZE    4            // number of bytes in a word
#define FREE_TAG         0xdead       // header tag for freed objects (not used)


///////////////////////////////////////////////////////////////////////////////
//
//  How to use:
//
//  o In a base class overload the new() and delete() operators:
//
//	void * operator new(size_t n)
//	  { return mem_default_group.Malloc(n); }
//	void operator delete (void *p, size_t n)
//        { mem_default_group.Free(p, n); }
//       
//      This has already been done to Object if MEM is #defined, see object.h
//    
//  o If you are only using the memory class from HomeBrew, you must construct 
//       one (or several) MemGroup instance.  Otherwise skip this step:
//
//       extern MemGroup *mem_default_group;  // typically a global variable 
//             
//       mem_default_group = new MemGroup;
//
//  o Allocate new objects as usual:
//     
//       Foo *f;
//       f = new foo;
//
//
///////////////////////////////////////////////////////////////////////////////

#ifdef MEM_STATS
extern int _global_num_alloc;
extern int _global_num_inuse;
extern int _global_num_freed;
#endif

class Tcl_Interp;

class MemGroup {  

  int    num_malloced;  // total num of bytes from system malloc
  int    num_inuse;     // total num of bytes in active use
  int    num_free;      // total num of recycled available mem
  int    num_big;       // total num of big chunks
  int    malloc_count;  // number of times system malloc called
  int    alloc_count;   // number of times MemGroup::Malloc() called
  int    alloc_watch;   // 

  char **free;        // free list
  int   *hits;        // 
  int   *allocated;   //
  int   *free_count;  //
#ifdef MEM_STATS
  MemHashTree verify;
#endif

public:
  
  MemGroup();
 ~MemGroup();
  void * Malloc(size_t);
  void Free(void *, size_t);
  void PrintStats( ostream& out = cout);
  void RecoverMem();

  void * operator new(size_t n)       // make sure this object is allocted
	  { return ::malloc(n); }     // with system malloc
  void operator delete (void *p)
         { ::free((char*) p); }
#ifdef MEM_STATS
  friend void *operator new(size_t n);			// global function
  friend void operator delete (void* p );	// global function
#endif
  friend Tcl_MMLinkVars(Tcl_Interp*);
};

///////////////////////////////////////////////
// Extra functions which uses pstat values

// look in memfunts.C for the source

#ifdef MEM_STATS
void *operator new(size_t n);
void operator delete (void* p );
#endif


#endif
