#ifndef TNODE_H
#define TNODE_H

#include <iostream.h>
#include "compable.h"
#include "classdef.h"

#define TNODE_REFERENCE (unsigned short) 0x3FFF
#define TNODE_MARK   (unsigned short) 0x8000
#define TNODE_OVERFLOW (unsigned short) 0x4000

class Tnode : public Comparable {

protected:
  Tnode 	   *left;
  Tnode		   *right;
  unsigned short    reference;
  short             index;

  void    SetLeft(Tnode *l)  { left  = l;}
  void    SetRight(Tnode *r) { right = r;}
  void	  SetReference( int r )  { reference &= ~TNODE_REFERENCE;
				   reference |= (r & TNODE_REFERENCE); }
  // these can set the overflow
  void    Reference()     { reference = ((reference & TNODE_REFERENCE)+1) |
					(reference & ~TNODE_REFERENCE)   ; }
  void    UnReference()   { reference = (reference & ~TNODE_REFERENCE) |
					((reference-1) & TNODE_REFERENCE); }

public:
  Tnode() { left = right = NULL; reference=0; }
 ~Tnode() {;}

  Tnode*  Left()          { return left;}
  Tnode*  Right()         { return right;}
  Tnode*  One()           { return left;}
  Tnode*  Zero()          { return right;}

  // this returns a positive value if overflow is set
  unsigned NumReferences(){ return reference & ~(TNODE_MARK); }

  void    Mark()          { reference |= TNODE_MARK;	}
  void    UnMark()        { reference &= ~TNODE_MARK;	}
  boolean Marked()        { return (reference & TNODE_MARK)!=0;}
  void    SetOverflow()   { reference |= TNODE_OVERFLOW; }
  boolean Overflow()      { return (reference & TNODE_OVERFLOW)!=0; }

  virtual void Print(ostream& out = cout) { out << "Tnode\n"; } 
};

inline ostream& operator << (ostream &out, Tnode& f)
{ f.Print(out); return out; }

inline ostream& operator << (ostream &out, Tnode* f)
{ out << "(ptr->)"; f->Print(out); return out; }

#endif // TNODE_H

