#include "window.h"
#include "bdd.h"

BubbleMgr::BubbleMgr( BDDmgr *bddm, Object* dag, char *window_title,
		      boolean sr, char **inames)
  {
  show_ref_count = sr;
 
  xdisplay  = new XDisplay;
  if(xdisplay->Get() == NULL)
    return;

  indexNames = inames;
  function = dag;
  NumInLevel = NULL;
  Depth = NULL;
  SetDepth(bddm);
  xwindow   = new XWindow( xdisplay, window_title, maxdepth );
  fname     = FONTNAME;
  font_info = XLoadQueryFont(xdisplay->Get(), fname);
  xgc       = new XGC(xdisplay, xwindow, font_info);

  coord2node = new HasherTree;
  node2coord = new HasherSort;

  // create a new manager WITHOUT caches
  bdd = new BDDmgr(bddm,0,0);


  MakeLabelWindows();
  MakeButtons();

  InitBubbles(bddm);

  }

void
BubbleMgr::InitBubbles(BDDmgr *oldbdd) {

  if( function->isa() == BDDNODE )
    {
    function = bdd->Copy2Manager((BDDnode*)function, oldbdd);
    bdd->Keep((BDDnode*)function);
    AddNodeS( (BDDnode*)function );
    }

  else if( function->isa() == BDDHASHERARRAY )
    {
    bdd->Dump2Manager( oldbdd );
    function = bdd->unique;
    BDDHasherArray_ForEach_Type( ((BDDHasherArray*)function), N) {
      if( N != bdd->One() && N != bdd->Zero() )
        AddNode( N );
      }BDDHasherArray_End_ForEach
    }

  else if( function->isa() == LIST )
    {
    List *newfunc = new List;
    HasherTree ht;
    List_ForEach_Type( (List*) function, BDDnode, F )
      {
      F = bdd->Copy2Manager(F, oldbdd, &ht);
      newfunc->Add(F);
      bdd->Keep(F);
      AddNodeS( F );
      }List_End_ForEach
    // leave function deletion to be handled by user outside XDump
    function = newfunc;
    }
  else
    FATAL("XDump plots only functions, function lists, or hasher values");
  }

void
BubbleMgr::InitBubbles() {

  if( function->isa() == BDDNODE )
    AddNodeS( (BDDnode*)function );

  else if( function->isa() == BDDHASHERARRAY )
    {
    BDDHasherArray_ForEach_Type( ((BDDHasherArray*)function), N) {
      if( N != bdd->One() && N != bdd->Zero() )
        AddNode( N );
      }BDDHasherArray_End_ForEach
    }

  else if( function->isa() == LIST )
    {
    List_ForEach_Type( (List*) function, BDDnode, F )
      {
      AddNodeS( F );
      }List_End_ForEach
    }
  }


BubbleMgr::~BubbleMgr() {
  if( !xdisplay ) return;	// aborted, no cleanup
  XUnmapWindow( xdisplay->Get(), xwindow->Get() );
  delete xgc;
  delete xwindow;	// must go after gc delete
  delete xdisplay;	// must go after window and gc delete
  delete NumInLevel;
  delete Depth;
  delete coord2node;	// HasherTree
  delete node2coord;	// HasherSort
  for( int i=0 ; i<maxdepth ; i++ )
    delete label_names[i];
  delete label_names;
  delete labels;
  if( function->isa() == LIST )
    delete function;
  delete bdd;		// after delete function (could be HasherArray)
  }

void
BubbleMgr::Print( ostream& out ) {
  out << "Not Done Yet!" << endl << flush;
  }


void
BubbleMgr::AddNodeS( BDDnode* f )
  {
  if( f == bdd->One() || f == bdd->Zero() )
    return;
 
  if( node2coord->Get( f ) )	// have we seen this one before?
    return;

  AddNodeS(f->Then() );
  AddNodeS(f->Else() );
  AddNode( f );
  }

void
BubbleMgr::AddNode( BDDnode* n )
  {
  int level = bdd->GetLevel(n);

  // if there is no label there, set it to the index of the nodes on that level
  if( label_names[Depth[level]-1][0] == '\0' )
    if( indexNames )
      sprintf( label_names[Depth[level]-1], "%.6s", indexNames[n->GetIndex()] );
    else
      sprintf( label_names[Depth[level]-1], "%d", n->GetIndex() );

  // add a margin of MARGIN_OFFSET
  BubbleObj* xy = new BubbleObj(horz * NumInLevel[level]++ + MARGIN_OFFSET, vert * Depth[level]);

  // we know the initilization will never place two nodes on top of one another
  // If this wasn't the case, it would be necessary to check and make sure,
  // since a Put will fail, if the locations are two close (same grid spot)
  node2coord->Put( xy, n );
  coord2node->Put( n, xy );
  }

void
BubbleMgr::SetDepth(BDDmgr *bddm)	// bdd is still unset, so pass the old 
  {
  HasherTree nodes;
  numcubes = 0;

  int i;

  // We must allow for highest level even among variables which are not
  // present in the bdd itself.

  for( maxlevel=i=0 ; i<bddm->NumInputs() ; i++ )
    maxlevel = MAX(maxlevel, bddm->unique->LookupLevel(i)+1);

  if( function->isa() == BDDNODE )
    {
    BDD2Hash( (BDDnode*) function, &nodes, bddm );
    numcubes = bddm->NumCubes((BDDnode*)function);

    }
  else if( function->isa() == LIST )
    {
    List_ForEach_Type( (List*) function, BDDnode, F )
      {

      BDD2Hash( F, &nodes, bddm );
      }List_End_ForEach
    }
  else if( function->isa() == BDDHASHERARRAY )
    {
    BDDHasherArray_ForEach_Type( ((BDDHasherArray*)function), F )
      {
      nodes.Put( F );
      }BDDHasherArray_End_ForEach
    }

  numnodes = 0;
  maxdepth = 0;

  if( NumInLevel ) delete NumInLevel;
  if( Depth ) delete Depth;
  NumInLevel = new int[maxlevel];
  Depth      = new int[maxlevel];

  for( i=0 ; i<maxlevel ; i++ )
    NumInLevel[i] = 0;

  Hasher_ForEach_Type( &nodes, BDDnode, k, N )
    {
    i = bddm->GetLevel(N);
    if( NumInLevel[i] == 0 )
      maxdepth++;
    NumInLevel[i]++;
    }Hasher_End_ForEach
  numnodes = nodes.Count();

  int depth = maxdepth;

  for( i=0 ; i<maxlevel ; i++ )
    {
    if( NumInLevel[i] )
      Depth[i] = depth--;
    else
      Depth[i] = -1;
    
    NumInLevel[i] = 0;		// these are initialized as the nodes are added
				// so that we can give each a unique x coordinate
    }
 
  }

void
BubbleMgr::Straighten( BDDnode* n )
  {

  HasherTree* redundant = new HasherTree;

  BubbleObj* coord = (BubbleObj*) node2coord->Get( n );

  StraightenR( n, coord->x, coord->y, redundant );

  // now we must delete the "redundant" ht, even though it has the correct
  // format, since we don't know if two or more bubbles are in the exact
  // same location! If they are, then they cannont both be put into the 
  // hashertree, so one must be moved slightly (the while loop below )
  delete redundant;

  HasherTree* coord2node_save = new HasherTree;
  HasherTree* node2coord_save = new HasherSort;

  Hasher_ForEach_Type( coord2node, BDDnode, K, F )
    {
    while( ! coord2node_save->Put( F, (BubbleObj*) K ) )
      ((BubbleObj*)K)->x += xgrid;
    node2coord_save->Put( K, F );
    }Hasher_End_ForEach
  

  delete coord2node;
  node2coord->SetDelete(FALSE,FALSE);
  delete (HasherTree*) node2coord;

  coord2node = coord2node_save;
  node2coord = node2coord_save;

  }


void
BubbleMgr::StraightenR( BDDnode* n, int x, int y, HasherTree* ht )
  {

  BubbleObj* left;

  if( left = (BubbleObj*) node2coord->Get( n->Then() ) )
    {
    left->x = x - STRAIGHTEN_X_OFFSET;
    left->y = y + vert;

    // terminating case
    if( ht->Put( left, n->Then() ) )
      StraightenR( n->Then() , left->x , left->y, ht );
    }

  BubbleObj* right;
  // check if n->Else is one or zero ( and not in the node2coord hasher )
  if( right = (BubbleObj*) node2coord->Get( n->Else() ) )
    {
    right->x = x + STRAIGHTEN_X_OFFSET;
    right->y = y + vert;

    // terminating case
    if( ht->Put( right, n->Else() ) )
      StraightenR( n->Else(), right->x, right->y, ht );
    }

  }
 

void
BubbleMgr::BDD2Hash( BDDnode* f, HasherTree* ht, BDDmgr *bddm )
  {
  if( f == bddm->One() || f == bddm->Zero() )
    return;

  if( ht->Put( f ) )
    {
    BDD2Hash( f->Then(), ht, bddm );
    BDD2Hash( f->Else(), ht, bddm );
    }
  }
 


void
BubbleMgr::expose()
  {

  // Load font and get font info structure for use in following loop
  // I use this long obnoxious name so that it ports to both DEC and Sparc

  if (font_info == NULL)
    {
    cerr << "Cannot open " << fname << " font" << endl;
    exit (1);
    }

  char line[80];
  if( (BDDnode*)function == bdd->One() )
    sprintf(line, "TRUE");
  else if( (BDDnode*)function == bdd->Zero() )
    sprintf(line, "FALSE");
  else if( numcubes != 0 )
    sprintf(line, "%d levels, %d nodes %d cubes:",maxdepth, numnodes, numcubes);
  else
    sprintf(line, "%d levels, %d nodes:", maxdepth, numnodes);

  XDrawString(xdisplay->Get(), xwindow->Get(), xgc->Get(), TITLE_X, TITLE_Y, line, strlen(line));

  XWindowAttributes xwa;
  XGetWindowAttributes( xdisplay->Get(), xwindow->Get(), &xwa );
  Hasher_ForEach_Type( coord2node, BDDnode, K, N )
    {
    DrawNode( (BubbleObj*) K, N, &xwa );
    }Hasher_End_ForEach
  }



void
BubbleMgr::DrawNode( BubbleObj* position, BDDnode* n, XWindowAttributes *xwa )
  {
  if( position->flags == INVISIBLE )
    return;
  int x, y;
  int width;

  x = position->x;
  y = position->y;

  boolean visible=FALSE;

  if( x-BUBBLE_WIDTH/2 < xwa->x+xwa->width &&
      y-BUBBLE_HEIGHT/2 < xwa->y+xwa->height &&
      x+BUBBLE_WIDTH/2 > x &&
      y+BUBBLE_HEIGHT/2 > y ) {
    visible = TRUE;

    char line[10];
    if( show_ref_count )
      sprintf( line, "%s|%d", label_names[Depth[bdd->GetLevel(n)] -1 ], n->NumReferences() );
    else 
      sprintf( line, "%s", label_names[Depth[bdd->GetLevel(n)] -1 ] );

    width = XTextWidth( font_info, line, strlen(line) );


    int height = font_info->max_bounds.ascent + font_info->max_bounds.descent;
    XDrawString (xdisplay->Get(), xwindow->Get(), xgc->Get(),
		  x-width/2, y+height/2, line, strlen(line));
    XDrawArc (xdisplay->Get(), xwindow->Get(), xgc->Get(),
	      x-BUBBLE_WIDTH/2 , y-BUBBLE_HEIGHT/2, BUBBLE_WIDTH, BUBBLE_HEIGHT, 0, 360*64);

  // Do this only if this is not a terminal node
  if( n != bdd->One() && n != bdd->Zero() )
    drawConnections( x, y, n );

    }	// end if visible
 
// #ifdef _XBDD_LABEL_
// // Specify small font and print a small "1" and "0" to denote the
// // true and false edges... Some people don't like this drawn, because it
// // confuses them
// XSetFont (xdisplay->Get(), xgc->Get(), font_info->fid);
// XDrawString (xdisplay->Get, xwindow->Get(), xgc->Get(), x-5, y+12, "1", 1);
// XDrawString (xdisplay->Get(), xwindow->Get(), xgc->Get(), x+33, y+12, "0", 1);
//
// // Specify GC's font
// XSetFont (xdisplay->Get(), xgc->Get(), xgc->GetFontInfo()->fid);
//#endif
  }

void
BubbleMgr::drawConnections( int x, int y, BDDnode* n )
  {

  //////////////////////////////////////////////////////////////////////////////////////////
  // Draw box for one or zero on either side

  if(n->Then() == bdd->One() ||n->Then() == bdd->Zero() )
    {
    XDrawLine( xdisplay->Get(), xwindow->Get(), xgc->Get(),
		x-PTR_OFF, y+BUBBLE_HEIGHT/2, x-BOX_X, y+BOX_Y);
    if( n->Then() == bdd->One() )
      XFillRectangle( xdisplay->Get(), xwindow->Get(), xgc->Get(),
			x-BOX_X-BOX_SIZE, y+BOX_Y, BOX_SIZE, BOX_SIZE);
    else			// zero
      XDrawRectangle( xdisplay->Get(), xwindow->Get(), xgc->Get(),
			x-BOX_X-BOX_SIZE, y+BOX_Y, BOX_SIZE, BOX_SIZE);
    }

  if( n->Else() == bdd->One() || n->Else() == bdd->Zero() )
    {
    XDrawLine( xdisplay->Get(), xwindow->Get(), xgc->Get(),
		x+PTR_OFF, y+BUBBLE_HEIGHT/2, x+BOX_X, y+BOX_Y);
    if( n->Else() == bdd->One() )
      XFillRectangle( xdisplay->Get(), xwindow->Get(), xgc->Get(),
		     x+BOX_X, y+BOX_Y, BOX_SIZE, BOX_SIZE);
    else			// zero
      XDrawRectangle( xdisplay->Get(), xwindow->Get(), xgc->Get(),
			x+BOX_X, y+BOX_Y, BOX_SIZE, BOX_SIZE);
    }

  //////////////////////////////////////////////////////////////////////////////////
  // Draw line to other nodes
 
  int xchild, ychild;
  BubbleObj*	position;

  if( position = (BubbleObj*) node2coord->Get(n->Then() ) )
    {
    // Check for visibility here
    if( position->flags == VISIBLE )
      {
      xchild = position->x;
      ychild = position->y;
      XDrawLine( xdisplay->Get(), xwindow->Get(), xgc->Get(), x-PTR_OFF,
		   y+BUBBLE_HEIGHT/2, xchild, ychild-BUBBLE_HEIGHT/2);
      }
    }
 
  if( position = (BubbleObj*) node2coord->Get(n->Else() ) )
    {
    // Check for visibility here
    if( position->flags == VISIBLE )
      {
      xchild = position->x;
      ychild = position->y;
      XDrawLine( xdisplay->Get(), xwindow->Get(), xgc->Get(), x+PTR_OFF,
		   y+BUBBLE_HEIGHT/2, xchild, ychild-BUBBLE_HEIGHT/2);
      }
    }

  }




int
BubbleMgr::EventLoop(int flag)
{
  // Show window
  XMapWindow(xdisplay->Get(), xwindow->Get());

  boolean move=FALSE; // Denotes valid motion of a bubble
  boolean relabel=FALSE; // Denotes typing is valid label number "index_num"
  boolean done = FALSE;
  boolean keydown = FALSE;

  int index_num = maxdepth;
  int i;
  int key_count;
  int string_length = 0;
  char buffer[10];
  int retval = CHILD;

  KeySym keysym;
  XComposeStatus compose;

  XEvent report; // Xevents recorded here
  BubbleObj startpos(0,0);
  BubbleObj* grabbubble;
  BDDnode* grabnode;

  expose();

  // Putting the fork call here ensures that windows are placed in order
  if (flag == FORK)
  {
    if (fork() != 0)
      return PARENT;
    retval = FORKDONE;
  XUnmapWindow( xdisplay->Get(), buttons[2] );	// fork
  }

  // Window event loop
  // This loop does not exit until right button is pressed (or
  // someone kills the program in some devious manner)
  // meaning, if there are consecutive XDump() calls, the first one
  // is displayed, and then the next one(s)
  while (!done)
    {
    if( index_num == maxdepth )
      relabel = FALSE;
    XNextEvent(xdisplay->Get(), &report);
    switch (report.type)
    {

    ///////////////////////////////////////////////////////////////////////////////

    case Expose:
	if( report.xexpose.count == 0 && report.xexpose.window == xwindow->Get() )
          expose ();
	else
	  {
	  for( i=0 ; i< NUM_BUTTONS ; i++ )
	    if( report.xexpose.window == buttons[i] )
	      XDrawString( xdisplay->Get(), buttons[i], xgc->Get(), 2,
			font_info->max_bounds.ascent+2,
			button_names[i], strlen(button_names[i]) );
	  if( report.xexpose.window == filename_button ) {
	    XDrawString( xdisplay->Get(), filename_button, xgc->GetInv(), 2,
			font_info->max_bounds.ascent+2,
			filename, strlen(filename) );
	    }
	  for( index_num=0 ; index_num<maxdepth ; index_num++ )
	    if( report.xexpose.window == labels[index_num] )
              XDrawString( xdisplay->Get(), labels[index_num], xgc->Get(), 2,
			font_info->max_bounds.ascent+2,
	       		label_names[index_num], strlen(label_names[index_num]) );
	  }
	  
      break;

    ///////////////////////////////////////////////////////////////////////////////


    case ButtonPress:
      if( report.xbutton.window == buttons[0] )		// exit
	done = TRUE;
      else if ( report.xbutton.window == buttons[2] )	// fork
	{
        if ( retval == CHILD )
          {
          if (fork() != 0)
            return PARENT;
          retval = FORKDONE;
          XUnmapWindow( xdisplay->Get(), buttons[2] );
          }
	}
      else if ( report.xbutton.window == buttons[1] ) // print cubes
	{
        if( function->isa() == BDDNODE )
          {
          bdd->PrintCubes((BDDnode*)function);
          cout << flush;
          }
        }
      else if( report.xbutton.window == buttons[3] ) // save bdd
	{
        SiftBDD( );
	}
      else if( report.xbutton.window == buttons[4] ) // save bdd
	{
	XMapWindow( xdisplay->Get(), filename_button );
	XUnmapWindow( xdisplay->Get(), buttons[4] );
	XMapWindow( xdisplay->Get(), buttons[5] );
	XMapWindow( xdisplay->Get(), buttons[6] );
	}

      else if( report.xbutton.window == buttons[5] ) // ok (to save)
	{
	SaveBDD();
	}
      else if( report.xbutton.window == buttons[6] )
	{
	XUnmapWindow( xdisplay->Get(), filename_button );
	XUnmapWindow( xdisplay->Get(), buttons[5] );
	XUnmapWindow( xdisplay->Get(), buttons[6] );
	XMapWindow( xdisplay->Get(), buttons[4] );
	}
	  
      if( relabel && index_num < maxdepth )
	{
        label_names[index_num][--string_length] = '\0';
        XClearWindow( xdisplay->Get(), labels[index_num] );
        XDrawString( xdisplay->Get(), labels[index_num], xgc->Get(), 2, font_info->max_bounds.ascent+2,
			  label_names[index_num], string_length );

        XClearWindow( xdisplay->Get(), xwindow->Get() );
        expose();
	}

	
      relabel = FALSE;

      // shift middle button make all nodes visible again
      if( (report.xbutton.state & ShiftMask) && report.xbutton.button == 2 )
	{
        Hasher_ForEach_Type( node2coord, BubbleObj, K, B )
          {
          B->flags = VISIBLE;
          }Hasher_End_ForEach
	// adjust numnodes to all nodes, again
	numnodes = node2coord->Count();
        XClearWindow (xdisplay->Get(), xwindow->Get());
	expose();
	break;
	}
      
      // Get position where mouse was pressed

      startpos.x = report.xbutton.x;
      startpos.y = report.xbutton.y;

      // Search for the bubble we are trying to move
      FindNode( &startpos, grabnode );

      // Found a node!
      if( grabnode != NULL )
        {
        grabbubble = (BubbleObj*) node2coord->Get( grabnode );

        // Check for middle button press
        if (report.xbutton.button == 2)
          {
          // Delete node and redraw window
          SoftRemove( grabnode );
          XClearWindow (xdisplay->Get(), xwindow->Get());
          expose ();
          break;
          }
        if (report.xbutton.button == 1 && report.xbutton.state & ControlMask )
	  {
	  Straighten( grabnode );
          XClearWindow (xdisplay->Get(), xwindow->Get());
	  expose();
	  break;
	  }
  
        // Else, left button pressed! Prepare to move it!
        // All of these variables are saved for use in the next event
        move = TRUE;
  
        // Draw initial arc
        XSetFunction (xdisplay->Get(), xgc->Get(), GXequiv);
        XDrawArc (xdisplay->Get(), xwindow->Get(), xgc->Get(),
                   grabbubble->x - BUBBLE_WIDTH/2, grabbubble->y - BUBBLE_HEIGHT/2,
		   BUBBLE_WIDTH, BUBBLE_HEIGHT, 0, 360*64);
        XSetFunction (xdisplay->Get(), xgc->Get(), GXcopy);
	break;
	}

      // No node found, check to see if labels are being adjusted...
      else
        {
	move = FALSE;

	// set index_num to be the appropriate label
	for( index_num=0 ; index_num<maxdepth ; index_num++ )
	  if( report.xbutton.window == labels[index_num] )
	    break;

	if( index_num < maxdepth )
	  {
	  relabel = TRUE;
	  string_length = strlen(label_names[index_num]);
	  label_names[index_num][string_length++] = '_';
	  label_names[index_num][string_length] = '\0';
          XClearWindow( xdisplay->Get(), labels[index_num] );
          XDrawString( xdisplay->Get(), labels[index_num], xgc->Get(), 2,
		       font_info->max_bounds.ascent+2,
	  	       label_names[index_num], string_length );
	  }
	}

      break;

    ///////////////////////////////////////////////////////////////////////////////

    case ButtonRelease:

  
      if (move)
        {

        // Stop movement
        move = FALSE;


        // Otherwise, assign new position, redraw window
        HardRemove( grabnode );
        grabbubble->x += report.xbutton.x - startpos.x;

        // Check for Hien's attempt to move bubble too far left!
	if( grabbubble->y < MIN_LEFT )
	  grabbubble->y = MIN_LEFT;
	if( grabbubble->x < MIN_LEFT )
	  grabbubble->x = MIN_LEFT;

	// holding down shift key disallows movement in the y dimension
	if( !(report.xbutton.state & ShiftMask) )
          grabbubble->y += report.xbutton.y - startpos.y; 
	while( ! coord2node->Put( grabnode, grabbubble) )
	  grabbubble->x += xgrid;
        node2coord->Put( grabbubble, grabnode );
        XClearWindow (xdisplay->Get(), xwindow->Get());
        expose ();
        }
      else if( relabel ) // valid starting point
	{
        unsigned int keys;
        int root_x, root_y;
        int newx, newy;
        Window root, child;

	// query as two which child of the main window is now pointed to by the mouse
        XQueryPointer (xdisplay->Get(), xwindow->Get(), &root, &child,
                        &root_x, &root_y, &newx, &newy, &keys);
	// set index_num to be the appropriate label

        int new_index_num;
	for( new_index_num=0 ; new_index_num<maxdepth ; new_index_num++ )
	  if( child == labels[new_index_num] )
	    break;
	
	if( new_index_num < maxdepth && index_num != new_index_num )
	  {
	  if( function->isa() != HASHER_TREE && function->isa() != BDDHASHERARRAY )
	    {
	    SwapBDD( index_num, new_index_num );
            index_num = new_index_num;
	    }
	  }
	}
      break;

    ///////////////////////////////////////////////////////////////////////////////

    case MotionNotify:
      {
      // Draw a ghost-arc to indicate where node will eventually be placed
      if (move)
        {
        unsigned int keys;
        int root_x, root_y;
        int newx, newy;
        Window root, child;

        // Undraw this arc
        XSetFunction (xdisplay->Get(), xgc->Get(), GXequiv);
        XDrawArc (xdisplay->Get(), xwindow->Get(), xgc->Get(),
        grabbubble->x - BUBBLE_WIDTH/2, grabbubble->y - BUBBLE_HEIGHT/2, BUBBLE_WIDTH, BUBBLE_HEIGHT, 0, 360*64);
        XQueryPointer (xdisplay->Get(), report.xmotion.window, &root, &child,
                        &root_x, &root_y, &newx, &newy, &keys);

        // Draw this arc
        grabbubble->x += newx - startpos.x;
        if( !(report.xbutton.state & ShiftMask) )
          grabbubble->y += newy - startpos.y;
        XDrawArc (xdisplay->Get(), xwindow->Get(), xgc->Get(),
        grabbubble->x - BUBBLE_WIDTH/2, grabbubble->y - BUBBLE_HEIGHT/2, BUBBLE_WIDTH, BUBBLE_HEIGHT, 0, 360*64);
        XSetFunction (xdisplay->Get(), xgc->Get(), GXcopy);
        startpos.x = newx;
        if( !(report.xbutton.state & ShiftMask) )
          startpos.y = newy;
        }
      break;
      }

    ///////////////////////////////////////////////////////////////////////////////

    case KeyPress:
	key_count = XLookupString(&report.xkey, buffer, 10, &keysym, &compose);
	if( !relabel )
	  {
	  if( report.xkey.window == filename_button ) {
	    UpdateFilename( keysym, buffer, key_count );
            }
	  if( keysym == XK_p || keysym == XK_P )
	    {
	    if( move )
	      {
	      bdd->PrintCubes(grabnode);
	      cout << flush;
	      }
	    else if( function->isa() == BDDNODE )
	      {
	      bdd->PrintCubes((BDDnode*)function);
	      cout << flush;
	      }
	    }
	  break;
	  }
	keydown = TRUE;
	
	// done?
	if( keysym == XK_Return || keysym == XK_KP_Enter || keysym == XK_Linefeed )
	  {
	  relabel = FALSE;
          label_names[index_num][--string_length] = '\0';
          XClearWindow( xdisplay->Get(), labels[index_num] );

	  XClearWindow( xdisplay->Get(), xwindow->Get() );
	  expose();
	  }
	// back up?
	else if( keysym == XK_BackSpace || keysym == XK_Delete )
	  {
	  if( string_length > 1 )
	    {
	    label_names[index_num][--string_length] = '\0';
	    label_names[index_num][string_length - 1] = '_';
	    XClearWindow( xdisplay->Get(), labels[index_num] );
	    }
	  else
	    XBell( xdisplay->Get(), 50 );
	  }
	else if( keysym == XK_Down ) {	// reorder: swap with level above
	  if( index_num < maxdepth-1 ) {
	    SwapBDD( index_num, index_num+1 );
            index_num++;
	    }

	  }
	else if( keysym == XK_Up ) {	// reorder: swap with level below
	  if( index_num > 0 ) {
	    SwapBDD( index_num, index_num-1 );
            index_num--;
	    }
	  }
	else if( key_count + string_length >= MAX_LABEL + 1)
	  {
	  XBell(xdisplay->Get(), 50);
	  }
	else
	  {
	  label_names[index_num][string_length-1] = '\0';	// erase the _
	  strncat( label_names[index_num], buffer, key_count );
	  string_length+=key_count;
	  label_names[index_num][string_length-1] = '_';
	  label_names[index_num][string_length] = '\0';
	  XClearWindow( xdisplay->Get(), labels[index_num] );
	  }
	
	XDrawString( xdisplay->Get(), labels[index_num], xgc->Get(), 2, font_info->max_bounds.ascent+2, label_names[index_num], string_length );
	break;
	
    ///////////////////////////////////////////////////////////////////////////////

    case EnterNotify:
      InvertWindow( report.xcrossing.window );
      break;

    ///////////////////////////////////////////////////////////////////////////////

    case LeaveNotify:
      UnInvertWindow( report.xcrossing.window );
      break;

    ///////////////////////////////////////////////////////////////////////////////

    case KeyRelease:
      break;

    ///////////////////////////////////////////////////////////////////////////////

    default:
      break;
      }
    }

  // Exit child process
  return retval;	// retval tells whether the processed forked (==FORKDONE) or not(==CHILD)
  }

void
BubbleMgr::InvertWindow( Window win ) {
  int index;
  for( index = 0 ; index < NUM_BUTTONS ; index++ )
    if( win == buttons[index] )
      break;
  if( index == NUM_BUTTONS )
    return;
  XSetWindowBackground(xdisplay->Get(), win, xdisplay->Black());
  XClearWindow( xdisplay->Get(), win );
  XDrawString( xdisplay->Get(), win, xgc->GetInv(), 2,
	    font_info->max_bounds.ascent+2,
	    button_names[index], strlen(button_names[index]) );
  }
void
BubbleMgr::UnInvertWindow( Window win ) {
  int index;
  for( index = 0 ; index < NUM_BUTTONS ; index++ )
    if( win == buttons[index] )
      break;
  if( index == NUM_BUTTONS )
    return;
  XSetWindowBackground(xdisplay->Get(), win, xdisplay->White());
  XClearWindow( xdisplay->Get(), win );
  XDrawString( xdisplay->Get(), win, xgc->Get(), 2,
	    font_info->max_bounds.ascent+2,
	    button_names[index], strlen(button_names[index]) );
  }

  
void
BubbleMgr::MakeLabelWindows( )
  {
  int width = (MAX_LABEL+1) * font_info->max_bounds.width + 4;
  int height = font_info->max_bounds.ascent + font_info->max_bounds.descent + 4;

  label_names = new char*[maxdepth];

  labels = new Window[maxdepth];

  for(int i = 0 ; i< maxdepth ; i++ )
    {
    // add 1 to i because the zeroth level is reserved for the title
    labels[i] = XCreateSimpleWindow(xdisplay->Get(), xwindow->Get(), MIN_LEFT, (i+1) * vert - 10,
					width, height, LABEL_BORDER, xdisplay->Black(),
					xdisplay->White() );
    XMapWindow( xdisplay->Get(), labels[i]);

    XSelectInput( xdisplay->Get(), labels[i], ExposureMask |
					      KeyPressMask | KeyReleaseMask |
					      ButtonPressMask | ButtonReleaseMask );
    label_names[i] = new char[MAX_LABEL+1];
    label_names[i][0] = '\0';
    }
  }
  
void
BubbleMgr::MakeButtons()
  {
  int width = 4 * font_info->max_bounds.width + 4;
  int height = font_info->max_bounds.ascent + font_info->max_bounds.descent + 4;

  button_names[0] = "Exit";
  button_names[1] = "Cube";
  button_names[2] = "Fork";
  button_names[3] = "Reor";
  button_names[4] = "Save";
  int i;

  for( i=0 ; i<NUM_BUTTONS-2 ; i++ )
    {
    buttons[i] = XCreateSimpleWindow(xdisplay->Get(), xwindow->Get(),
			BUTTON_LEFT + (width + 6) * i, 10,
			width, height,
			LABEL_BORDER, xdisplay->Black(), xdisplay->White() );

    XSelectInput( xdisplay->Get(), buttons[i], ExposureMask |
				       KeyPressMask | KeyReleaseMask |
				       ButtonPressMask | ButtonReleaseMask |
				       EnterWindowMask | LeaveWindowMask );
    }
  XMapWindow( xdisplay->Get(), buttons[0] );
  XMapWindow( xdisplay->Get(), buttons[1] );
  if( function->isa() == BDDNODE )
    XMapWindow( xdisplay->Get(), buttons[2] );
  if( function->isa() == BDDNODE || function->isa() == LIST )
  XMapWindow( xdisplay->Get(), buttons[3] );
  XMapWindow( xdisplay->Get(), buttons[4] );
  button_names[5] = " OK ";
  button_names[6] = "CANC";
  for( i=0 ; i<2 ; i++ ) {
    buttons[i+NUM_BUTTONS-2] = XCreateSimpleWindow(xdisplay->Get(), xwindow->Get(),
			BUTTON_LEFT + (width + 6) * i, 22+height+height,
			width, height,
			LABEL_BORDER, xdisplay->Black(), xdisplay->White() );

    XSelectInput( xdisplay->Get(), buttons[i+NUM_BUTTONS-2], ExposureMask |
				       KeyPressMask | KeyReleaseMask |
				       ButtonPressMask | ButtonReleaseMask |
				       EnterWindowMask | LeaveWindowMask );
    }
  sprintf(filename, "bdd%d", (int)getpid());
  filename_button = XCreateSimpleWindow(xdisplay->Get(), xwindow->Get(),
	      BUTTON_LEFT , 16+height,
	      20* font_info->max_bounds.width + 4, height,
	      LABEL_BORDER, xdisplay->White(), xdisplay->Black() );

  XSelectInput( xdisplay->Get(), filename_button, ExposureMask |
				     KeyPressMask | KeyReleaseMask |
				     ButtonPressMask | ButtonReleaseMask );
  
  }


boolean
BubbleMgr::SaveBDD() {
  List savelist;
  if( function->isa() == BDDNODE ) 
    savelist.Add(function);
  else if( function->isa() == LIST )
    savelist.AppendList( (List*)function );
  else // but the save button shouldn't be visible
    cerr << "Can only save a single function or a list of functions\n";
  
  ofstream outFile( filename, ios::out );
  if( !outFile ) {
    cerr << "BDD Save operation failed to create file \""<<filename<<"\"\n";
    return FALSE;
    }
  bdd->WriteBinary( &savelist, outFile );
  outFile.close();
  cerr << "Saved \"" << filename << "\"\n";
  XUnmapWindow( xdisplay->Get(), filename_button );
  XUnmapWindow( xdisplay->Get(), buttons[5] );
  XUnmapWindow( xdisplay->Get(), buttons[6] );
  XMapWindow( xdisplay->Get(), buttons[4] );
  return TRUE;
  }
  
void
BubbleMgr::UpdateFilename( KeySym keysym, char* buffer, int key_count ) {
  if( keysym == XK_Return || keysym == XK_KP_Enter || keysym == XK_Linefeed ){
    SaveBDD();
    return;
    }
  int len = strlen(filename);
  if( keysym == XK_BackSpace || keysym == XK_Delete ) {
    if( len>0 )
      filename[len-1] = '\0';
    else
      XBell( xdisplay->Get(), 50 );
    }
  else {
    strncat( filename, buffer, key_count );
    filename[len+key_count] = '\0';
    }
  
  XClearWindow( xdisplay->Get(), filename_button );
  XDrawString( xdisplay->Get(), filename_button, xgc->GetInv(), 2,
      font_info->max_bounds.ascent+2, filename, strlen(filename) );

  }
  
void
BubbleMgr::FindNode( BubbleObj* position, BDDnode* &node )
  {
  
  if( node = (BDDnode*) coord2node->Get( position ))
    return;

  BubbleObj neighbor_check( position->x, position->y );

  if( position->x % xgrid >= xgrid / 2 )
    neighbor_check.x += xgrid;
  else
    neighbor_check.x -= xgrid;

  if( node = (BDDnode*) coord2node->Get( &neighbor_check ))
    return;


  if( position->y % ygrid >= ygrid / 2 )
    neighbor_check.y += ygrid;
  else
    neighbor_check.y -= ygrid;

  if( node = (BDDnode*) coord2node->Get( &neighbor_check ))
    return;

  neighbor_check.x = position->x;
  node = (BDDnode*) coord2node->Get( &neighbor_check );

  return;
  }

/////////////////////////////////////////////////////////////////////

void
BubbleMgr::HardRemove( BDDnode *n )
  {
  HasherTree *coord2node_save = new HasherTree;
  HasherTree *node2coord_save = new HasherSort;

  Hasher_ForEach_Type( coord2node, BDDnode, K, N )
    {
    if( N != n )
      {
      coord2node_save->Put( N, (BubbleObj*)K );
      node2coord_save->Put( (BubbleObj*)K, N );
      }
    }Hasher_End_ForEach

  delete coord2node;
  node2coord->SetDelete(FALSE,FALSE);
  delete node2coord;

  coord2node = coord2node_save;
  node2coord = node2coord_save;
  }

/////////////////////////////////////////////////////////////////////

void
BubbleMgr::SoftRemove( BDDnode *n )
  {
  BubbleObj *bubble;
  bubble = (BubbleObj*) node2coord->Get( n );

  bubble->flags = INVISIBLE;
  // adjust numnodes to be one less (one more invisible node)
  numnodes--;
  }

////////////////////////////////////////////////////////////////////////

void
BubbleMgr::SiftBDD() {
  int i, max = MAX(maxlevel, bdd->NumInputs());
  int *oldlevel = new int[max];
  int *newindex = new int[max];
  for( i=0 ; i<maxlevel ; i++ )
    oldlevel[i] = bdd->unique->LookupLevel(i);

  bdd->Reorder_Sift();
  
  for( i=0 ; i<maxlevel ; i++ )
    newindex[i] = bdd->unique->LookupIndex(i);
  
  char **label_names2 = new char*[maxdepth];

  int newdepth=0, olddepth;
  for( i=0 ; i<maxlevel ; i++ ) {
    olddepth = Depth[oldlevel[newindex[i]]]-1;
    if( olddepth < 0 ) continue;
    
    label_names2[newdepth] = label_names[olddepth];

    XClearWindow( xdisplay->Get(), labels[newdepth] );

    XDrawString( xdisplay->Get(), labels[newdepth], xgc->Get(), 2,
		 font_info->max_bounds.ascent+2, label_names2[newdepth],
		 strlen( label_names2[newdepth] ) );
    newdepth++;
    }
  
  delete coord2node;
  delete node2coord;

  coord2node = new HasherTree;
  node2coord = new HasherSort;

  delete label_names;
  label_names = label_names2;

  SetDepth(bdd);
  InitBubbles();
  XClearWindow( xdisplay->Get(), xwindow->Get() );
  expose();
  delete oldlevel;
  delete newindex;
  }


void
BubbleMgr::SwapBDD( int a, int b )
  {
  delete coord2node;
  delete node2coord;

  coord2node = new HasherTree;
  node2coord = new HasherSort;


  int level1, level2;
  for( level1=0 ; level1<maxlevel ; level1++)
    if( Depth[level1] == a+1 )
      break;

  for( level2=0 ; level2<maxlevel ; level2++)
    if( Depth[level2] == b+1 )
      break;


  bdd->Swap2Var( level1, level2 );


  SetDepth(bdd);
  InitBubbles();

  // assume it did it right, and just swap the labels
  SWAP( char*, label_names[a], label_names[b] );


  XClearWindow( xdisplay->Get(), labels[a] );
  XClearWindow( xdisplay->Get(), labels[b] );

  XDrawString( xdisplay->Get(), labels[a], xgc->Get(), 2, font_info->max_bounds.ascent+2,
      label_names[a], strlen( label_names[a] ) );   // -1 to remove the _

  XDrawString( xdisplay->Get(), labels[b], xgc->Get(), 2, font_info->max_bounds.ascent+2,
      label_names[b], strlen(label_names[b] ) );

  XClearWindow( xdisplay->Get(), xwindow->Get() );
  int i;
  for( i=0 ;label_names[b][i] != '_' ; i++ );

  label_names[b][i] = '\0';

  expose();
  label_names[b][i] = '_';
  }


XDisplay::XDisplay()
{
  name = NULL;
  display = XOpenDisplay (name);
  if (display == NULL)
    {
    cerr << "Cannot connect to X server " << XDisplayName(name) << endl;
    return;
    }
  screen_num = DefaultScreen (display);
  width = DisplayWidth (display, screen_num);
  height = DisplayHeight (display, screen_num);
  root = RootWindow (display, screen_num);
  black = BlackPixel(display, screen_num);
  white = WhitePixel(display, screen_num);
  }

XDisplay::~XDisplay()
  {
  XCloseDisplay(display);
  }

XWindow::XWindow(XDisplay *disp, char *windowName, int maxdepth)
  {
  display = disp;

  int x = 0;
  int y = 0;
  int borderWidth = 2;
  width = WWIDTH;
  if( (maxdepth+1)*vert > display->Height()/2 )
    height = display->Height()/2;
  else
    height = (maxdepth+1)*vert;
 
  win = XCreateSimpleWindow (display->Get(), display->Root(), x, y,
		width, height, borderWidth, display->Black(), display->White());

  XWMHints wmHints;
  XClassHint classHints;
  char *iconName = "BDD";
  XTextProperty name, icon;
  XSetWindowAttributes setwinattr;
  long unsigned valuemask;

  valuemask = CWBitGravity | CWSaveUnder | CWBackingStore ; //settings for faster graphics

  setwinattr.bit_gravity = NorthWestGravity; // on resize, place visible contents in NW corner
  setwinattr.save_under = True;
  setwinattr.backing_store = Always;

  XChangeWindowAttributes(display->Get(), win, valuemask, &setwinattr);


  sizeHints.flags = PPosition | PSize | PMinSize;
  sizeHints.min_width = 300;
  sizeHints.min_height = 200;

  if (XStringListToTextProperty (&windowName, 1, &name) == 0)
    {
    cerr << "Structure allocation for windowName failed." << endl;
    exit (1);
    }
  if (XStringListToTextProperty (&iconName, 1, &icon) == 0)
    {
    cerr << "Structure allocation for iconName failed." << endl;
    exit (1);
    }

  wmHints.initial_state = NormalState;
  wmHints.input = True;
  wmHints.flags = StateHint | InputHint;

  classHints.res_name = "BDD TREE";
  classHints.res_class = "BDD Tree";

  XSetWMProperties (display->Get(), win, &name, &icon, NULL, 0,
				&sizeHints, &wmHints, &classHints);

  XFree( name.value );
  XFree( icon.value );
  // These are the X events that we pay attention to
  XSelectInput (display->Get(), win, ExposureMask | ButtonPressMask |
		ButtonReleaseMask | PointerMotionHintMask | ButtonMotionMask |
 		StructureNotifyMask | KeyPressMask | KeyReleaseMask );
  }

XWindow::~XWindow()
  {
  XDestroyWindow (display->Get(), win);
  }

XGC::XGC(XDisplay *d, XWindow *w, XFontStruct *fi)
  {
  win = w;
  font_info = fi;

  // Create default Graphics Context
  unsigned long valuemask = GCFont | GCForeground | GCBackground | GCLineWidth;
  XGCValues values;

  values.font = font_info->fid;
  values.foreground = d->Black();
  values.background = d->White();
  values.line_width = 0;	// fastest line drawing
  gc = XCreateGC (d->Get(), win->Get(), valuemask, &values);

  values.foreground = d->White();
  values.background = d->Black();
  inv_gc = XCreateGC (d->Get(), win->Get(), valuemask, &values);
 
  }

XGC::~XGC() {
  XFreeFont (win->Display()->Get(), font_info);
  XFreeGC (win->Display()->Get(), gc);
  XFreeGC (win->Display()->Get(), inv_gc);
  }



void
BDDmgr::XDump(Object *bn, char* window_title, boolean show_ref, char **inames)
  {
  BubbleMgr* bmgr;

  if( bn == NULL)
    bn = unique;
  bmgr = new BubbleMgr( this, bn, window_title, show_ref, inames );
 
  int forker;
  if(bmgr->xdisplay->Get() == NULL)
    bmgr->Abort(this,bn);
  else {
    if( (forker = bmgr->EventLoop(NO_FORK)) == FORKDONE)
      {
      delete bmgr;
      exit(0);
      }
    if (forker == CHILD)	// no fork
      delete bmgr;
    // if forker == PARENT then do nothing (program continues on)
    }
  }

void BDDmgr::XDump_nowait(Object *bn, char *window_title, boolean show_ref, char **inames)
{
  BubbleMgr* bmgr;

  if( bn == NULL)
    bn = unique;
  bmgr = new BubbleMgr( this, bn, window_title, show_ref, inames );
 
  if(bmgr->xdisplay->Get() != NULL) {
   
    int forker;
    if( (forker = bmgr->EventLoop(FORK)) == FORKDONE) {
      // Child must wipe itself then exit when it is done taking a dump!
      delete bmgr;
      exit (0);
    }
  }
  else
    cout << "continuing\n" << flush;
}


void
BubbleMgr::Abort(BDDmgr *bddm,Object *f) {
  cout << "XDump failed.";
  if( f->isa() == BDDNODE ) {
    cout << " There are " << bddm->NumNodes((BDDnode*)f);
    cout << " nodes and " << bddm->NumCubes((BDDnode*)f);
    cout << " cubes.Print cubes y/n? ";
    char x=cin.get();
    while( x!='y' && x != 'n' )
      x=cin.get();
    if(x=='y')
      bddm->PrintCubes((BDDnode*)f);
    }
  }
