
#ifndef WINDOW_H
#define WINDOW_H

#include "bddm.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>

#include "classdef.h"
#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>

// vert is the vertical distance between nodes
#define vert 36
#define ygrid 8
// horz is the horizontal distance between nodes
#define horz 48
#define xgrid 8

// WWIDTH = initial width of window
#define WWIDTH 425	
#define BUBBLE_HEIGHT 16
#define BUBBLE_WIDTH 48
#define STRAIGHTEN_X_OFFSET 24
#define MARGIN_OFFSET 100
#define LABEL_MARGIN 60
#define LABEL_BORDER 1
#define MAX_LABEL 6
#define NUM_BUTTONS 7
#define BUTTON_LEFT 240

#define TITLE_X 20
#define TITLE_Y 20

// minimum x value allowed when moving nodes
#define MIN_LEFT 10
// flags for forking
#define NO_FORK 0
#define FORK 1
#define CHILD 0
#define PARENT 1
#define FORKDONE 2

// width and height of a terminal "one" or "zero" pointer box
#define BOX_SIZE 5
// distance away from the center node to put the terminal boxes
#define BOX_Y 20
#define BOX_X 15
// offset from center of bubble
#define PTR_OFF 4


#define RIGHT 0
#define LEFT 1

#define VISIBLE 1
#define INVISIBLE 2

#define BUBPRIME 7919
#define FONTNAME "-adobe-courier-medium-r-normal--11-80-100-100-m-60-iso8859-1"

class XGC;
class BubbleMgr;
class XDisplay;
class XWindow;


class XDisplay
{
	Display *display;
	char *name;
	int screen_num;
	int width, height;
	Window root;
	unsigned long black, white;
public:
	XDisplay();
	~XDisplay();

	Display *Get() { return display; }
	int Width() { return width; }
	int Height() { return height; }
	int ScreenNum() { return screen_num; }
	Window Root() { return root; }
	unsigned long Black() { return black; }
	unsigned long White() { return white; }
};

class XWindow
{
	XDisplay *display;
	Window win;
	XSizeHints sizeHints;
	int width, height;
public:
	XWindow(XDisplay *, char *, int w);
	~XWindow();
	Window Get() { return win; }
	XDisplay *Display() { return display; }
};

class XGC
{
	XWindow *win;
	GC gc, inv_gc;
	XFontStruct *font_info;
public:
	XGC(XDisplay *, XWindow *, XFontStruct *font_info);
	~XGC();
	GC &Get() { return gc; }
	GC &GetInv() { return inv_gc; }
	XFontStruct *GetFontInfo() { return font_info; }
};
#endif // WINDOW_H

class BubbleObj : public Comparable {

public:
  int flags;
//  BDDnode* node;
  char* ident;
  int x;
  int y;

  BubbleObj(int xpos, int ypos)	{
					flags = VISIBLE;
					ident = NULL;
					x = xpos;
					y = ypos;
					}
 ~BubbleObj() {}

  void Name( char *name)	{ ident = name; }

  virtual integer isa() { return BUBBLE_OBJ;}

  virtual int Compare( Comparable* c ) {
    ASSERT( c->isa() == BUBBLE_OBJ, "BubbleObj expected!");

    if( x / xgrid != ((BubbleObj*)c)->x / xgrid )
      return -1;
    if( y / ygrid != ((BubbleObj*)c)->y / ygrid )
      return 1;
    return 0;
   } 

  virtual unsigned Hash( unsigned i=0 ) {
    return ( (x / xgrid)^((y/ygrid)<<3)^(i<<7)%BUBPRIME );
    }
  friend class BubbleMgr;
  };


class BubbleMgr : public Object {

  void            SetUpBDD( BDDmgr* oldbdd, Object* dag);
public:
  // X stuff
  XDisplay	 *xdisplay;
  XWindow	 *xwindow;
  XGC		 *xgc;
  char		 *fname;
  XFontStruct	 *font_info;
  Window	 *labels;
  Window	 buttons[NUM_BUTTONS];
  Window         filename_button;


  // algorithm stuff
  HasherTree	 *coord2node;
  HasherTree	 *node2coord;
  int		 *NumInLevel;
  int		 *Depth;
  int		  maxdepth;
  int		  numnodes;
  int		  numcubes;
  int		  maxlevel;
  char		**indexNames;		// provided by call to XDump
  char		**label_names;		// allocated inside
  char		 *button_names[NUM_BUTTONS];
  char           filename[100];
  boolean	 show_ref_count;

  // BDD stuff
  BDDmgr	 *bdd;
  Object	 *function;


  BubbleMgr( BDDmgr*, Object*, char*, boolean show_ref, char **inames);

 ~BubbleMgr();
  void		  InitBubbles(BDDmgr *oldbdd);
  void		  InitBubbles();

  virtual	  integer isa() { return BUBBLEMGR; }
  virtual	  void Print( ostream& out = cout );

  void		  AddNodeS( BDDnode* function );
  void		  AddNode( BDDnode* n);
  void		  SetDepth( BDDmgr *oldbddm );
  void		  BDD2Hash( BDDnode* f, HasherTree* ht, BDDmgr *oldbdd);
  void		  expose();
  void		  DrawNode( BubbleObj* b, BDDnode* n, XWindowAttributes *xwa );
  void		  drawConnections( int x, int y, BDDnode* mynode );
  int		  EventLoop(int flag);
  void		  HardRemove( BDDnode* erase_node );
  void		  SoftRemove( BDDnode* erase_node );
  void		  FindNode( BubbleObj*, BDDnode*& );
  void		  Straighten( BDDnode* n );
  void		  StraightenR( BDDnode* n, int x, int y, HasherTree* ht );
  void		  MakeLabelWindows();
  void		  MakeButtons();
  void		  SwapBDD( int a, int b );
  void		  SiftBDD( );
  void		  Abort(BDDmgr *, Object *o);
  void		  InvertWindow( Window win );
  void		  UnInvertWindow( Window win );
  boolean         SaveBDD();
  void            UpdateFilename( KeySym, char *, int );

  };



