#include <fstream.h>
#include <iostream.h>
#include "zbdd.h"
 
// this line doesn't make much sense, since srand return void, so
// I am executing srand in ZBDDmanager init... --ALC
//  integer ZRand = srand((integer) time((time_t*) NULL));
integer ZRand;

/////////////////////////////////////////////////////////////
//
//  ZBDDnode structure : ( index, right, left )
//                       root
//                        /\
//                       /  \
//                     Left  Right
//                   True()  False()
//	Michael Mo
//	Modified by: F. Brewer, I. Radivojevic, C. Monahan
//	Copyright Regents of the University of California, 1994
/////////////////////////////////////////////////////////////

void
ZBDDnode::Print(ostream& out) {

  out << "ZBDDnode " << (void *) this 
      << " v=" << index << " ref=" << NumReferences() << "\n";
  out << "  true=" << (void *) Then() << " false=" << (void *) Else()  << "\n";

}

ZBDDmanager::ZBDDmanager(integer ) {

  one  = new ZBDDnode(-1, 0, 0, TRUE);  //if both edges point to one.. dc
  zero = new ZBDDnode(-1, 0, 0, FALSE); //if both edges point to zero.. not dc
  num_inputs = 0;
  num_nodes  = 2;
  num_cubes  = 0;
  unique     = new HasherTree*[MAXINDEX+1];
  computed   = new HashCache(ZBDD_PRIME, FALSE, TRUE);
  srand((integer) time((time_t*) NULL));
  ZRand = rand();

}

ZBDDmanager::~ZBDDmanager() {
  delete one;
  delete zero;
  for( int i=0 ; i<num_inputs ; i++ )
    {
    unique[i]->SetDelete(TRUE,FALSE);	// delete the zbddnode during tree deletion
    delete unique[i];
    }
  delete unique;
  computed->Clear();
  delete computed;
  }


ZBDDnode*
ZBDDmanager::GetNode( integer v, ZBDDnode *g, ZBDDnode *h ) {

  ZBDDnode *p;

  ASSERT( v >= 0, "Negative variable index!");

  ASSERT( v > g->index && v > h->index ,
	  "GetNode index smaller than children's.");

  if( g == zero ) return h;   //Node elimination

  boolean    dc = ( g == h && h->isDC() && h->Index()+1 == v);
  ZBDDnode*  r = new ZBDDnode(v, g, h, dc);
  while( v>=num_inputs )
    unique[num_inputs++]=new HasherTree;

  if(  p = (ZBDDnode*) unique[v]->Get( r) ) {    //If already in the unique table
    delete r;
    return p;
  } else{
    unique[v]->Put(*r, r);                    //Input new node in unique table
    num_nodes++;                           //# of nodes increases
    Keep(g);
    Keep(h);
    return r;
  }
    
}


ZBDDnode*
ZBDDmanager::Subset1( ZBDDnode* P, integer var) {
  
  ZBDDcomputed2 *c;
  ZBDDnode* f;
  Object* o;

  ASSERT( var >= 0, "Negative variable index!");

  if( P->index < var ) return zero;

  if( P->index == var ) 
    return P->Then();
  else {
    c = new ZBDDcomputed2( SUBSET1, P, var );
    if( computed->Get( &o, c )) {
      delete c;
      return (ZBDDnode*) o;
    }
    else
      f =  GetNode( P->Index(), 
		    Subset1(P->Then(), var),
		    Subset1(P->Else(), var));
  }
  computed->Put( f, c );
  return f;
}



ZBDDnode*
ZBDDmanager::Subset0( ZBDDnode* P, integer var) {
  
  ZBDDcomputed2 *c;
  ZBDDnode* f;
  Object* o;

  ASSERT( var >= 0, "Negative variable index!");

  if( P->index < var ) return P;

  if( P->index == var ) 
    return ( P->Else() );
  else {
    c = new ZBDDcomputed2( SUBSET0, P, var );
    if( computed->Get( &o, c )) {
      delete c;
      return (ZBDDnode*) o;
    }
    else
      f = GetNode( P->Index(), 
		   Subset0( P->Then(), var),
		   Subset0( P->Else(), var));
  }
  computed->Put( f, c );
  return f;
}
 
ZBDDnode*
ZBDDmanager::Change( ZBDDnode* P, integer var) {

  ZBDDcomputed2 *c;
  ZBDDnode* f;
  Object* o;

  ASSERT( var >= 0, "Negative variable index!");

  c = new ZBDDcomputed2( CHANGE, P, var);
  if( computed->Get( &o, c )) {
    delete c;
    return (ZBDDnode*) o;
  }

  if( P->index < var )
    f = GetNode( var, P, zero);
  else if ( P->index == var ) 
    f = GetNode( var, 
		 P->Else(), 
		 P->Then());
  else 
    f = GetNode( P->Index(),
		 Change(P->Then(), var),
		 Change(P->Else(), var));

  computed->Put( f, c );
  return f;
}


ZBDDnode*
ZBDDmanager::Union( ZBDDnode* P, ZBDDnode* Q) {
  
  ZBDDcomputed1 *c;
  ZBDDnode* f;
  Object* o;

  if( P == zero ) return Q;
  if( Q == zero ) return P;
  if( P == Q ) return P;
  
  if( P->isDC() && P->Index() >= Q->Index()) return P;
  if( Q->isDC() && Q->Index() >= P->Index()) return Q;

  c = new ZBDDcomputed1( UNION, P, Q);
  if( computed->Get( &o, c )) {
    //ASSERT( (((ZBDDnode*) o)->index >= P->index)&&(((ZBDDnode*) o)->index >= Q->index), "Invalid Cache Hit in Union!");
    delete c;
    return (ZBDDnode*) o;
  }
  if( P->index > Q->index )
    f = GetNode( P->index, P->Then(), Union( P->Else(), Q) );
  else if ( P->index < Q->index )
    f = GetNode( Q->index, Q->Then(), Union( P, Q->Else()) );
  else
    f =  GetNode( P->Index(),
                  Union(P->Then(), Q->Then()),
                  Union(P->Else(), Q->Else()));

  computed->Put( f, c );
  return f;

}

ZBDDnode*
ZBDDmanager::Intsect( ZBDDnode* P, ZBDDnode* Q ) {

  ZBDDcomputed1 *c;
  ZBDDnode* f;
  Object* o;

  if( P == zero ) return zero;
  if( Q == zero ) return zero;
  if( P == Q ) return P;
  
  if( P->isDC() && P->Index() >= Q->Index()) return Q;
  if( Q->isDC() && Q->Index() >= P->Index()) return P;

  c = new ZBDDcomputed1( INTSECT, P, Q);
  if( computed->Get( &o, c )) {
    delete c;
    return (ZBDDnode*) o;
  }

  if ( P->index < Q->index )
    f = Intsect( P, Q->Else() );
  else if( P->index == Q->index )
    f = GetNode( P->Index(),
		Intsect(P->Then(), Q->Then()),
		Intsect(P->Else(), Q->Else()));
  else 
    f = Intsect( P->Else(), Q );

  computed->Put( f, c );
  return f;
}


ZBDDnode*
ZBDDmanager::Diff( ZBDDnode* P, ZBDDnode* Q ) {

  ZBDDcomputed1 *c;
  ZBDDnode* f;
  Object* o;

  if( P == zero ) return zero;
  if( Q == zero ) return P;
  if( P == Q ) return zero;

  c = new ZBDDcomputed1( DIFF, P, Q);
  if( computed->Get( &o, c )) {
    delete c;
    return (ZBDDnode*) o;
  }
  if( P->index > Q->index )
    f = GetNode(P->Index(), 
		P->Then(),
		Diff(P->Else(), Q));
  else if( P->index < Q->index )
    f =  Diff( P, Q->Else());
  else 
    f = GetNode(P->Index(),
		Diff(P->Then(), Q->Then()),
		Diff(P->Else(), Q->Else()));

  computed->Put( f, c );
  return f;
}


integer
ZBDDnode::Compare(Comparable *p) {
 
  ASSERT( p->isa() == ZBDDNODE, "not a ZBDDnode Object");
 
  ZBDDnode *n = (ZBDDnode *)p;
 
  if (( index == n->index) && (left == n->left) &&
      ( right == n->right))
    return 0;
  if ((integer)this > (integer)n)
    return 1;
  if ((integer)this < (integer)n)
    return -1;
  return 0;
}


//
// NodesInFunction, PathsInFunction (i.e. Count)...
//

integer
ZBDDmanager::NodesInFunction(ZBDDnode* f) {

  integer n;
  HasherTree *ht    = new HasherTree(FALSE);
  NodesInFunctionR(f, ht);
  n = ht->Count();
  delete ht;
  return n;
}

void
ZBDDmanager::NodesInFunctionR(ZBDDnode *f, HasherTree *ht) {
  
  if (f == one || f == zero)
    return;
  if (ht->Put(*f,f)) {
    if (f->Then())
      NodesInFunctionR(f->Then(), ht);
    if (f->Else())
      NodesInFunctionR(f->Else(), ht);
  } else
    return;
}


float
ZBDDmanager::Count( ZBDDnode* P ) {
  HasherTree *ht    = new HasherTree(TRUE);
  float       count = CountR( P, ht );

  delete ht;
  return count;
}

float
ZBDDmanager::CountR( ZBDDnode* P, HasherTree* ht) {
  Object *o;

  if (P == zero ) 
    return 0;
  if (P == one ) 
    return 1;
  if (ht->Get(&o, P))
    return ((FloatObj*)o)->val;

  float n;
  n = CountR(P->Then(), ht) +
      CountR(P->Else(), ht);
    
  ht->Put(new FloatObj(n), P);
  return n;
}

void
ZBDDmanager::PrintSets( ZBDDnode* P, ostream& out) {

  int i;
  
  ASSERT( P, "No Zbdd to print!" );

  out << "{\n";
  if ( P == one )
    out << "Base\n";
  else if ( P == zero )
    out << "Empty\n";
  else {
    string set = new char[num_inputs + 1];
    for (i = 0; i < num_inputs ; ++i)
      set[i] = '0';
    set[i] = '\0';
    PrintSetsR( P, set, out );
  }
  out << "}\n";
}

void
ZBDDmanager::PrintSetsR( ZBDDnode* P, string set, ostream& out) {
  
 
  if( P == one ) {
    out << set << "\n";
    num_cubes++;
    return;
  } else 
    if ( P == zero ) return;
    else {
      char c;
      c = set[ P->index ];
      set[ P->index ] = '1';
      PrintSetsR( P->Then(), set, out);
      set[ P->index ] = '0';
      PrintSetsR( P->Else(), set, out);
      set[ P->index ] = c;
    }
}
 
///////////////////////////////////////////////////////////////////////
//
// Universe routine generates universal set vith variables from
// [0, universe_top]
//

ZBDDnode*
ZBDDmanager::Universe(integer universe_top) {

  ZBDDnode*	temp_node = GetNode(0,one,one);
  for(integer i=1; i<=universe_top; i++)
    temp_node = GetNode(i,temp_node,temp_node);
  return temp_node;
}


///////////////////////////////////////////////////////////////////////
//
// B2Z (BDD-to-ZBDD) routine converts OBDD to ZBDD. Input parameter
// 'universe_top' provides info about the possible 'don't care' 
// variables that have index higher than the one corresponding to
// the top variable in OBDD function (f->Indedx()).
//

ZBDDnode*
ZBDDmanager::B2Z( integer universe_top, BDDnode* f, BDDmgr* bdd ) {

  HasherTree* ht = new HasherTree;
  ZBDDnode* zf = ZBDDCover( f, bdd, ht );
  delete ht;

  for(integer i = f->GetIndex()+1; i <= universe_top; i++ )
    zf = GetNode( i, zf, zf);
  return  zf;	
}	


ZBDDnode*
ZBDDmanager::ZBDDCover( BDDnode* g, BDDmgr* bdd , HasherTree* ht ) { 

  ZBDDnode *result;
  ZBDDnode *t, *f;
  integer  i;	
  Object*  o;

  if( g == bdd->Zero() )   return( Empty() );
  if( g == bdd->One() )    return( Base() );

  if (ht->Get(&o, g)) 
    return (ZBDDnode*) o;

  if( g->Then() == bdd->Zero() ) {
    f =  ZBDDCover( g->Else(), bdd, ht);
    for( i = (g->Else()->GetIndex())+1; i < g->GetIndex(); i++ )
      f =  GetNode( i, f, f);
    result = f;
  } else {
    t = ZBDDCover( g->Then(), bdd, ht ) ;
    for( i = (g->Then()->GetIndex()) +1; i < g->GetIndex(); i++ )
      t = GetNode( i, t, t );	
    f = ZBDDCover( g->Else(), bdd, ht );
    for( i = (g->Else()->GetIndex())+1; i < g->GetIndex(); i++ )
      f = GetNode( i, f, f );
    result = GetNode( g->GetIndex(), t, f );
  }

  ht->Put( result, g);
  return result;
}


/////////////////////////////////////////////////////////////////////// 
//
// Z2B (ZBDD-to-BDD) routine converts ZBDD to OBDD. It calls BDDcover
// routine to perform the actual conversion and appends the OBDD with
// complemented variables with indices from (f->Index ... universe_top]  
//

BDDnode*
ZBDDmanager::Z2B(integer universe_top,  ZBDDnode* g, BDDmgr* bdd) {
	
  HasherTree* ht = new HasherTree;
  BDDnode* temp = BDDCover( g, bdd, ht);
  delete ht;
	
  for (integer i=(g->Index()+1); i<=universe_top; i++) 
    temp = bdd->and( temp, bdd->not( bdd->Input(i)));

  return temp;
}

// BDDCover converts ZBDD f into a BDD built with BDDmanager m.
// It uses MakeNode-right-left traversal. Before the BDDmanager builds
// a node on the next level, it checks ZBDD for missing nodes and 
// does insertion if necessary. Before BDDmanager travers to the right,
// it checks ZBDD for right=left. BDDCover returns a BDDnode by using
// the BDDmanager::ite() functions to maintain ROBDD. 

BDDnode*
ZBDDmanager::BDDCover( ZBDDnode* g, BDDmgr* bdd, HasherTree* ht) {

  if(g==zero) return bdd->Zero();
  if(g==one)  return bdd->One();

  if( g->isDC())                       // check for tautology
    return bdd->One();

  Object*  o;
  if (ht->Get(&o, g))                  // check if we've seen it before
    return (BDDnode*) o;
 
  if ( g->Then() == g->Else() &&      // check if present node is don't care
      (g->Then()->Index()) == (g->Index()-1))
    return BDDCover( g->Then(), bdd, ht);

  integer  i;
  BDDnode* t = BDDCover( g->Then(),  bdd, ht);
  BDDnode* f = BDDCover( g->Else(), bdd, ht);

  for ( i= (g->Then()->Index())+1; i<g->Index(); i++)
    t = bdd->and( t, bdd->not( bdd->Input(i)));

  for ( i= (g->Else()->Index())+1; i<g->Index(); i++)
    f = bdd->and( f, bdd->not( bdd->Input(i)));

  BDDnode* result = bdd->MakeNode( g->Index(), t, f);
  ht->Put( result, g);
  return result;
}



void
ZBDDmanager::PrintNodes( ZBDDnode* P, ostream& out ) {

  if( P == zero  || P == one ){
    out << (void*)P <<"\n";
    exit(1);
  }  else {
    P->Print();
    if( P->left != one )
      PrintNodes( P->Then() );
    if( P->right != one && P->right != zero ) 
      PrintNodes( P->Else() );
  }
}

void
ZBDDmanager::TermPrint(ostream& out) {

  out << "Terminal Nodes: \n"
      << "one = " << (void *) one << "\t"
      << "zero = " << (void *) zero << "\n";
}


void
ZBDDmanager::PrintStats(ostream& out) {
/*
  out << "\n\nUNIQUE TABLE:\n\n";
  unique->ShortPrint(out);
//  unique->Print(out);
*/
  out << "\n\nCOMPUTED CACHE:\n";
  computed->Print(out);
}

///////////////////////////////////////////////////////////////////          
//
//  routine to perform garbage collection - ZBDDnodes supporting only
//  'unneeded' functions are deleted.
//
//
 
 
void
ZBDDmanager::GC() {

  ZRand = (integer) lrand48();
  for(integer i=(num_inputs-1); i>=0; i--)
    if(unique[i]->Count()!=0) 
      doGC(i);
  computed->Clear();
}
 
void
ZBDDmanager::ForceGC() {

  ZRand = (integer) lrand48();
  for(integer i=(num_inputs-1); i>=0; i--) 
    if(unique[i]->Count()!=0)
      doGC(i);
  computed->Clear();
}
 
void
ZBDDmanager::doGC(integer i) {

//  ZRand = (integer) lrand48();  
  ZBDDnode *node;
 
  HasherTree *unique_save = new HasherTree;
 
  Hasher_ForEach_Object(*unique[i], key, object) {
    node = (ZBDDnode *) key;
    if (node->NumReferences() == 0 && ! node->Marked())
      RGC(node);
  } Hasher_End_ForEach;
 
  Hasher_ForEach_Object(*unique[i], key, object) {
    node = (ZBDDnode *) key;
    if (node->NumReferences() > 0) {
      unique_save->Put(*node, node);
      node->UnMark();
    } else {
      ASSERT(node->Marked(), "deleting an unmarked node");
      delete node;
      --num_nodes;
    }
  } Hasher_End_ForEach;
 
  delete unique[i];
  unique[i] = unique_save;
//  computed->Clear();
}

void
ZBDDmanager::RGC(ZBDDnode *i) {
 
  if (i == one || i == zero)
    return;
 
  integer a, b;
  i->Mark();
 
  ZBDDnode *t = (ZBDDnode*) i->Then();
  ZBDDnode *e = (ZBDDnode*) i->Else();
 
  a = t->NumReferences();
  t->UnReference();
  if (a == 1)
    RGC(t);
 
  b = e->NumReferences();
  e->UnReference();
  if (b == 1)
    RGC(e);
}

///////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////
//
//    ZBDDcomputed functions
//


integer
ZBDDcomputed1::Compare(Comparable *p) {

  if (p->isa() == ZBDDCOMPUTED2) return -1;
  ASSERT( p->isa() == ZBDDCOMPUTED1, "not a BDDcomputed Object");
  
  ZBDDcomputed1 *n = (ZBDDcomputed1 *)p;
  
  if ((f == n->f) && (g == n->g) && ( type == n->type))
    return 0;
  if ((integer)this > (integer)n)
    return 1;
  if ((integer)this < (integer)n)
    return -1;
  return 0;
}

////////////////////////////////////////////////////////

integer
ZBDDcomputed2::Compare(Comparable *p) {

  if (p->isa() == ZBDDCOMPUTED1) return 1;
  ASSERT( p->isa() == ZBDDCOMPUTED2, "not a BDDcomputed Object");
  
  ZBDDcomputed2 *n = (ZBDDcomputed2 *)p;
  
  if ((f == n->f) && (type == n->type) && (var == n->var))
    return 0;
  if ((integer)this > (integer)n)
    return 1;
  if ((integer)this < (integer)n)
    return -1;
  return 0;
}


/////////////////////////////////////////////////////////////////////
//
// ZBDD Unate Product as described in "Binary Decision Diagrams and 
// Applications For VLSI CAD" Minato page 75

ZBDDnode*
ZBDDmanager::Product( ZBDDnode* P, ZBDDnode* Q ) {

  if (P == zero || Q == zero)                
    return zero;
  if (P == one)
    return Q;
  if (Q == one)
    return P;

  if( P->index < Q->index) return Product(Q,P);    

  ZBDDcomputed1* c = new ZBDDcomputed1( PRODUCT, P, Q);
  Object* o;
  if( computed->Get( &o, c )) {
    //ASSERT( (((ZBDDnode*) o)->index >= P->index)&&(((ZBDDnode*) o)->index >= Q->index), "Invalid Cache Hit in Product!");
    delete c;
    return (ZBDDnode*) o;
  }

  integer x = P->index;

  ZBDDnode* P0=Subset0(P,x);
  ASSERT( x > P0->index, "P0 index is larger than x!");
  ZBDDnode* P1=Subset1(P,x);
  ASSERT( x > P1->index, "P1 index is larger than x!");
  ZBDDnode* Q0=Subset0(Q,x);
  ASSERT( x > Q0->index, "Q0 index is larger than x!");
  ZBDDnode* Q1=Subset1(Q,x);
  ASSERT( x > Q1->index, "P1 index is larger than x!");

  ZBDDnode* t1=Product(P1,Q1);
  ASSERT( x > t1->index, "t1 index is larger than x!");
  ZBDDnode* t2=Product(P1,Q0);
  ASSERT( x > t2->index, "t2 index is larger than x!");
  ZBDDnode* t3=Product(P0,Q1);
  ASSERT( x > t3->index, "t3 index is larger than x!");
  ZBDDnode* t4=Product(P0,Q0);
  ASSERT( x > t4->index, "t4 index is larger than x!");
  ZBDDnode* sum1=Union(t1,t2);
  ASSERT( x > sum1->index, "sum1 index is larger than x!");

  ZBDDnode* sum2=Union(sum1,t3);
  ASSERT( x > sum2->index, "sum2 index is larger than x!");

  ZBDDnode* r=GetNode(x,sum2,t4);

  computed->Put( r, c );
  return r;
}



