#ifndef ZBDD_H
#define ZBDD_H
 
////////////////////////////////////////////////////
//
// zbdd.h : C++ ZBDD (0-Suppressed-BDD) package based
//          on paper by Minato SASIMI 93'.
//
//	Michael Mo
//	Modified by: F. Brewer, I. Radivojevic, C. Monahan
//	Copyright Regents of the University of California, 1994
// 
//
///////////////////////////////////////////////////

#include "bddm.h"
#include "hashcach.h"
#include "floatobj.h"

#define	ZBDD_PRIME  104729	
#define UNION       1 
#define INTSECT     2 
#define SUBSET1     3
#define SUBSET0     5
#define CHANGE      7
#define DIFF        11
#define PRODUCT     13 

#define MAXINDEX   	4095	

class ZBDDnode;
class ZBDDmanager;
class ZBDDcomputed1;
class ZBDDcomputed2;
//**
  extern integer ZRand;

class ZBDDnode : public Tnode {
  integer index;
  boolean dc;     // if dc is true, all variables are don't care
  
public:

  ZBDDnode(integer v, ZBDDnode *a, ZBDDnode *b, boolean dcare)
           {index=v; left=a; right=b; dc=dcare;}
  ZBDDnode( ZBDDnode* a ) 
	   { index=a->index; left=a->left; right=a->right; dc=a->dc;}
  ~ZBDDnode() {;}

  //Virtual hash functions
  virtual unsigned Hash( unsigned h = 0 ) {
    return ( ((integer)index)   //<<  5)
	   ^ ((integer)Then()   <<  5)
	   ^ ((integer)Else()   << 11)
           ^ ZRand
	     ^ (h << 1))  % ZBDD_PRIME;
  }
  
  virtual integer isa() { return ZBDDNODE;}
  virtual integer Compare(Comparable *);
  integer Index() {return index;}
  boolean isDC()  {return dc;}
  ZBDDnode* Then() { return (ZBDDnode *) left;}
  ZBDDnode* Else() { return (ZBDDnode *) right;}

  friend class ZBDDmanager;
  
  void Print( ostream& out = cout );
};

class ZBDDmanager : public Object {

protected:

  ZBDDnode *one;           // terminal 1
  ZBDDnode *zero;          // terminal 0
  integer num_inputs;      // total number of variables (levels)      
  integer num_nodes;       // total number of allocated nodes
  integer num_cubes;       // total number of cubes (1-pathes)
  

  HasherTree** unique;         //Keeps each node unique
  HashCache *computed;         //Cache for computed nodes

  float CountR( ZBDDnode* P, HasherTree* ht);
  void NodesInFunctionR(ZBDDnode *f, HasherTree *ht);
   
  void doGC(integer);
  void RGC(ZBDDnode*);

public:

  BDDmgr *robdd;

  ZBDDmanager( integer n=0 );
  ~ZBDDmanager();


  integer NumInputs()   { return num_inputs; }
  integer NumLiterals() { return num_nodes; }
  integer NumCubes()    { return num_cubes; }
  integer NumNodes()    { return num_nodes; }   

  ZBDDnode* Empty()     { return zero; }
  ZBDDnode* Base()      { return one;  }

//	Generates a node with index v, and subgraphs g and h
  ZBDDnode* GetNode( integer v, ZBDDnode* g, ZBDDnode* h );
  ZBDDnode* InputNode( integer v ){ return GetNode( v, one, zero );}

  ZBDDnode* Subset1( ZBDDnode* P, integer var ); //Returns the 1 subset of P at var
  ZBDDnode* Subset0( ZBDDnode* P, integer var ); //Returns the 0 subset of P at var
  ZBDDnode* Change(  ZBDDnode* P, integer var );  //Returns P when var is inverted
//
  ZBDDnode* ExistentialAbstraction( ZBDDnode* P, integer var ) {
	ZBDDnode* temp = Union(Subset0(P,var),Subset1(P,var));
	return ( Union(temp,Change(temp,var)));
  }
//
  ZBDDnode* Union( ZBDDnode *P, ZBDDnode *Q );   //Returns P * Q
  ZBDDnode* Intsect( ZBDDnode* P, ZBDDnode* Q ); //Returns P + Q
  ZBDDnode* Diff( ZBDDnode* P, ZBDDnode* Q );    //Returns P - Q
  float     Count( ZBDDnode* P );                //Returns # of elements
  float     PathsInFunction(ZBDDnode* P) { return Count( P ); }
  integer   NodesInFunction(ZBDDnode* P);

  ZBDDnode* One()  { return one; }
  ZBDDnode* Zero() { return zero; }

  ZBDDnode* And( ZBDDnode* P, ZBDDnode* Q ) { return Intsect( P, Q ); }
  ZBDDnode* Or ( ZBDDnode* P, ZBDDnode* Q ) { return Union( P, Q ); }

  ZBDDnode* Product( ZBDDnode* P, ZBDDnode* Q );

//  ZBDDnode* Read( istream& in=cin );

//  ZBDDnode* ISOP( BDDnode* P );

  ZBDDnode* Universe(integer);
//  ZBDDnode* Distinguish(BDDmgr *bdd, BDDnode *s1, BDDnode *s2, int max, HasherTree *ht);
//  ZBDDnode* DistinguishR(BDDnode *s1, BDDnode *s2, int max, int univ, HasherTree *ht, ZBDDnode **t);

  BDDnode*  Z2B(integer, ZBDDnode* P, BDDmgr* M);
  ZBDDnode* B2Z(integer, BDDnode* f, BDDmgr* bdd);
  BDDnode*  BDDCover ( ZBDDnode* P, BDDmgr* bdd, HasherTree* ht);
  ZBDDnode* ZBDDCover(  BDDnode* f, BDDmgr* bdd, HasherTree* ht);
//  ZBDDnode* ReadPLA( char* filename );

  void Keep  ( ZBDDnode *P) { P->Reference(); }
  void UnKeep( ZBDDnode *P) { P->UnReference(); }
  void ForceGC();
  void GC();
  
  void PrintComputed() { computed->Print(); }
  void PrintSets( ZBDDnode* P, ostream& out = cout );   //Prints Comb. Sets
  void PrintSetsR( ZBDDnode* P, string set, ostream& out = cout );
  void PrintNodes( ZBDDnode* P, ostream& out = cout );  //Prints ZBDD structure P
  void TermPrint( ostream& out = cout );         //Prints terminal nodes

  void PrintStats( ostream& out = cout );        //unique/computed stats... 
};


class ZBDDcomputed1 : public Comparable {

  ZBDDnode *f, *g;
  integer type;

public:
  
  ZBDDcomputed1(integer i, ZBDDnode *x, ZBDDnode* y)
    { f    = MaxNode( x,y ); 
      g    = MinNode( x,y ); 
      type = i;}
  ZBDDcomputed1( ZBDDcomputed1 *p )
    { f    = p->f;
      g    = p->g;
      type = p->type;
    }
  ~ZBDDcomputed1() {;}

// utility functions
  ZBDDnode* MaxNode( ZBDDnode* P, ZBDDnode* Q ) {   //Returns the larger of two
    return ((integer)P > (integer)Q)? P:Q;
  }

  ZBDDnode* MinNode( ZBDDnode* P, ZBDDnode* Q ) {  //Returns the smaller of two
    return ((integer)P < (integer)Q)? P:Q;
  }

  virtual unsigned Hash( unsigned n=0 ){
    return( ((integer) type) //  << 5 )   
	    ^ ((integer) f   << 5)
	    ^ ((integer) g   << 11)
	    ^ ( n << 1 )) % ZBDD_PRIME;
  }

  virtual integer isa() { return ZBDDCOMPUTED1;}
  virtual integer Compare(Comparable *);    
};


class ZBDDcomputed2 : public Comparable {

  ZBDDnode *f;
  integer type, var;

public:
  
  ZBDDcomputed2(integer i, ZBDDnode *x, integer j)
    { f    = x; 
      type = i; 
      var  = j; }
  ZBDDcomputed2( ZBDDcomputed2 *p )
    { f    = p->f;
      var  = p->var;
      type = p->type;
    }
  ~ZBDDcomputed2() {;}

  virtual unsigned Hash( unsigned n=0 ){
    return( ((integer) type << 5 )
	   ^ ((integer) f   << 7)
	   ^ ((integer) var << 11)
	    ^ ( n << 8 )) % ZBDD_PRIME;
  }

  virtual integer isa() { return ZBDDCOMPUTED2;}
  virtual integer Compare(Comparable *);
};

#endif


