#include "zbdd.h"

////////////////////////////////////////////
// 
//       SMALL COMBINATORIAL PROBLEM
//
//   (see "cout" statements for comments) 
//  
////////////////////////////////////////////

XZBDDmanager*	xzbdd;
BDDmanager*	bdd;
integer		universe_top;


main() {

  void	PrintInfo(XZBDDnode*);
  void  CheckGC(integer);
  
  xzbdd = new XZBDDmanager;
  bdd   = new BDDmanager;

  integer GC_level;
  cout << "\n\nEnter GC level: "; cin >> GC_level;
  
  XZBDDnode* X0 = xzbdd->InputNode(0);
  XZBDDnode* X1 = xzbdd->InputNode(1);
  XZBDDnode* X2 = xzbdd->InputNode(2);
  XZBDDnode* X3 = xzbdd->InputNode(3);
  
  universe_top=xzbdd->NumInputs()-1; 

  cout << "\n\n\nEMPTY SET:\n";
  PrintInfo(xzbdd->Empty());
  
  cout << "\n\n\nUNIVERSE:\n";
  XZBDDnode*   universe = xzbdd->Universe(universe_top);
  xzbdd->Keep(universe);
  PrintInfo(universe);
	 
  cout << "\n\n\nUNIVERSE (Positive Half):\n";
  PrintInfo(xzbdd->Subset1(universe,universe_top));
  cout << "\n\n\nUNIVERSE (Negative Half):\n";
  PrintInfo(xzbdd->Subset0(universe,universe_top));

  cout << "\n\n\nONE_OUT_OF_FOUR:\n";
  XZBDDnode* one_of_four = xzbdd->Union(X0,X1);
  one_of_four = xzbdd->Union(one_of_four,X2);
  one_of_four = xzbdd->Union(one_of_four,X3);
  PrintInfo(one_of_four);
  
  XZBDDnode*	one_or_less = xzbdd->Union(one_of_four,xzbdd->Base());
  cout << "\n\n\nONE_OR_LESS:\n";
  PrintInfo(one_or_less);
 
  XZBDDnode*	more_than_one = xzbdd->Diff(universe,one_or_less);
  cout << "\n\n\nMORE_THAN_ONE:\n";
  PrintInfo(more_than_one);

  XZBDDnode*   two_of_four;
  two_of_four = xzbdd->Change(X0,1);
  two_of_four = xzbdd->Union(two_of_four,xzbdd->Change(X0,2));
  two_of_four = xzbdd->Union(two_of_four,xzbdd->Change(X0,3));
  two_of_four = xzbdd->Union(two_of_four,xzbdd->Change(X1,2));
  two_of_four = xzbdd->Union(two_of_four,xzbdd->Change(X1,3));
  two_of_four = xzbdd->Union(two_of_four,xzbdd->Change(X2,3));
  cout << "\n\n\nTWO_OUT_OF_FOUR:\n";
  PrintInfo(two_of_four);
  
  XZBDDnode*	two_or_less = xzbdd->Union(two_of_four,one_or_less);
  cout << "\n\n\nTWO_OR_LESS:\n";
  PrintInfo(two_or_less);

  XZBDDnode*	more_than_two = xzbdd->Diff(universe,two_or_less);
  cout << "\n\n\nMORE_THAN_TWO:\n";
  PrintInfo(more_than_two);

  xzbdd->PrintStats(cout);
  CheckGC(GC_level);
  xzbdd->PrintStats(cout);
}


void	PrintInfo(XZBDDnode* Z) {
  
  xzbdd->PrintSets(Z);
  cout << "#nodes (XZBDD): " << xzbdd->NodesInFunction(Z) << "\n";
  cout << "#paths (XZBDD): " << xzbdd->PathsInFunction(Z) << "\n";

  BDDnode* print_node;
  print_node = xzbdd->Z2B(universe_top,Z,bdd);
  cout << "\nOBDD form:\n";
  bdd->PrintCubes(print_node);
  cout << "#nodes (OBDD): " << bdd->NodesInFunction(print_node) << "\n";
  cout << "#paths (OBDD): " << bdd->PathsInFunction(print_node) << "\n";
  cout.flush();

  XZBDDnode* new_Z = xzbdd->B2Z(universe_top,print_node,bdd);
  cout << "\n... CONVERT BACK TO XZBDD AND COMPARE TO ORIGINAL... ";
  if(Z==new_Z)
    cout << "OK!\n";
  else
    cout << "FAILED!\n";
  xzbdd->PrintSets(new_Z);
  cout.flush();
}

void  CheckGC(integer i) {
  
  if(xzbdd->NumNodes()>i) {
    cout << "\n\n#nodes before GC: " << xzbdd->NumNodes();
    xzbdd->ForceGC();
    cout << "\n\n#nodes after GC: " << xzbdd->NumNodes();
  }
}

   
   
